//******************************************************************************
//*                                  uBee512                                   *
//*         An emulator for the Microbee Z80 FDD and ROM based models          *
//*                                                                            *
//*                         Floppy disk support module                         *
//*                                                                            *
//*                   Copyright (C) 2007-2008 Stewart Kay                      *
//******************************************************************************
//
// This module implements functions to represent the actual physical disks.
// It consists of support functions that manipulate the disk image files.
//
// To do:
//    - Check that seek position is within range of image.  If not then put
//      message out and flag as an error.
//
//==============================================================================
// ChangeLog (most recent entries are at top)
//==============================================================================
// v2.3.0 - 21 January 2008, S.J.Kay
// - Added preliminary support for direct floppy access for Unices.
// - Added modio_t structure.
//
// v2.2.0 - 12 December 2007, S.J.Kay
// - added image write protect feature.  A trailing underscore '_' character
//   makes the image write protected.
// - disk->imagerec.datatrack value changed to 1 for non DS80 formats. Not
//   an issue as the system and data tracks have the same sector offsets for
//   either tracks anyway.
//
// v2.1.0 - 24 October 2007, S.J.Kay
// - Added new function disk_read_idfield() to obtain the ID field from the
//   track, side.
// - Undid the changes made in v2.0.0 to the imagerec.datatrack back to how it
//   is in v1.4.0.
// - Microbee DS80 format fixed,  this format has 4 reserved cylinders not 2.
//   disk->imagerec.datatrack = 4
//
// v2.0.0 - 18 October 2007, S.J.Kay
// - DSK format should now be fully complete.
// - Changes to imagerec.datatrack used by all raw and DIP formats now holds
//   the byte offset instead of number of tracks for first data track.
//
// v1.4.0 - 6 October 2007, S.J.Kay
// - Added alternative shortened 3 character filename extentions for raw images.
// - The '.dsk' format now includes 3.5" Modular and should work with other
//   unknown disk formats.  Sector offset information is now read per track.
// - Fixed bug preventing '.DSnn' and '.SSnn' raw images from being opened.
//   Changed image_types[][5] to be image_types[][6].
//
// v1.1.0 - 14 August 2007, S.J.Kay
// - Added Microbee ds40, ss80, ds80, ds82, ds84 raw disk image types.
// - Modified DSK code to detect all Microbee formats.  This needs testing for
//   some formats. (warning in README)
//
// v1.0.0 - 19 June 2007, S.J.Kay
// The following changes/additions have been made:
// - Added new disk image structures,  this now allows for different floppy
//   images to be used instead of only the 40T DS (SBC) images.
// - disk_open function now uses open with write enabled (r+b). This is
//   required to allow disk_write to work.
// - Added an assert to test max_size variable in disk_read function.
// - Implement the disk_write function.
//
// v0.0.0 - 7 June 2007, S.J.Kay
// Start with "nanowasp" source distribution version 0.22. An emulator for the
// microbee 128k. Copyright (C) 2000-2003  David G. Churchill.

//==============================================================================
/*
 *  uBee512 - An emulator for the Microbee Z80 FDD and ROM based models. 
 *  Copyright (C) 2007-2008 Stewart Kay
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/ 
//==============================================================================

#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <string.h>

#include "disk.h"
#include "ubee512.h"

//==============================================================================
// structures and variables
//==============================================================================
extern modio_t modio;

//==============================================================================
// Disk open.
//
// Determine the disk format and structure being used.  This can one of the
// following:
//
// * original v0.22 raw disk format (40T DS DD) Side0=1st 200K, Side1=2nd 200K
// * CPC-EMU dsk format
// * dip format (disk image plus) This is an in-house spec not finalised.
// * ds40 - Microbee 40T DS DD (SBC) S0-T0, S1-T0, S0-T1, S1-T1 ...
// * ss80 - Microbee 80T SS DD (CIAB) S0-T0, S0-T1, S0-T2 ...
// * ds80 - Microbee 80T DS DD (modular) S0-T0, S1-T0, S0-T1, S1-T1 ...
// * ds82 - Microbee 80T DS DD (dream) S0-T0, S1-T0, S0-T1, S1-T1 ...
// * ds84 - Microbee 80T DS DD (pjb) S0-T0, S1-T0, S0-T1, S1-T1 ...
//
//   pass: const char *name
//         disk_t *disk
// return: int				0 if no error, else -1 
//==============================================================================
int disk_open (const char *name, disk_t *disk)
{
 int type_start = 0;
 
 dski_t dski;			// dsk information block
 dskt_t dskt;			// dsk track information block

 char uname[512];
 
 char const image_types[][6] =
 {
  ".DIP",			// Disk Image Plus (in-house format)
  ".DSK",			// standard CPC DSK format
  ".IMG",			// Nanowasp v0.22 format
  ".NW",			// Nanowasp v0.22 format
  ".DS40",			// raw images
  ".D40",  
  ".SS80",
  ".S80",
  ".DS80",
  ".D80",  
  ".DS82",
  ".D82",  
  ".DS84",
  ".D84",
  ""
 };

// make an upper case copy of the file name  
 int i = 0;
 strcpy(uname, name);
 while (uname[i])
    {
     uname[i] = toupper(uname[i]);
     i++;
    }

// image is write protected if a trailing '_' is found
 disk->wrprot = uname[strlen(uname)-1] == '_';
 if (disk->wrprot)
    uname[strlen(uname)-1] = 0;		//remove the '_'

 assert(disk != NULL);

// direct floppy access. ie /dev/fd0u800/.ds84
#ifdef MINGW
#else
 if (strstr(name, "/dev/") == name)
    {
     char *ch = strstr(name, "/.");
     if (ch)
        *ch = 0;
     type_start = 4;   
    }
#endif 

// open the image of floppy
 disk->fdisk = fopen(name, "r+b");
 if (disk->fdisk == NULL)
    return -1;

// find the disk type
 for (i = type_start; image_types[i][0] != 0; i++)
    {
     if (strstr(uname, image_types[i]) != NULL)
        break;
    }
 if (image_types[i][0] == 0)
    return -1;    

 disk->itype = i;

// configure parameters for the disk type 
 switch (disk->itype)
    {
     case 0 : // DIP (in house spec)
              fseek(disk->fdisk, -sizeof(disk->imagerec), SEEK_END);
              fread(&disk->imagerec, sizeof(disk->imagerec), 1, disk->fdisk);
              if (strcmp(disk->imagerec.id, "DISK-IMAGE-PLUS") != 0)
                 return -1;
              if ((disk->imagerec.secsize != 128) && (disk->imagerec.secsize != 256) &&
                 (disk->imagerec.secsize != 512) && (disk->imagerec.secsize != 1024))
                 return -1;
              if (! disk->wrprot)   // if not already '_' write protected
                 disk->wrprot = disk->imagerec.wrprot;
#ifdef use_info_disk_open
              printf("Image: %s   Vers: %s   Type: %s\n", disk->imagerec.id,
              disk->imagerec.ver, disk->imagerec.type);
#endif
              break;
     case 1 : // DSK
// get the disk information block
              fread(&dski, sizeof(dski), 1, disk->fdisk);
              disk->imagerec.tracks = dski.tracks;
              disk->imagerec.heads = dski.heads; 

// get the first track information block to get some initial values
              fread(&dskt, sizeof(dskt), 1, disk->fdisk);
              disk->imagerec.sectrack = dskt.spt;
              disk->imagerec.secsize = 128 << dskt.bps;
              disk->imagerec.systsecofs = dskt.sect_numb;
#if 0
              printf("dskt.sect_numb=%d\n", dskt.sect_numb);
              printf("dskt.hnumb_sectid=%d\n", dskt.hnumb_sectid);
#endif

#ifdef use_info_disk_open
              printf("Image format is DSK\n");
#endif
              break;
     case 2 : // IMG (nanowasp v0.22 format)
     case 3 : // NW (nanowasp v0.22 format)
              disk->imagerec.tracks = 40;
              disk->imagerec.heads = 2;
              disk->imagerec.sectrack = 10;
              disk->imagerec.secsize = 512;           
              disk->imagerec.datatrack = 1;
              disk->imagerec.datasecofs = 1;
              disk->imagerec.systsecofs = 1;

#ifdef use_info_disk_open
              printf("Image format is RAW 40T DS DD\n");
#endif
              break;
     case 4 : // DS40
     case 5 : // D40
              disk->imagerec.tracks = 40;
              disk->imagerec.heads = 2;
              disk->imagerec.sectrack = 10;
              disk->imagerec.secsize = 512;           
              disk->imagerec.datatrack = 1;
              disk->imagerec.datasecofs = 1;
              disk->imagerec.systsecofs = 1;
#ifdef use_info_disk_open
              printf("Image format is RAW DS40\n");
#endif
              break;
     case 6 : // SS80
     case 7 : // S80     
              disk->imagerec.tracks = 80;
              disk->imagerec.heads = 1;
              disk->imagerec.sectrack = 10;
              disk->imagerec.secsize = 512;           
              disk->imagerec.datatrack = 1;
              disk->imagerec.datasecofs = 1;
              disk->imagerec.systsecofs = 1;

#ifdef use_info_disk_open
              printf("Image format is RAW SS80\n");
#endif
              break;
     case 8 : // DS80 (Microbee modular)
     case 9 : // D80 (Microbee modular)     
              disk->imagerec.tracks = 80;
              disk->imagerec.heads = 2;
              disk->imagerec.sectrack = 10;
              disk->imagerec.secsize = 512;           
              disk->imagerec.datatrack = 2;         // 4 reserved tracks
              disk->imagerec.datasecofs = 21;       // sectors are 21 to 30
              disk->imagerec.systsecofs = 1;

#ifdef use_info_disk_open
              printf("Image format is RAW DS80\n");
#endif
              break;
     case 10 : // DS82
     case 11 : // D82     
              disk->imagerec.tracks = 80;
              disk->imagerec.heads = 2;
              disk->imagerec.sectrack = 10;
              disk->imagerec.secsize = 512;           
              disk->imagerec.datatrack = 1;
              disk->imagerec.datasecofs = 1;
              disk->imagerec.systsecofs = 1;
#ifdef use_info_disk_open
              printf("Image format is RAW DS82\n");
#endif
              break;
     case 12 : // DS84
     case 13 : // D84     
              disk->imagerec.tracks = 80;
              disk->imagerec.heads = 2;
              disk->imagerec.sectrack = 10;
              disk->imagerec.secsize = 512;           
              disk->imagerec.datatrack = 1;
              disk->imagerec.datasecofs = 1;
              disk->imagerec.systsecofs = 1;
#ifdef use_info_disk_open
              printf("Image format is RAW DS84\n");
#endif
              break;
    } 

#ifdef use_info_disk_open
     printf("Tracks: %d   Heads: %d   S/T: %d   B/S: %d\n", 
     disk->imagerec.tracks, disk->imagerec.heads,
     disk->imagerec.sectrack, disk->imagerec.secsize);
#endif

 return 0;    
}

//==============================================================================
// Disk close.
//
//   pass: disk_t *disk
// return: void
//==============================================================================
void disk_close (disk_t *disk)
{
 assert(disk != NULL);

 fclose(disk->fdisk);
 disk->fdisk = NULL;
}

//==============================================================================
// Disk create.
//
//   pass: const char *name
//         disk_t *disk
// return: int				-1
//==============================================================================
int disk_create(const char *name, disk_t *disk)
{
 return -1;
}

//==============================================================================
// Disk read.
//
//   pass: disk_t *disk
//         char *buf
//         int max_size
//         int side
//         int track
//         int sect
// return: int				1 
//==============================================================================
int disk_read (disk_t *disk, char *buf, int max_size, int side, int track, int sect)
{
 dskt_t dskt;			// dsk track information block

 int trackofs;
 int sectofs;
 int sectuse;
 int dskofs;

 assert(disk != NULL);
 assert(disk->fdisk != NULL);

#ifdef use_debug_disk_read_abort
 printf("disk_read: forcing an abort here\n");
 assert(1 != 1);
#endif

 switch (disk->itype)
    {
     case 0 : // DIP (in house spec)
              trackofs = track * disk->imagerec.heads * disk->imagerec.sectrack * disk->imagerec.secsize;
              if (track >= disk->imagerec.datatrack)
                 sectuse = sect - disk->imagerec.datasecofs;
              else
                 sectuse = sect - disk->imagerec.systsecofs;
              sectofs = side * disk->imagerec.sectrack * disk->imagerec.secsize + (sectuse * disk->imagerec.secsize);

              fseek(disk->fdisk, trackofs + sectofs, SEEK_SET);
              fread(buf, disk->imagerec.secsize, 1, disk->fdisk);
#ifdef use_info_disk_read
              printf("Image: %s   Vers: %s   Type: %s\n", disk->imagerec.id,
              disk->imagerec.ver, disk->imagerec.type);
#endif
              break;
     case 1 : // DSK
              trackofs = track * disk->imagerec.heads * disk->imagerec.sectrack * disk->imagerec.secsize;
              dskofs = (((track * disk->imagerec.heads) + side + 1) * 0x100) + 0x100;

// read the track information header for the track
              sectofs = side * disk->imagerec.sectrack * disk->imagerec.secsize;
              fseek(disk->fdisk, trackofs + sectofs + dskofs - 0x100, SEEK_SET);
              fread(&dskt, sizeof(dskt), 1, disk->fdisk);
              disk->imagerec.datasecofs = dskt.sect_numb;
              disk->imagerec.systsecofs = dskt.sect_numb;
              sectuse = sect - disk->imagerec.datasecofs;              
//            printf("\nTrk: %d   Sde: %d   Sec: %d\n",  track, side, sect);
//            printf("imagerec.systsecofs=%d  imagerec.datasecofs=%d\n", disk->imagerec.systsecofs, disk->imagerec.datasecofs);

// seek and read the sector
              sectofs = side * disk->imagerec.sectrack * disk->imagerec.secsize + (sectuse * disk->imagerec.secsize);
              fseek(disk->fdisk, trackofs + sectofs + dskofs, SEEK_SET);
              fread(buf, disk->imagerec.secsize, 1, disk->fdisk);
              break;
     case 2 : // IMG (nanowasp v0.22 format)
     case 3 : // NW (nanowasp v0.22 format)
              fseek(disk->fdisk, ((side * 40 + track) * 10 + (sect - 1)) * 512, SEEK_SET);
              fread(buf, 512, 1, disk->fdisk);
              break;
     case 4 : // DS40
     case 5 : // D40     
     case 6 : // SS80
     case 7 : // S80     
     case 8 : // DS80 (Microbee modular)
     case 9 : // D80 (Microbee modular)     
     case 10 : // DS82
     case 11 : // D82     
     case 12 : // DS84
     case 13 : // D84

//            printf("Trk: %d   Sde: %d   Sec: %d\n",  track, side, sect);

              trackofs = track * disk->imagerec.heads * disk->imagerec.sectrack * disk->imagerec.secsize;
              sectofs = side * disk->imagerec.sectrack * disk->imagerec.secsize;

              if (track >= disk->imagerec.datatrack)
                 sectuse = sect - disk->imagerec.datasecofs;
              else
                 sectuse = sect - disk->imagerec.systsecofs;

              sectofs += sectuse * disk->imagerec.secsize;
        
              fseek(disk->fdisk, trackofs + sectofs, SEEK_SET);
              fread(buf, disk->imagerec.secsize, 1, disk->fdisk);
              break;
    }  

#ifdef use_info_disk_read
 printf("Tracks: %d   Heads: %d   S/T: %d   B/S: %d\n", 
 disk->imagerec.tracks, disk->imagerec.heads,
 disk->imagerec.sectrack, disk->imagerec.secsize);
 printf("Trk: %d   Sde: %d   Sec: %d\n",  track, side, sect);
 if (disk->itype == 1)
    printf("File offset: %d\n\n", trackofs + sectofs + dskofs);
 else   
    printf("File offset: %d\n\n", trackofs + sectofs);
#endif
 
 return 1;
}

//==============================================================================
// Disk write.
//
// A flush call is used on each sector write.  This may be moved out at a later
// stage when image (floppy) swapping is implemented.
//
//   pass: disk_t *disk
//         char *buf
//         int size
//         int side
//         int track
//         int sect
// return: int				1 
//==============================================================================
int disk_write (disk_t *disk, char *buf, int size, int side, int track, int sect)
{
 dskt_t dskt;			// dsk track information block

 int trackofs;
 int sectofs;
 int sectuse; 
 int dskofs;
 
 assert(disk != NULL);
 assert(disk->fdisk != NULL);

 switch (disk->itype)
    {
     case 0 : // DIP (in house spec)
              trackofs = track * disk->imagerec.heads * disk->imagerec.sectrack * disk->imagerec.secsize;
              if (track >= disk->imagerec.datatrack)
                 sectuse = sect - disk->imagerec.datasecofs;
              else
                 sectuse = sect - disk->imagerec.systsecofs;       
              sectofs = side * disk->imagerec.sectrack * disk->imagerec.secsize + (sectuse * disk->imagerec.secsize);
     
              fseek(disk->fdisk, trackofs + sectofs, SEEK_SET);
              fwrite(buf, disk->imagerec.secsize, 1, disk->fdisk);
              fflush(disk->fdisk);
#ifdef use_info_disk_read
              printf("Image: %s   Vers: %s   Type: %s\n", disk->imagerec.id,
              disk->imagerec.ver, disk->imagerec.type);
#endif
              break;
     case 1 : // DSK
              trackofs = track * disk->imagerec.heads * disk->imagerec.sectrack * disk->imagerec.secsize;
              dskofs = (((track * disk->imagerec.heads) + side + 1) * 0x100) + 0x100;

// read the track information header for the track
              sectofs = side * disk->imagerec.sectrack * disk->imagerec.secsize;
              fseek(disk->fdisk, trackofs + sectofs + dskofs - 0x100, SEEK_SET);
              fread(&dskt, sizeof(dskt), 1, disk->fdisk);
              disk->imagerec.datasecofs = dskt.sect_numb;
              disk->imagerec.systsecofs = dskt.sect_numb;
              sectuse = sect - disk->imagerec.datasecofs;
//            printf("\nTrk: %d   Sde: %d   Sec: %d\n",  track, side, sect);
//            printf("imagerec.systsecofs=%d  imagerec.datasecofs=%d\n", disk->imagerec.systsecofs, disk->imagerec.datasecofs);

// seek and write the sector
              sectofs = side * disk->imagerec.sectrack * disk->imagerec.secsize + (sectuse * disk->imagerec.secsize);
              fseek(disk->fdisk, trackofs + sectofs + dskofs, SEEK_SET);
              fwrite(buf, disk->imagerec.secsize, 1, disk->fdisk);
              fflush(disk->fdisk);
              break;
     case 2 : // IMG (nanowasp v0.22 format)
     case 3 : // NW (nanowasp v0.22 format)
              fseek(disk->fdisk, ((side * 40 + track) * 10 + (sect - 1)) * 512, SEEK_SET);
              fwrite(buf, 512, 1, disk->fdisk);
              fflush(disk->fdisk);
              break;
     case 4 : // DS40
     case 5 : // D40     
     case 6 : // SS80
     case 7 : // S80     
     case 8 : // DS80 (Microbee modular)
     case 9 : // D80 (Microbee modular)     
     case 10 : // DS82
     case 11 : // D82     
     case 12 : // DS84
     case 13 : // D84
              trackofs = track * disk->imagerec.heads * disk->imagerec.sectrack * disk->imagerec.secsize;
              sectofs = side * disk->imagerec.sectrack * disk->imagerec.secsize;

              if (track >= disk->imagerec.datatrack)
                 sectuse = sect - disk->imagerec.datasecofs;
              else
                 sectuse = sect - disk->imagerec.systsecofs;

              sectofs += sectuse * disk->imagerec.secsize;

              fseek(disk->fdisk, trackofs + sectofs, SEEK_SET);
              fwrite(buf, disk->imagerec.secsize, 1, disk->fdisk);
              fflush(disk->fdisk);
              break;
    }  

#ifdef use_info_disk_read
 printf("Tracks: %d   Heads: %d   S/T: %d   B/S: %d\n", 
 disk->imagerec.tracks, disk->imagerec.heads,
 disk->imagerec.sectrack, disk->imagerec.secsize);
 printf("Trk: %d   Sde: %d   Sec: %d\n",  track, side, sect);
 if (disk->itype == 1)
    printf("File offset: %d\n\n", trackofs + sectofs + dskofs);
 else   
    printf("File offset: %d\n\n", trackofs + sectofs);

#endif
    
 return 1;    
}

//==============================================================================
// disk read ID field.
//
// Fills the read_addr_t structure id with the ID field
//
//   pass: disk_t *disk
//         read_addr_t *idfield
//         int side
//         int track
// return: void
//==============================================================================
void disk_read_idfield (disk_t *disk, read_addr_t *idfield, int side, int track)
{
 dskt_t dskt;			// dsk track information block

 int trackofs;
 int sectofs;
 int dskofs;

 idfield->track = track;
 idfield->side = side;
 idfield->seclen = disk->imagerec.secsize >> 8; 
 idfield->crc1 = 0xff;
 idfield->crc2 = 0xff;

 if (disk->itype == 1)          // if DSK format
    {
     trackofs = track * disk->imagerec.heads * disk->imagerec.sectrack * disk->imagerec.secsize;
     dskofs = (((track * disk->imagerec.heads) + side + 1) * 0x100) + 0x100;

// read the track information header for the track
     sectofs = side * disk->imagerec.sectrack * disk->imagerec.secsize;
     fseek(disk->fdisk, trackofs + sectofs + dskofs - 0x100, SEEK_SET);
     fread(&dskt, sizeof(dskt), 1, disk->fdisk);

     idfield->secaddr = dskt.sect_numb;
    }
 else   
    {
     if (track >= disk->imagerec.datatrack)
        idfield->secaddr = disk->imagerec.datasecofs;
     else
        idfield->secaddr = disk->imagerec.systsecofs;
    }
}

//==============================================================================
// disk get sector header ???.
//
//   pass:
// return:
//==============================================================================

/* int disk_getsecthdr(...) ?? */

//==============================================================================
// Disk write protected query.
//
//   pass: disk_t *disk
// return: int				write protect status
//==============================================================================
int disk_iswrprot (disk_t *disk)
{
   if (disk == NULL)
      return 0;

   return disk->wrprot;
}
