/* Floppy Disk Support Header */

#ifndef HEADER_DISK_H
#define HEADER_DISK_H

#include <inttypes.h>

#include "ubee512.h"

// Disk image record information (adjust filler for 512 bytes)
struct diski_s
{
 char id[32];		// disk image ID
 char ver[16];		// disk image version
 char type[32];		// disk image type

 uint16_t wrprot;	// write protected
 uint16_t tracks;	// number of tracks
 uint16_t heads;	// number of heads (sides)
 uint16_t secsize;	// bytes per sector
 uint16_t sectrack;	// sectors per track
 uint16_t datatrack;	// start of data tracks (after system)
 uint16_t systsecofs;	// offset to apply to system sectors
 uint16_t datasecofs;	// offset to apply to data sectors
 uint8_t skewsa[3];	// skew algorithm info for system tracks
 uint8_t skewda[3];	// skew algorithm info for data tracks
 uint8_t skewsd[128]; 	// skew data for system tracks
 uint8_t skewdd[128]; 	// skew data for data tracks
 uint16_t secterrs;	// size of sector error table (0 if not used) 
 char fill[152];	// filler to make structure 512 bytes long
};

struct disk_s
{
 FILE *fdisk;
 int wrprot;
 int itype;		// image type 0=raw, 10=dsk, 20=dip
 struct diski_s imagerec;
};

// dsk disk structure, first 0x100 bytes of image is header the disk
// information block.
struct dski_s
{
 char cpcemu[0x22];		// 0x00 - 0x21
 char unused1[0x0e];		// 0x22 - 0x2f
 uint8_t tracks;		// 0x30
 uint8_t heads;			// 0x31
 uint8_t size_one_ta;		// 0x32
 uint8_t size_one_tb;		// 0x33 
 char unused2[0x100-0x34];      // 0x34 - 0xff
};

// dsk disk structure, track information block. We use this to figure out
// the SPT and BPS for all other tracks on the disk.
struct dskt_s
{
 char track_info[0x0d];		// 0x00 - 0x0c
 char unused3[0x03];            // 0x0d - 0x0f
 uint8_t tracks_n;		// 0x10
 uint8_t heads_n;		// 0x11
 char unused4[0x02];		// 0x12 - 0x13
 uint8_t bps;			// 0x14
 uint8_t spt;			// 0x15
 uint8_t gap3;			// 0x16
 uint8_t filling_byte;		// 0x17
 uint8_t track_numb;		// 0x18
 uint8_t hnumb_sectid; 		// 0x19
 uint8_t sect_numb;		// 0x1a
 uint8_t bps_x;			// 0x1b
 uint8_t state1_errcode; 	// 0x1c
 uint8_t state2_errcode;  	// 0x1d
 char unused5[0x02];		// 0x1e - 0x1f
 char unused6[0x100-0x20];	// 0x20 - 0xff
};

typedef struct read_addr_t
{
 uint8_t track;
 uint8_t side;
 uint8_t secaddr;
 uint8_t seclen;
 uint8_t crc1;
 uint8_t crc2;
}read_addr_t;

typedef struct dski_s dski_t;
typedef struct dskt_s dskt_t;

typedef struct disk_s disk_t;

int disk_open (const char *name, disk_t *disk);
void disk_close (disk_t *disk);
int disk_create (const char *name, disk_t *disk);

int disk_read (disk_t *disk, char *buf, int max_size, int side, int track, int sect);
int disk_write (disk_t *disk, char *buf, int size, int side, int track, int sect);

void disk_read_idfield (disk_t *disk, read_addr_t *idfield, int side, int track);

/* int disk_getsecthdr(...) ?? */
int disk_iswrprot (disk_t *disk);

#endif  /* HEADER_DISK_H */
