//******************************************************************************
//*                                  uBee512                                   *
//*         An emulator for the Microbee Z80 FDD and ROM based models          *
//*                                                                            *
//*                              function module                               *
//*                                                                            *
//*                   Copyright (C) 2007-2008 Stewart Kay                      *
//******************************************************************************
//
// Built in functions able to be called from Z80 code.
//
// General purpose functions that don't belong to any particular module.
//
//==============================================================================
// ChangeLog (most recent entries are at top)
//==============================================================================
// v2.3.0 - 23 January 2007, S.J.Kay
// - Added function_emu_command() function.
// - Added printf, hex and decimal output functions to the stdio group.
// - Added dump function to diagnostics group.
// - Added additional functions to status group.
// - Some function values that are intended to return 0 or 1 were missing
//   the little_endian conversion, so 1 would have returned as some other non
//   zero value on big endian platforms. If the test in an application was for
//   a non zero value then there won't be a problem.  Constant values of 0 now
//   also use the little_endian conversion for consistency.
// - Added modio_t debug_t and regdump_t structures.
//
// v2.2.0 - 11 December 2007, S.J.Kay
// - Added status, control, keyboard, stdio, and file and directory functions.
// - Changed the method for Application and Version reporting.
// - Added a Z80 memory dump function to stdio.
//
// v1.4.0 - 22 September 2007, S.J.Kay
// - Added copy_file() function.
//
// v1.2.0 - 19 August 2007, S.J.Kay
// - Created new module to implement the emulator function ports.

//==============================================================================
/*
 *  uBee512 - An emulator for the Microbee Z80 FDD and ROM based models. 
 *  Copyright (C) 2007-2008 Stewart Kay
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/ 
//==============================================================================

#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <dirent.h>
#include <sys/stat.h>

#ifdef MINGW
#include <time.h>
#else
#include <sys/times.h>
#endif

#include "ubee512.h"
#include "function.h"
#include "gui.h"
#include "crtc.h"
#include "rtc.h"
#include "z80.h"
#include "mz80.h"

int dump_addr;
static int dump_addr_x;

static const char emulator_id_str[] = APPIDSTR;  // defined in the Makefile
static const char emulator_ver_str[] = APPVER;   // defined in the Makefile
static int emulator_id_pos;

static int ubee_req;
static int ubee_addr;
static int ubee_command;

extern int key_tape;

extern int model;
extern model_t modelx;
extern int key_reset;
extern int done;

extern char z80mem[];

extern int z80_blocks;
extern modio_t modio;
extern debug_t debug;
extern regdump_t regdump;

//==============================================================================
// Copy a file from one location to another.
//
//   pass: char *dest			destination file path and name
//         char *src			source file path and name
// return: int				0 if no error, else -1
//==============================================================================
int copy_file (char *dest, char *src)
{
 FILE *srcfp;
 FILE *destfp; 
 char *buf[0x4000];
 int len;

 srcfp = fopen(src, "rb");
 if (srcfp == NULL)
    return -1;

 destfp = fopen(dest, "wb"); 
 if (destfp == NULL)
    {
     fclose(srcfp);		// close the source file
     return -1;
    }

 while (! feof(srcfp))
    {
     len = fread(buf, 1, sizeof(buf), srcfp);
     if (len)
        fwrite(buf, 1, len, destfp);     
    } 
 
 fclose(srcfp);
 fclose(destfp); 

 return 0;
}

//==============================================================================
// Test if system uses Big Endian or Little Endian format.
//
//   pass: void
// return: int				1 if Big Endian, 0 if Little Endian
//==============================================================================
static int IsBigEndian (void)
{
#if defined(__aix__)
 return _true;
#else
 int i = 1;
 int c = *((char*)&i);
 return c == 0;
#endif
}

//==============================================================================
// Swap Endianess. Only works for even number variables sizes.
//
//   pass: char *n
//         int size			size in bytes of the variable
// return: void
//==============================================================================
void swap_endian (char *n, int size) 
{
 int i;
 int x;

 size /= 2;
 
 for (i = 0; i < size; i++)
    {
     x = *(n+i);
     *(n+i) = *(n+(size-1)-i);
     *(n+(size-1-i)) = x;
    } 
}

//==============================================================================
// Little Endian to host endian,  and host endian to Little Endian functions.
// Swaps byte order if on a Big Endian host.
//
//   pass: xintxx_t n			little endian, or host endian format
// return: xintxx_t			host endian, or little endaan format
//==============================================================================
int16_t little_endian_16 (int16_t n)
{
 if (IsBigEndian())
    swap_endian((char *)&n, sizeof(n));
 return(n);
}
          
uint16_t little_endian_u16 (uint16_t n)
{
 if (IsBigEndian())
    swap_endian((char *)&n, sizeof(n));
 return(n);
}

int32_t little_endian_32 (int32_t n) 
{
 if (IsBigEndian())
    swap_endian((char *)&n, sizeof(n));
 return(n);
}
                                              
uint32_t little_endian_u32 (uint32_t n)
{
 if (IsBigEndian())
    swap_endian((char *)&n, sizeof(n));
 return(n);
}

int64_t little_endian_64 (int64_t n) 
{
 if (IsBigEndian())
    swap_endian((char *)&n, sizeof(n));
 return(n);
}

uint64_t little_endian_u64 (uint64_t n) 
{
 if (IsBigEndian())
    swap_endian ((char *)&n, sizeof(n));
 return(n);
}

//==============================================================================
// Big Endian to host endian,  and host to Big Endian functions.
// Swaps byte order if on a Little Endian host.
//
//   pass: xintxx_t n			big endian, or host endian format
// return: xintxx_t			host endian, or big endian format
//==============================================================================
int16_t big_endian_16 (int16_t n)
{
 if (! IsBigEndian())
    swap_endian((char *)&n, sizeof(n));
 return(n);
}
          
uint16_t big_endian_u16 (uint16_t n)
{
 if (! IsBigEndian())
    swap_endian((char *)&n, sizeof(n));
 return(n);    
}

int32_t big_endian_32 (int32_t n) 
{
 if (! IsBigEndian())
    swap_endian((char *)&n, sizeof(n));
 return(n);
}
                                              
uint32_t big_endian_u32 (uint32_t n)
{
 if (! IsBigEndian())
    swap_endian((char *)&n, sizeof(n));
 return(n);
}

int64_t big_endian_64 (int64_t n) 
{
 if (! IsBigEndian())
    swap_endian((char *)&n, sizeof(n));
 return(n);
}

uint64_t big_endian_u64 (uint64_t n) 
{
 if (! IsBigEndian())
    swap_endian ((char *)&n, sizeof(n));
 return(n);
}

//==============================================================================
// set Z80 memory 8, 16 and 32 bit signed and unsigned variables in little
// endian format.  If host uses big endian format the values are swapped first.
//
//   pass: int addr
//         xintxx_t  n
// return: void
//==============================================================================
#if 0
static void z80_set_8 (int addr, int8_t n)
{
 z80mem[addr] = n;
}

static void z80_set_u8 (int addr, uint8_t n)
{
 z80mem[addr] = n;
}

static void z80_set_16 (int addr, int16_t n)
{
 n = little_endian_16(n);
 memcpy(z80mem+addr, &n, 2);
}

static void z80_set_u16 (int addr, uint16_t n)
{
 n = little_endian_u16(n);
 memcpy(z80mem+addr, &n, 2);
}

static void z80_set_32 (int addr, int32_t n)
{
 n = little_endian_32(n);
 memcpy(z80mem+addr, &n, 4);
}

static void z80_set_u32 (int addr, uint32_t n)
{
 n = little_endian_u32(n);
 memcpy(z80mem+addr, &n, 4);
}

static void z80_set_64 (int addr, int64_t n)
{
 n = little_endian_64(n);
 memcpy(z80mem+addr, &n, 8);
}

static void z80_set_u64 (int addr, uint64_t n)
{
 n = little_endian_u64(n);
 memcpy(z80mem+addr, &n, 8);
}
#endif

//==============================================================================
// get Z80 memory 8, 16 and 32 bit signed and unsigned variables in little
// endian format and convert to host endian format.
//
//   pass: int addr
// return: xintxx_t 			in host endian format
//==============================================================================
#if 0
static int8_t z80_get_8 (int addr)
{
 return z80mem[addr];
}

static uint8_t z80_get_u8 (int addr)
{
 return z80mem[addr];
}

static int16_t z80_get_16 (int addr)
{
 int16_t n;
 memcpy(&n, z80mem+addr, 2);
 return (little_endian_16(n));
}

static uint16_t z80_get_u16 (int addr)
{
 uint16_t n;
 memcpy(&n, z80mem+addr, 2);
 return (little_endian_u16(n));
}

static int32_t z80_get_32 (int addr)
{
 int32_t n;
 memcpy(&n, z80mem+addr, 4);
 return (little_endian_32(n));
}

static uint32_t z80_get_u32 (int addr)
{
 uint32_t n;
 memcpy(&n, z80mem+addr, 4);
 return (little_endian_u32(n));
}

static int64_t z80_get_64 (int addr)
{
 int64_t n;
 memcpy(&n, z80mem+addr, 8);
 return (little_endian_64(n));
}

static uint64_t z80_get_u64 (int addr)
{
 uint64_t n;
 memcpy(&n, z80mem+addr, 8);
 return (little_endian_u64(n));
}
#endif

//==============================================================================
// Dump Z80 memory
//
//   pass: int addr		Z80 memory address
//         int lines            number of lines (16 bytes per line), 0=64K
//         int header           header type, 0=none
// return: void
//==============================================================================
void function_dump (int addr, int lines, int htype)
{
 int c;
 int a;
 char x;

 if ((lines > 4096) || (lines <= 0))
    lines = 4096;			// 64K maximium

 if (htype)
    {
     printf("     ");
     for (c = 0; c < 16; c++)
        printf(" %02X", c);
     printf("  ");
     for (c = 0; c < 16; c++)
        printf("%1X", c);
     printf("\n");
    } 

 while (lines--)
    {
     addr &= 0xFFFF;
     a = addr;
     
     printf("%04X:", a);
     for (c = 0; c < 16; c++)
        {
         a &= 0xFFFF;
         printf(" %02X", (uint8_t)z80mem[a]);
         a++;
        } 
     printf(" ");        
     
     putchar('"');
     a = addr;
     for (c = 0; c < 16; c++)
        {
         a &= 0xFFFF;
         x = z80mem[a];
         a++;
         if (isalpha(x) || isdigit(x))
            putchar(x);
         else
            putchar('.');
        }
     printf("\"\n");
     
     addr += 16;
    }
}    

//==============================================================================
// Emulator commands
//
//   pass: int cmd			command number
// return: void
//==============================================================================
void function_emu_command (int cmd)
{
 switch (cmd)
    {
     case   EMU_CMD_FULLSCR : gui_toggledisplay();
                              break;
     case   EMU_CMD_TAPEREW : key_tape = 2;
                              printf("Tape rewind.\n");
                              break;
     case      EMU_CMD_DUMP : dump_addr_x = dump_addr;
                              printf("\n");     
                              function_dump(dump_addr_x, 8, 1);
                              break;
     case   EMU_CMD_DUMP_N1 : dump_addr_x += 0x0080;
                              printf("\n");
                              function_dump(dump_addr_x, 8, 1);
                              dump_addr_x &= 0xFFFF;
                              break;
     case   EMU_CMD_DUMP_N2 : dump_addr_x += 0x1000;
                              printf("\n");
                              function_dump(dump_addr_x, 8, 1);
                              dump_addr_x &= 0xFFFF;
                              break;
     case   EMU_CMD_DUMP_B1 : dump_addr_x -= 0x0080;
                              printf("\n");
                              function_dump(dump_addr_x, 8, 1);
                              dump_addr_x &= 0xFFFF;
                              break;
     case   EMU_CMD_DUMP_B2 : dump_addr_x -= 0x1000;
                              printf("\n");
                              function_dump(dump_addr_x, 8, 1);
                              dump_addr_x &= 0xFFFF;
                              break;
     case  EMU_CMD_DUMP_REP : printf("\n");
                              function_dump(dump_addr_x, 8, 1);
                              dump_addr_x &= 0xFFFF;
                              break;
     case  EMU_CMD_DUMPREGS : if (regdump.crtc)
                                 crtc_regdump();
                              if (regdump.rtc)
                                 rtc_regdump();
                              if (regdump.z80)
                                 z80_regdump();
                              break;   
     case    EMU_CMD_DBGOFF : debug.debug = 0;
                              debug.mode = 0;
                              debug.step = 0;
                              z80_blocks = 0;			// restart normal execution
                              printf("Debug mode is now off\n");
                              break;
     case     EMU_CMD_DBGON : debug.debug = 1;
                              debug.mode = 0;
                              debug.step = 0;
                              printf("Debug mode is now on\n");
                              break;
     case  EMU_CMD_DBGTRACE : if (debug.mode == 2)
                                 debug.mode = 0;
                              else
                                 {   
                                  debug.debug = 1;
                                  debug.mode = 2;
                                 } 
                              break;
     case EMU_CMD_DBGSTEP01 : debug.step = 1;
                              break;
     case EMU_CMD_DBGSTEP10 : debug.step = 10;
                              break;                               
     case EMU_CMD_DBGSTEP20 : debug.step = 20;
                              break;
    }                           

 if ((cmd >= EMU_CMD_DBGSTEP01) && (cmd <= EMU_CMD_DBGSTEP20))
    {
     if (debug.step > 1)
        printf("\n");
     debug.debug = 1;
     debug.mode = 1;
    }
}                         

//==============================================================================
// Wildcards
//
// Source obtained from GPLv2 wildcards v1.2 project.
// Copyright (C) 1996, 1997, 1998, 1999, 2000 Florian Schintke.
//
// Only structural changes have been made to keep inline with the rest of
// the emulator's style.
//
// int set (char **wildcard, char **test)
// Scans a set of characters and returns 0 if the set mismatches at this
// position in the teststring and 1 if it is matching wildcard is set to the
// closing ] and test is unmodified if mismatched and otherwise the char
// pointer is pointing to the next character
//
// int asterisk (char **wildcard, char **test)
// scans an asterisk
//
// int wildcardfit (char *wildcard, char *test)
// The function to call to compare wildcards to the test string.
//
//   pass: char *wildcard
//         char *test
// return: int			1 if wildcards fit, else 0
//==============================================================================
int asterisk (char **wildcard, char **test);
int wildcardfit (char *wildcard, char *test);

int set (char **wildcard, char **test)
{
 int fit = 0;
 int negation = 0;
 int at_beginning = 1;

 if ('!' == **wildcard)
    {
     negation = 1;
     (*wildcard)++;
    }
 while ((']' != **wildcard) || (1 == at_beginning))
    {
     if (0 == fit)
        {
         if (('-' == **wildcard) 
            && ((*(*wildcard - 1)) < (*(*wildcard + 1)))
            && (']' != *(*wildcard + 1))
	    && (0 == at_beginning))
            {
             if (((**test) >= (*(*wildcard - 1))) && ((**test) <= (*(*wildcard + 1))))
                {
                 fit = 1;
                 (*wildcard)++;
                }
            }
         else
            if ((**wildcard) == (**test))
               {
                fit = 1;
               }
        }
     (*wildcard)++;
     at_beginning = 0;
    }
 if (1 == negation)
// change from zero to one and vice versa
 fit = 1 - fit;
 if (1 == fit) 
    (*test)++;
 return (fit);
}

int asterisk (char **wildcard, char **test)
{
// Warning: uses multiple returns
 int fit = 1;

// erase the leading asterisk
 (*wildcard)++; 
 while (('\000' != (**test)) && (('?' == **wildcard) || ('*' == **wildcard)))
    {
     if ('?' == **wildcard) 
	(*test)++;
     (*wildcard)++;
    }

// Now it could be that test is empty and wildcard contains */
// aterisks. Then we delete them to get a proper state */
 while ('*' == (**wildcard))
    (*wildcard)++;

 if (('\0' == (**test)) && ('\0' != (**wildcard)))
    return (fit = 0);
 if (('\0' == (**test)) && ('\0' == (**wildcard)))
    return (fit = 1); 
 else
    {
     // Neither test nor wildcard are empty!
     // the first character of wildcard isn't in [*?]
     if (0 == wildcardfit(*wildcard, (*test)))
	{
         do 
	    {
	     (*test)++;
	     // skip as much characters as possible in the teststring
	     // stop if a character match occurs
	      while (((**wildcard) != (**test)) && ('['  != (**wildcard)) && ('\0' != (**test)))
		 (*test)++;
	    }
	 while ((('\0' != **test))? (0 == wildcardfit (*wildcard, (*test))) : (0 != (fit = 0)));
	}
     if (('\0' == **test) && ('\0' == **wildcard))
	fit = 1;
     return (fit);
    }
}

int wildcardfit (char *wildcard, char *test)
{
 int fit = 1;
  
 for (; ('\000' != *wildcard) && (1 == fit) && ('\000' != *test); wildcard++)
    {
     switch (*wildcard)
        {
         case '[': wildcard++; // leave out the opening square bracket
                   fit = set (&wildcard, &test);
                   // we don't need to decrement the wildcard as in case
                   // of asterisk because the closing ] is still there
                   break;
         case '?': test++;
                   break;
         case '*': fit = asterisk(&wildcard, &test);
                   // the asterisk was skipped by asterisk() but the loop will
                   // increment by itself. So we have to decrement
                   wildcard--;
                   break;
          default: fit = (int) (*wildcard == *test);
                   test++;
        }
    }
 while ((*wildcard == '*') && (1 == fit)) 
// here the teststring is empty otherwise you cannot
// leave the previous loop
 wildcard++;
 return (int) ((1 == fit) && ('\0' == *test) && ('\0' == *wildcard));
}
 
//==============================================================================
// Function Initialise
//
//   pass: void
// return: int			0
//==============================================================================
int function_init (void)
{
 ubee_req = 0;
 dump_addr_x = dump_addr;
 return 0;
}

//==============================================================================
// Function de-initialise
//
//   pass: void
// return: int			0
//==============================================================================
int function_deinit (void)
{
 return 0;
}

//==============================================================================
// Function reset
//
//   pass: void
// return: int			0
//==============================================================================
int function_reset (void)
{
 ubee_req = 0;
 return 0;
}

//==============================================================================
// uBee512 read - Special Ubee512 Port function
//
// This function allows a Z80 program to identify that it is running in the
// uBee512 emulator.  This is provided so that some extra functionality can
// be made taken advantage off and must be called and checked before using any
// special uBee512 write function calls to prevent unknown results on a real
// Microbee.
//
// This port call just returns the program name as a string one character at
// a time in a circular loop, the APPIDSTR in the Makefile ultimately
// determines the name to be used.
//
// Version information can be returned by a using a status function call.
//
//   pass: UINT16 port
//         struct z80PortRead *port_s
// return: UINT16		returns a null terminated string in a loop
//==============================================================================
UINT16 function_ubee_r (UINT16 port, struct z80PortRead *port_s)
{
 char c;

 c = emulator_id_str[emulator_id_pos++];
 if (! c)
    emulator_id_pos = 0;
 
 if (modio.func)
    {
     printf("function_ubee_r: port=0x%02x id char=%c\n", port, c);
     if (modio.level)
        fprintf(modio.log, "function_ubee_r: port=0x%02x id char=%c\n", port, c);
    }    

 return (UINT16)(c);
}

//==============================================================================
// status functions
//
//   pass: int addr		address of Z80 structure
// return: void
//==============================================================================
static void function_status (int addr)
{
 int i;
 int x;
 char vers[6];
 int res; 
 
 ub_func_u *func = (ub_func_u*)(z80mem + addr);
 
 switch (little_endian_u16(func->cmdres.cmd))
    {
     case 0x00 : // Get version information
                 i = 0;
                 
                 x = 5;
                 vers[0] = 0;
                 while (isdigit(emulator_ver_str[i]) && x--)
                    strncat(vers, &emulator_ver_str[i++], 1);
                 sscanf(vers, "%d", &res);
                 func->version.ver1 = little_endian_u16(res);
                 i++;

                 x = 5;
                 vers[0] = 0;
                 while (isdigit(emulator_ver_str[i]) && x--)
                    strncat(vers, &emulator_ver_str[i++], 1);
                 sscanf(vers, "%d", &res);
                 func->version.ver2 = little_endian_u16(res);
                 i++;

                 x = 5;
                 vers[0] = 0;
                 while (isdigit(emulator_ver_str[i]) && x--)
                    strncat(vers, &emulator_ver_str[i++], 1);
                 sscanf(vers, "%d", &res);
                 func->version.ver3 = little_endian_u16(res);
                 i++;

                 break;

     case 0x01 : // Get host system information
#ifdef MINGW
                 func->cmdres.res = little_endian_u16(0);	// win32
#else
                 func->cmdres.res = little_endian_u16(1);	// unices
#endif                     
                 break;
     case 0x02 : // Get host integer size
                 func->cmdres.res = little_endian_u16(sizeof(i));
                 break;
     case 0x03 : // Get host endian format
                 func->cmdres.res = little_endian_u16(IsBigEndian());
                 break;

     case 0x10 : // Get model number emulated (subject to change)
                 func->cmdres.res = little_endian_u16(model);
                 break;
     case 0x11 : // Get alpha+ model emulation status
                 func->cmdres.res = little_endian_u16(modelx.alphap);
                 break;
     case 0x12 : // Get ROM model emulation status
                 func->cmdres.res = little_endian_u16(modelx.rom);
                 break;                 
     case 0x13 : // Get model boot address status
                 func->cmdres.res = little_endian_u16(modelx.bootaddr);
                 break;
     case 0x14 : // Get RAM emulation size
                 func->cmdres.res = little_endian_u16(modelx.ram);
                 break;                 
     case 0x15 : // Get colour emulation status
                 func->cmdres.res = little_endian_u16(modelx.colour);
                 break;
     case 0x16 : // Get hardware flash emulation status
                 func->cmdres.res = little_endian_u16(modelx.hwflash);
                 break;
     case 0x17 : // Get light pen keys emulation status
                 func->cmdres.res = little_endian_u16(modelx.lpen);
                 break;
     case 0x18 : // Get speed switch emulation status
                 func->cmdres.res = little_endian_u16(modelx.speed);
                 break;
     case 0x19 : // Get speed piob7 emulation status
                 func->cmdres.res = little_endian_u16(modelx.piob7);
                 break;
     case 0x1A : // Get RTC emulation status
                 func->cmdres.res = little_endian_u16(modelx.rtc);
                 break;
       default : break;
    }
}    

//==============================================================================
// control functions
//
//   pass: int addr		address of Z80 structure
// return: void
//==============================================================================
static void function_control (int addr)
{
 ub_func_u *func = (ub_func_u*)(z80mem + addr);
 
 if (little_endian_u16(func->cmdres.id) == 0xAA55)
    {
     func->file.id = 0;  // clear it for next time (must be set each time)
     switch (little_endian_u16(func->cmdres.cmd))
        {
         case 0x00 : key_reset = 2;	// action 2 is reset, no confirmation
                     break;
         case 0x01 : done = 1;
                     break;
           default : break;
        }
    }
}    

//==============================================================================
// Diagnostics functions
//
//   pass: int addr		address of Z80 structure
// return: void
//==============================================================================
static void function_diagnostics (int addr)
{
 int addr1;
 int lines;
 int htype; 
  
 ub_func_u *func = (ub_func_u*)(z80mem + addr);
 
 if (little_endian_u16(func->cmdres.id) == 0xAA55)
    {
     func->file.id = 0;  // clear it for next time (must be set each time)    
     switch (little_endian_u16(func->cmdres.cmd))
        {
         case 0x00 : addr1 = little_endian_u16(func->dump.addr);
                     lines = little_endian_u16(func->dump.lines);
                     htype = little_endian_u16(func->dump.htype);
                     function_dump(addr1, lines, htype);
                     break;
           default : break;
        }
    }
}

//==============================================================================
// keyboard functions
//
//   pass: int addr		address of Z80 structure
// return: void
//==============================================================================
static void function_keyboard (int addr)
{
// ub_func_u *func = (ub_func_u*)(z80mem + addr);
}    

//==============================================================================
// stdio functions
//
//   pass: int addr		address of Z80 structure
// return: void
//==============================================================================
static void function_stdio (int addr)
{
 int addr1;
 ub_func_u *func = (ub_func_u*)(z80mem + addr);

 switch (little_endian_u16(func->cmdres.cmd))
    {
     case 0x00 : // getchar function
                 func->getchar.res = little_endian_16(getchar());
                 break;
     case 0x01 : // putchar function
                 func->putchar.res = little_endian_16(putchar(func->putchar.val));
                 break;
     case 0x02 : // puts function
                 addr1 = little_endian_u16(func->puts.addr);
                 func->puts.res = little_endian_16(puts((char*)(z80mem+addr1)));
                 break;
     case 0x03 : // printf function (no variables)
                 addr1 = little_endian_u16(func->puts.addr);
                 printf("%s", (char*)(z80mem+addr1));
                 break;

     case 0x80 : // write 16 bit value as signed decimal
                 printf("%d", little_endian_16(func->putchar.val));
                 break;
     case 0x81 : // write 16 bit value as unsigned decimal
                 printf("%ud", little_endian_u16(func->putchar.val));
                 break;
     case 0x82 : // write 16 bit value as 2 hex characters
                 printf("%02x", little_endian_u16(func->putchar.val));
                 break;
     case 0x83 : // write 16 bit value as 4 hex characters
                 printf("%04x", little_endian_u16(func->putchar.val));
                 break;

       default : break;
    }
}    

//==============================================================================
// File functions.
//
//   pass: int addr		address of Z80 structure
// return: void
//==============================================================================
static void function_files (int addr)
{
 int addr1;
 int addr2;
 int addr3; 

 char *ptrres;

 ub_func_u *func = (ub_func_u*)(z80mem + addr);
   
 if (little_endian_u16(func->file.id) == 0xAA55)
    {
     func->file.id = 0;  // clear it for next time (must be set each time)
     switch (little_endian_u16(func->file.cmd))
        {
         case 0x00 : // fopen
                     addr1 = little_endian_u16(func->file.addr1);
                     addr2 = little_endian_u16(func->file.addr2);
                     func->file.fp.p = fopen((char*)(z80mem+addr1),
                     (char*)(z80mem+addr2));
                     func->file.res = little_endian_u16(func->file.fp.p == NULL);
                     break;
         case 0x01 : // fclose
                     func->file.res = little_endian_16(fclose(func->file.fp.p));
                     break;
         case 0x02 : // fflush
                     func->file.res = little_endian_16(fflush(func->file.fp.p));
                     break;
         case 0x03 : // feof
                     func->file.res = little_endian_16(feof(func->file.fp.p));
                     break;
         case 0x04 : // ferror
                     func->file.res = little_endian_16(ferror(func->file.fp.p));
                     break;
         case 0x05 : // fgetc
                     func->file.res = little_endian_16(fgetc(func->file.fp.p));
                     break;
         case 0x06 : // fgetpos (RESERVED)
                     break;
         case 0x07 : // fgets
                     addr3 = little_endian_u16(func->file.addr3);
                     ptrres = fgets((char*)(z80mem+addr3), func->file.num, func->file.fp.p);
                     func->file.res = little_endian_u16(ptrres != NULL);
                     break;
         case 0x08 : // fputc
                     func->file.res = little_endian_16(fputc(func->file.val1, func->file.fp.p));
                     break;
         case 0x09 : // fputs
                     addr3 = little_endian_u16(func->file.addr3);
                     func->file.res = little_endian_16(fputs((char*)(z80mem+addr3), func->file.fp.p));
                     break;                     
         case 0x0A : // fread
                     addr3 = little_endian_u16(func->file.addr3);
                     func->file.res = little_endian_16(fread((char*)(z80mem+addr3),
                     func->file.size, func->file.num, func->file.fp.p));         
                     break;
         case 0x0B : // freopen (RESERVED)
                     break;
         case 0x0C : // fseek (RESERVED)
                     break;
         case 0x0D : // fsetpos (RESERVED)
                     break;
         case 0x0E : // ftell (RESERVED)
                     break;
         case 0x0F : // fwrite
                     addr3 = little_endian_u16(func->file.addr3);
                     func->file.res = little_endian_16(fwrite((char*)(z80mem+addr3),
                     func->file.size, func->file.num, func->file.fp.p));
                     break;
         case 0x10 : // clearerr
                     clearerr(func->file.fp.p);
                     break;
         case 0x11 : // rewind
                     rewind(func->file.fp.p);
                     break;
         case 0x12 : // remove
                     addr1 = little_endian_u16(func->file.addr1);
                     func->file.res = little_endian_16(remove((char*)(z80mem+addr1)));
                     break;
         case 0x13 : // rename
                     addr1 = little_endian_u16(func->file.addr1);
                     addr2 = little_endian_u16(func->file.addr2);
                     func->file.res = little_endian_16(rename((char*)(z80mem+addr1),
                     (char*)(z80mem+addr2)));
                     break;
         case 0x14 : // tmpfile
                     func->file.fp.p = tmpfile();
                     break;
         case 0x15 : // tmpnam (RESERVED)
                     break;
         case 0x16 : // ugetc (RESERVED)
                     break;
         case 0x17 : // eofval constant
                     func->file.res = little_endian_16(EOF);
                     break;
           default : break;
        }
    }
}       

//==============================================================================
// Directory functions.
//
//   pass: int addr		address of Z80 structure
// return: void
//==============================================================================
static void function_directory (int addr)
{
 char const illegal_cpm_char[] = "<>,;:=?*[]\%|()/\\";
 struct dirent *de;
 struct stat st;

 char filepath[256]; 
 char filename[256];

 int l;
 int i;
 int addr1;
 int addr2;
 int addr3;
 int addr4;

 int res;
 int dot;
 int match;
 int dir_type;
 
 ub_func_u *func = (ub_func_u*)(z80mem + addr);
   
 if (little_endian_u16(func->file.id) == 0xAA55)
    {
     func->file.id = 0;  // clear it for next time (must be set each time)
     switch (little_endian_u16(func->file.cmd))
        {
         case 0x00 : // open a directory for reading
                     addr1 = little_endian_u16(func->file.addr1); // pass directory path/name
                     addr2 = little_endian_u16(func->file.addr2); // return file name or wildcards
                     addr3 = little_endian_u16(func->file.addr3); // return modified file path here
                     z80mem[addr2] = 0;
                     func->file.val1 = little_endian_u16(0);
                     res = 1;

                     strcpy(filepath, (char*)(z80mem+addr1));
                     l = strlen(filepath);

                     if (! l)                                  // add '.' if an empty path
                        strcat(filepath, ".");
                     else   
                        if (strcmp(filepath, "."SLASHCHAR_STR) == 0)  // if './' or '.\'
                           filepath[l-1] = 0;                  // remove the slash character
#ifdef MINGW
                        else
                           if (filepath[l-1] == ':')           // if last char is ':'
                              strcat(filepath, SLASHCHAR_STR); // add a '\'
#endif                              

                     if (stat(filepath, &st) == 0)
                        {
                         res = ((st.st_mode & S_IFDIR) == 0);
                         if (! res)
                            {
                             z80mem[addr2] = '*';
                             z80mem[addr2+1] = 0;
                            }
                        } 
                     else
                        res = 1;   

                     if (res)   
                        {
                         l = strlen(filepath);
#ifdef MINGW                         
                         while ((l) && (filepath[l-1] != SLASHCHAR) && (filepath[l-1] != ':'))
#else
                         while ((l) && (filepath[l-1] != SLASHCHAR))
#endif
                            l--;

                         strcpy((char*)z80mem+addr2, filepath+l); // copy the name part
                         if ((l > 1) && (filepath[l-1] == SLASHCHAR))  // if '/' or '\'
                            l--;
                         filepath[l] = 0;                         // remove the name part

                         if (! strlen((char*)z80mem+addr2))       // if file name is blank
                            {
                             z80mem[addr2] = '*';
                             z80mem[addr2+1] = 0;
                            }                         

                         l = strlen(filepath);
                         if (! l)                                 // add '.' if empty path
                            strcat(filepath, ".");

                         if (stat(filepath, &st) == 0)            // try again without name
                            res = ((st.st_mode & S_IFDIR) == 0);
                         else
                            res = 1;       
                        }

                     if (! res)
                        {
                         func->file.fp.d = opendir(filepath);

                         l = strlen(filepath);
                         if ((l) && (filepath[l-1] != SLASHCHAR))  // add '/' or '\' if no end slash
                            strcat(filepath, SLASHCHAR_STR);

                         res = (func->file.fp.d == NULL);
                        }
                     
                     func->file.val1 = little_endian_16(strlen((char*)z80mem+addr2));
                     strcpy((char*)z80mem+addr3, filepath);  // return the file path
                     func->file.val2 = little_endian_16(strlen(filepath));
                     func->file.res = little_endian_16(res);
                     
                     break;    
         case 0x01 : // close directory
                     func->file.res = little_endian_16(closedir(func->file.fp.d));
                     break;
         case 0x02 : // return the next directory entry (with wildcards)
                     addr1 = little_endian_u16(func->file.addr1); // return name match
                     addr2 = little_endian_u16(func->file.addr2); // pass file name or wildcards
                     addr3 = little_endian_u16(func->file.addr3); // pass file path
                     addr4 = little_endian_u16(func->file.addr4); // return full path and file name match
                     strcpy(filepath, (char*)z80mem+addr3);
                     match = 0;
                     do
                        {
                         de = readdir(func->file.fp.d);
                         if (de)
                            {
                             match = wildcardfit((char*)(z80mem+addr2), de->d_name);
                             if (match)
                                {
                                 strcpy((char *)z80mem+addr1, de->d_name);
                                 snprintf(filename, sizeof(filename), "%s%s", filepath, de->d_name);
                                 if (stat(filename, &st) == 0)
                                    {
                                     dir_type = 0;
#ifdef MINGW
#else
                                     dir_type |= (S_ISLNK(st.st_mode) != 0);
#endif                                     
                                     dir_type |= ((S_ISREG(st.st_mode) != 0) << 1);
                                     dir_type |= ((S_ISDIR(st.st_mode) != 0) << 2);
                                     dir_type |= ((S_ISCHR(st.st_mode) != 0) << 3);
                                     dir_type |= ((S_ISBLK(st.st_mode) != 0) << 4);
                                     dir_type |= ((S_ISFIFO(st.st_mode) != 0) << 5);
#ifdef MINGW
#else
                                     dir_type |= ((S_ISSOCK(st.st_mode) != 0)  << 6);
#endif                                        
                                    }
                                 else
                                    dir_type = 0;
                                 strcpy((char *)z80mem+addr4, filename);
                                 func->file.val2 = little_endian_16(strlen(filename));                                 
                                 func->file.res = little_endian_16(dir_type);
                                 func->file.val1 = little_endian_16(strlen((char*)z80mem+addr1));
                                }
                            }    
                        }
                     while ((de) && (! match));
                     
                     if (! match)
                        {
                         z80mem[addr1] = 0;
                         z80mem[addr4] = 0;                         
                         func->file.res = little_endian_u16(0);
                         func->file.val1 = little_endian_u16(0);
                         func->file.val2 = little_endian_u16(0);                         
                        }
                     break;
         case 0x03 : // wildcards test
                     addr1 = little_endian_u16(func->file.addr1); // pass file name to be tested
                     addr2 = little_endian_u16(func->file.addr2); // pass wildcards
                     func->file.res = little_endian_16(wildcardfit((char*)(z80mem+addr2), (char*)(z80mem+addr1)));
                     break;
         case 0x04 : // check if filename is a legal CP/M 8.3 format
                     addr1 = little_endian_u16(func->file.addr1); // pass file name to be tested
                     res = 0;
                     dot = 0;
                     i = 0;
                     l = strlen((char*)z80mem+addr1);
                     res = (l > 12);
                     if (! res)
                        {
                         while ((i < l) && (! res))
                            {
                             if (z80mem[addr1+i] == '.')
                                {
                                 dot++;
                                 res = (dot > 1) * 2;
                                 if ((! res) && (l > 3))
                                    res = (i < (l-4)) * 3;
                                } 
                             else
                                res = (strchr(illegal_cpm_char, z80mem[addr1+i]) != NULL) * 4;
                             i++;
                            }
                        }
                     if ((! res) && (! dot) && (l > 8))
                        res = 5;
                     func->file.res = little_endian_16(res);                        
           default : break;
        }
    }
}       

//==============================================================================
// uBee512 write - Special uBee512 Port function
//
// All commands pass a Z80 address to a structure in memory.  The structure
// contents is dependent on the command but all share the same overhead.  Data
// passed and returned in the structure is always in little endian format.
//
// IMPORTANT: This function does not currently address banking issues so it
// is important to realise that data may get lost if ANY bank swapping occurs
// after this function call.  To avoid problems do not switch banks or call
// any OS calls before the returned data is used.  Essentially the main area
// of concern is the 32K DRAM bank at 0x8000 that may also be at 0x0000.
// Normally writes to these banks will be copied from one to the other, but
// this does not occur here as we are acting on Z80 memory outside of the
// MZ80 context.  Also do not have the structure in video memory for similar
// reasons,  unlikely you would want to do that any way.
//
// It is important to verify that the application is running under the emulator
// and not a real Microbee before calling this function.  Use the uBee512 read
// port function to check first.
//
// The calling sequence is:
// port[n] = command number (0..255)
// port[n] = LSB of structure address (0 if not using a structure) 
// port[n] = MSB of structure address (0 if not using a structure)
//
//
//   pass: UINT16 port
//         UINT8 data
//         struct z80PortWrite *port_s
// return: void
//==============================================================================
void function_ubee_w (UINT16 port, UINT8 data, struct z80PortWrite *port_s)
{
#if 0
 printf("function_ubee__w port 0x%02x = 0x%02x\n", port, data);
 return; 
#endif

 if (modio.func)
    {
     printf("function_ubee_w: port=0x%02x data=0x%02x\n", port, data);
     if (modio.level)
        fprintf(modio.log, "function_ubee_w: port=0x%02x data=0x%02x\n", port, data);
    }    

 if (ubee_req)		// if we need to get the structure address
    {
     ubee_req--;
     if (ubee_req == 1)
        ubee_addr = data;	// LSB of the address
     else   
        if (ubee_req == 0)
           {
            ubee_addr = little_endian_u16(ubee_addr | (data << 8));
            switch (ubee_command)
               {
                case 0x00 : // status functions
                            function_status(ubee_addr);
                            break;
                case 0x10 : // control functions
                            function_control(ubee_addr);
                            break;
                case 0x20 : // diagnostics functions
                            function_diagnostics(ubee_addr);
                            break;
                case 0x30 : // keyboard functions
                            function_keyboard(ubee_addr);
                            break;                            
                case 0x40 : // stdio functions
                            function_stdio(ubee_addr);
                            break;                            
                case 0x50 : // file functions
                            function_files(ubee_addr);
                            break;
                case 0x60 : // directory functions
                            function_directory(ubee_addr);
                            break;
                  default : break;          
               }
           } 
    }
 else
    {
     ubee_command = data;
     ubee_req = 2;
    }
}
