//******************************************************************************
//*                                  uBee512                                   *
//*         An emulator for the Microbee Z80 FDD and ROM based models          *
//*                                                                            *
//*                              Keyboard module                               *
//*                                                                            *
//*                   Copyright (C) 2007-2008 Stewart Kay                      *
//******************************************************************************
//
// Overhead keyboard code for the CRTC6545 and 256TC keyboard encoders.
// The 256TC model uses both encoder modules.
//
//==============================================================================
// ChangeLog (most recent entries are at top)
//==============================================================================
// v2.3.0 - 9 January 2008, S.J.Kay
// - Added key commands.  Activated by pressing the HOME+KEY or ALT+KEY
//   key combinations.
// - Added conditional test of (modelx.lpen) for the 256TC model.
// - Added modio_t structure.
//
// v2.0.0 - 14 October 2007, S.J.Kay
// - Code specific for standard keyboard (CRT6545) has been moved to
//   keystd.c, this module is now simply used as an overhead to call
//   functions in keystd.c and keytc.c modules.
// - Moved revision history to stdkey.c as it directly relates to that module.
//
// v0.0.0 - 11 June 2007, S.J.Kay
// Start with "nanowasp" source distribution version 0.22. An emulator for the
// microbee 128k. Copyright (C) 2000-2003  David G. Churchill.

//==============================================================================
/*
 *  uBee512 - An emulator for the Microbee Z80 FDD and ROM based models. 
 *  Copyright (C) 2007-2008 Stewart Kay
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/ 
//==============================================================================

#include <stdio.h>
#include <stdlib.h>
#include <SDL.h>

#include "keyb.h"
#include "keystd.h"
#include "keytc.h"
#include "function.h"
#include "ubee512.h"

//==============================================================================
// structures and variables
//==============================================================================
static int func_key_down;

extern int model;
extern model_t modelx;

extern SDL_Event event;

extern modio_t modio;

//==============================================================================
// Keyboard initialise
//
//   pass: void
// return: int 				0
//==============================================================================
int keyb_init (void)
{
 if (model == MOD_256TC)
    {
     if (modelx.lpen)
        return keytc_init() | keystd_init();
     else
        return keytc_init();
    }
 else
    return keystd_init();
}

//==============================================================================
// Keyboard de-initialise
//
//   pass: void
// return: int 				0
//==============================================================================
int keyb_deinit (void)
{
 if (model == MOD_256TC)
    {
     if (modelx.lpen)
        return keytc_deinit() | keystd_deinit();
     else
        return keytc_deinit();
    }    
 else
    return keystd_deinit();
}

//==============================================================================
// Keyboard reset
//
//   pass: void
// return: int 				0
//==============================================================================
int keyb_reset (void)
{
 if (model == MOD_256TC)
    {
     if (modelx.lpen)
        return keytc_reset() | keystd_reset();
     else
        return keytc_reset();
    }    
 else
    return keystd_reset();
}

//==============================================================================
// key down event handler.
//
//   pass: void
// return: void
//==============================================================================
void keyb_keydown_event (void)
{
 SDLKey key; 

 key = event.key.keysym.sym; 

 if ((key == SDLK_HOME) || (modelx.lpen && ((key == SDLK_LALT) || (key == SDLK_RALT))))
    func_key_down = 1;

 if (func_key_down)
    {
     if (key == SDLK_RETURN)
        function_emu_command(EMU_CMD_FULLSCR);
     if (key == SDLK_t)
        function_emu_command(EMU_CMD_TAPEREW);
     if (key == SDLK_d)
        function_emu_command(EMU_CMD_DUMP);
     if (key == SDLK_1)
        function_emu_command(EMU_CMD_DUMP_N1);
     if (key == SDLK_2)
        function_emu_command(EMU_CMD_DUMP_N2);
     if (key == SDLK_3)
        function_emu_command(EMU_CMD_DUMP_B1);
     if (key == SDLK_4)
        function_emu_command(EMU_CMD_DUMP_B2);
     if (key == SDLK_5)
        function_emu_command(EMU_CMD_DUMP_REP);
     if (key == SDLK_r)
        function_emu_command(EMU_CMD_DUMPREGS);        
     if (key == SDLK_EQUALS)
        function_emu_command(EMU_CMD_DBGON);
     if (key == SDLK_MINUS)
        function_emu_command(EMU_CMD_DBGOFF);        
     if (key == SDLK_BACKSLASH)
        function_emu_command(EMU_CMD_DBGTRACE);        
     if (key == SDLK_BACKSPACE)
        function_emu_command(EMU_CMD_DBGSTEP01);
     if (key == SDLK_LEFTBRACKET)
        function_emu_command(EMU_CMD_DBGSTEP10);
     if (key == SDLK_RIGHTBRACKET)
        function_emu_command(EMU_CMD_DBGSTEP20);
     return;   
    }    

 if (model == MOD_256TC)
    {
     keytc_keydown_event();
     if (modelx.lpen)
        keystd_keydown_event();
    } 
 else
    keystd_keydown_event();
}         

//==============================================================================
// key up event handler.
//
//   pass: void
// return: void
//==============================================================================
void keyb_keyup_event (void)
{
 SDLKey key;

 key = event.key.keysym.sym; 

 if ((key == SDLK_HOME) || (modelx.lpen && ((key == SDLK_LALT) || (key == SDLK_RALT))))
    {
     func_key_down = 0;
     return;
    }

 if (model == MOD_256TC)
   {
    keytc_keyup_event();
    if (modelx.lpen)
       keystd_keyup_event();
   } 
 else
    keystd_keyup_event();
}

//==============================================================================
// Force a character to be returned.
//
// This is useful for placing the boot code into monitor mode by forcing the 'M'
// key (0x0d) to be returned or for pasting.
//
//   pass: int scan			scan code 0-63
//   pass: int counts			number of times to be forced
// return: void
//==============================================================================
void keyb_force (int scan, int counts)
{
 if (model == MOD_256TC)
    keytc_force(scan, counts);
 else
    keystd_force(scan, counts);
}    

//==============================================================================
// Force none.
//
// Forces no scan matches for counts
//
//   pass: int counts			number of times to be forced
// return: void
//==============================================================================
void keyb_force_none (int counts)
{
 if (model == MOD_256TC)
    keytc_force_none(counts);
 else
    keystd_force_none(counts);
}    
