//******************************************************************************
//*                                  uBee512                                   *
//*         An emulator for the Microbee Z80 FDD and ROM based models          *
//*                                                                            *
//*                            Memory mapper module                            *
//*                                                                            *
//*                   Copyright (C) 2007-2008 Stewart Kay                      *
//******************************************************************************
//
// Emulates the 128-512K memory bank switching and video active signal.
// Emulates the 32-56K RAM and 64K DRAM models.
//
//==============================================================================
// ChangeLog (most recent entries are at top)
//==============================================================================
// v2.3.0 - 4 January 2008, S.J.Kay
// - Recoded this module in many areas to fix known banking issues.
// - Added RAM/DRAM initilization patterns to emulate real hardware, modio
//   option allows sequential bank numbers to be used instead.
// - Fixed 256TC mapping,  bit 5 of port 50h is used to select the extra 128K
//   of DRAM and there is no ROM2 and ROM3 used.
// - All DRAM models (except 256TC) now have ROM1 size changed so that 16K
//   ROMS can be loaded instead of just 8K.  Relaxed the rules in the loading
//   section of the code on what size is allowed.
// - Added modio_t structure.
//
// v2.1.0 - 27 October 2007, S.J.Kay
// - boot_data had incorrect ordering of p256k, 256k, and p128k, 128k models,
//   moved the 256k models above the 128k models.
// - Added optional DRAM model ROMs for each model type.
// - Implement the modelx information structure.
//
// v2.0.0 - 17 October 2007, S.J.Kay
// - All FDD models can now load seperate boot ROMs and a fall-back ROM will be
//   loaded for some models if the default ROM is not found.
// - Added the Teleterm 256TC, PJB P256K, and 256K models.
// - APC model name changed to 56K for ROM loading.
//
// v1.4.0 - 26 September 2007, S.J.Kay
// - Changes for banked and non-banked models.  Models of 64K or less are
//   set up for VDU to fixed at address 0xF000.
// - Added alternive boot ROMs to be loaded depending on the model being
//   emulated,  for ROM machines this means a BASIC ROM.
// - Added memmap_basicrom() function for Premium ROM model basic upper/lower
//   half selection.
// - loadrom() function modified and is now global function memmap_loadrom()
// - Removed the non aplpha+ source build option.
// - Added changes to error reporting.
//
// v1.0.0 - 7 July 2007, S.J.Kay
// - added vdu_switchvideo_in() and vdu_switchvideo_out() function calls.
// - ROMs 1-3 required seperate handlers as there will be DRAM at 0xC000-0xDFFF
//   when ROM3 is selected.  This change allows the DRAM area to be written to.
//   A smaller dupwrite handler is also activated if the top and bottom banks
//   are the same and ROM3 is selected. (there is an 8K DRAM window).
// - According to the Technical manual ROM3 appears to be placed into wrong
//   memory location (0xC000) and should be placed into 0xE000.  Also ROM3 is
//   documented as being 8K and ROM1 and ROM2 are 16K.  Have made changes in
//   various locations for this and have modified the load_rom function and
//   optioned out the roundup function for now.
//
// - Converted to 512K DRAM model Microbee (as used for PJB 512K BIOS).  Now
//   uses array of pointers for banks.
//
// v0.0.0 - 5 June 2007, S.J.Kay
// Start with "nanowasp" source distribution version 0.22. An emulator for the
// microbee 128k. Copyright (C) 2000-2003  David G. Churchill.

//==============================================================================
/*
 *  uBee512 - An emulator for the Microbee Z80 FDD and ROM based models. 
 *  Copyright (C) 2007-2008 Stewart Kay
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/ 
//==============================================================================

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "memmap.h"
#include "ubee512.h"
#include "z80.h"
#include "vdu.h"
#include "mz80.h"

#include "macros.h"

static void memmap_init4164 (void);
static void memmap_init4256 (void);

static void switchin (UINT8 m);
static void switchout (void);

static void romxwrite (UINT32 addr, UINT8 data, struct MemoryWriteByte *mem_s);
static void dupwrite_lo (UINT32 addr, UINT8 data, struct MemoryWriteByte *mem_s);
static void dupwrite_hi (UINT32 addr, UINT8 data, struct MemoryWriteByte *mem_s);

static struct MemoryWriteByte z80_mem_w[MAXMEMHANDLERS] =
   { { -1, -1, NULL, NULL } };

static struct MemoryReadByte z80_mem_r[MAXMEMHANDLERS] =
   { { -1, -1, NULL, NULL } };

static char rom1[ROM1_SIZE],
            rom2[ROM2_SIZE],
            rom3[ROM3_SIZE];

static char block00[BLOCK_SIZE],
            block01[BLOCK_SIZE],
            block02[BLOCK_SIZE],
            block03[BLOCK_SIZE],
            block04[BLOCK_SIZE],
            block05[BLOCK_SIZE],
            block06[BLOCK_SIZE],
            block07[BLOCK_SIZE],
            block08[BLOCK_SIZE],
            block09[BLOCK_SIZE],
            block10[BLOCK_SIZE],
            block11[BLOCK_SIZE],
            block12[BLOCK_SIZE],
            block13[BLOCK_SIZE],
            block14[BLOCK_SIZE],
            block15[BLOCK_SIZE];

char const *block_ptrs[] =
   {
    block00, block01, block02, block03,
    block02, block03, block00, block01,
    block04, block05, block06, block07,
    block06, block07, block04, block05,
    block08, block09, block10, block11,
    block10, block11, block08, block09,
    block12, block13, block14, block15,
    block14, block15, block12, block13
   };

int mem_mode;		// current memory map configuration (port 0x50)

static struct mz80context z80;

static char basic[0x4000];	// storage area for Premium basic in ROM

static int port50h;
static int x_upper;
static int blocksel_x;
static int mem_mode_x;

static char maphigh[9];

extern int model;
extern model_t modelx;

extern char userhome_romspath[];

extern modio_t modio;

//==============================================================================
// Define all the possible ROM boot up options for disk and ROM based models.
// 
// The boot_data table must match the enumeration table order containing
// MOD_NAMES (in ubee512.h)
//==============================================================================
 static boot_t boot_data [MOD_TOTAL] =
    {
// MOD_256TC
     {
      {"256TC.ROM", "", "", ""},
      rom1,
      0,
      0x4000,
      "",
      ""
     },

// MOD_P512K
     {
      {"P512K.ROM", "rom1.bin", "", ""},
      rom1,
      0,
      0x4000,
      "P512K_2.ROM",
      "P512K_3.ROM"
     },

// MOD_512K
     {
      {"512K.ROM", "rom1.bin", "", ""},
      rom1,
      0,
      0x4000,
      "512K_2.ROM",
      "512K_3.ROM"
     },

// MOD_P256K
     {
      {"P256K.ROM", "rom1.bin", "", ""},
      rom1,
      0,
      0x4000,
      "P256K_2.ROM",
      "P256K_3.ROM"
     },

// MOD_256K
     {
      {"256K.ROM", "rom1.bin", "", ""},
      rom1,
      0,
      0x4000,
      "256K_2.ROM",
      "256K_3.ROM"
     },

// MOD_P128K
     {
      {"P128K.ROM", "rom1.bin", "", ""},
      rom1,
      0,
      0x4000,
      "P128K_2.ROM",
      "P128K_3.ROM"
     },

// MOD_128K    
     {
      {"128K.ROM", "rom1.bin", "", ""},
      rom1,
      0,
      0x4000,
      "128K_2.ROM",
      "128K_3.ROM"
     },

// MOD_P64K
     {
      {"P64K.ROM", "", "", ""},
      rom1,
      0,
      0x4000,
      "P64K_2.ROM",
      "P64K_3.ROM"
     },

// MOD_64K
     {
      {"64K.ROM", "", "", ""},
      rom1,
      0,
      0x4000,
      "64K_2.ROM",
      "64K_3.ROM"
     },

// MOD_56K
     {
      {"56K.ROM", "", "", ""},
      z80mem,
      0xE000,
      0x1000,
      "",
      ""
     },

// MOD_PPC85
     {
      {"PREM-A.ROM", "", "", ""},
      basic,
      0,
      0x4000,
      "",
      ""
     },

// MOD_PC85
     {
      {"PC85_BASIC.ROM", "", "", ""},
      z80mem,
      0x8000,
      0x4000,
      "",
      ""
     },

// MOD_PC
     {
      {"PC_BASIC.ROM", "PC_BASIC_A.ROM", "PC_BASIC_B.ROM", ""},
      z80mem,
      0x8000,
      0x4000,
      "",
      ""
     },

// MOD_IC
     {
      {"IC_BASIC.ROM", "IC_BASIC_A.ROM", "IC_BASIC_B.ROM", ""},
      z80mem,
      0x8000,
      0x4000,
      "",
      ""
     },

// MOD_2MHZ
     {
      {"2MHZ_BASIC.ROM", "2MHZ_BASIC_A.ROM", "2MHZ_BASIC_B.ROM", "2MHZ_BASIC_C.ROM", "2MHZ_BASIC_D.ROM", ""},
      z80mem,
      0x8000,
      0x4000,
      "",
      ""
     }
    };

//==============================================================================
// Load a boot or BASIC ROM image from file.
//
// Each model uses a different file name so that the user can link to the
// actual ROM file concerned.
//
//   pass: void
// return: int			0 if no errors, -1 if error
//==============================================================================
static int memmap_loadboot (void)
{
 int i;
 int loaded;
 int offset;
 int size;

 char romimage1[512];
 char romimage2[512];
 char romimage3[512]; 

 i = 0;
 loaded = 0;
 offset = boot_data[model].offset;
 size = boot_data[model].size;

 while (size > 0)
    {
     if (strcmp(boot_data[model].romimage[i], "") == 0)
        {
         if (! loaded)
            {
             console_output();
             printf("memmap_loadboot: Unable to continue ROM load operation (no ROM file specified).\n");
             return -1;
            }
         else
            size = 0;   
        }
     else
        {   
         snprintf(romimage1, sizeof(romimage1), "%s%s", userhome_romspath, boot_data[model].romimage[i]);
         loaded += memmap_loadrom(romimage1, boot_data[model].dest + offset + loaded, size);
         size -= loaded;
         i++;
#if 0
         printf("memmap_loadboot: file=%s size=%d loaded=%d\n", romimage1, size, loaded);
#endif
        } 
    }    

 if (! loaded)
    {
     console_output();
     printf("memmap_loadboot: Unable to load %s (required to boot).\n", romimage1);
     return -1;
    }
 else
    if (! modelx.rom)
       printf("memmap_loadboot: boot ROM image loaded: %s\n", romimage1);     

// special treatment for Premium PC85 model
 if (model == MOD_PPC85)
    {
     memcpy(z80mem + 0x8000, basic, 0x2000);	// copy the 1st half of BASIC
     snprintf(romimage1, sizeof(romimage1), "%s%s", userhome_romspath, "PREM-B.ROM");

     if (memmap_loadrom(romimage1, z80mem + 0xA000, 0x2000) != 0x2000)
        {
         console_output();
         printf("memmap_loadboot: Unable to load PREM-B.ROM (required to boot).\n");
         return -1;
        }
    }

// Load the optional ROMs for the DRAM models except 256TC
 if ((modelx.ram >= 64) && (model != MOD_256TC))
    {
     snprintf(romimage2, sizeof(romimage2), "%s%s", userhome_romspath, boot_data[model].romimage2);
     snprintf(romimage3, sizeof(romimage3), "%s%s", userhome_romspath, boot_data[model].romimage3);     

     if (memmap_loadrom(romimage2, rom2, ROM2_SIZE))
        printf("memmap_loadboot: optional ROM 2 image loaded: %s\n", romimage2);
     else
        memset(rom2, 0xFF, sizeof(rom2));   
     if (memmap_loadrom(romimage3, rom3, ROM3_SIZE))
        printf("memmap_loadboot: optional ROM 3 image loaded: %s\n", romimage3);
     else
        memset(rom3, 0xFF, sizeof(rom3));   
    }

 return 0;
}

//==============================================================================
// Initialize the memory map.
//
// Bank memory handling is used for models with more than 64K of RAM/DRAM.
// Models with less than 64K do not have bank switching and the VDU is fixed
// at 0xF000.
//
// DRAM and RAM is initialized to typical patterns for with each type of
// memory used for the emulated model.
//
//   pass: void
// return: int			0 if no errors, -1 if error
//==============================================================================
int memmap_init (void)
{
 int rindex = 0;
 int windex = 0;

 int i;

 mz80GetContext(&z80);
 z80.z80MemRead = z80_mem_r;
 z80.z80MemWrite = z80_mem_w;
 mz80SetContext(&z80);

 if (modelx.ram >= 64)
    {
     if (modio.raminit)
// for debugging purposes, fills each 32K DRAM block with numbers.
        {
         for (i = 0; i < 16; i++)
            memset(block_ptrs[i], i, BLOCK_SIZE);
        }
     else
        {
         if (modelx.ram <= 128)
            memmap_init4164();
         else
            memmap_init4256();
        }
        
     if (memmap_loadboot() == -1)
        return -1;
     switchin(0);
     port50h = 0;
    }
 else
// initialize for RAM models (< 64K)
    {
     if (modio.raminit)
// for debugging purposes, fills each 32K RAM section with numbers.
        {
         memset(z80mem+0x0000, 0, 0x8000);
         memset(z80mem+0x8000, 1, 0x8000);
        }
     else
// initialize static RAM
        {
         switch (model)
            {
             case   MOD_56K : memmap_init6116(z80mem, 28);
                              break;
             case MOD_PPC85 :
             case  MOD_PC85 :
             case    MOD_PC : memmap_init6264(z80mem, 16);
                              break;
                    default : memmap_init6116(z80mem, 16);
                              break;
            }
        }   

// VDU write handler
     z80_mem_w[windex].memoryCall = vdu_vidmem_w;
     z80_mem_w[windex].pUserArea  = NULL;
     z80_mem_w[windex].lowAddr = 0xF000;
     z80_mem_w[windex].highAddr = 0xFFFF;
     windex++;

// ROM write handler
     if (modelx.rom)
        {
         z80_mem_w[windex].lowAddr = 0x8000;
         z80_mem_w[windex].highAddr = 0xEFFF;
        } 
     
     if (model == MOD_56K)
        {
         z80_mem_w[windex].lowAddr = 0xE000;
         z80_mem_w[windex].highAddr = 0xEFFF;
        } 

     z80_mem_w[windex].memoryCall = romxwrite;
     z80_mem_w[windex].pUserArea = NULL;
     windex++;

// Terminate write structure
     z80_mem_w[windex].lowAddr = -1;
     z80_mem_w[windex].highAddr = -1;
     z80_mem_w[windex].memoryCall = NULL;
     z80_mem_w[windex].pUserArea = NULL;

// VDU read handler
     z80_mem_r[rindex].memoryCall = vdu_vidmem_r;
     z80_mem_r[rindex].pUserArea = NULL;
     z80_mem_r[rindex].lowAddr = 0xF000;
     z80_mem_r[rindex].highAddr = 0xFFFF;
     rindex++;

// Terminate read structure
     z80_mem_r[rindex].lowAddr = -1;
     z80_mem_r[rindex].highAddr = -1;
     z80_mem_r[rindex].memoryCall = NULL;
     z80_mem_r[rindex].pUserArea = NULL;

// Load the BASIC or 56K boot ROMs
     if (memmap_loadboot() == -1)
        return -1;

     vdu_switchvideo_in();
    }

 return 0;
}

//==============================================================================
// De-init the memory map (N/A)
//
//   pass: void
// return: int			0
//==============================================================================
int memmap_deinit (void)
{
 return 0;
}

//==============================================================================
// Set memory map to the reset condition
//
//   pass: void
// return: int			0
//==============================================================================
int memmap_reset (void)
{
 if (modelx.ram >= 64)
    {
     switchout();
     switchin(0);
     port50h = 0;
    }
 else
    if (modelx.rom)
       {
        x_upper = 1;
        memmap_basicrom(0);
        vdu_switchvideo_in();
       } 
 return 0;
}

//==============================================================================
// Load a ROM image from file.
//
// The messages reported in this function are not considered as fatal so are
// not forced to the console.
//
//   pass: char *name  		pointer to file name
//         char *dest  		pointer to destination
//         int size 		expected size of the ROM to be loaded
// return: int			number of bytes loaded
//==============================================================================
int memmap_loadrom (char *name, char *dest, int size)
{
 FILE *romfp;
 long len;

 if (dest == NULL)
    return 0;

 memset(dest, 0, size);

 romfp = fopen(name, "rb");
 if (romfp == NULL)
    return 0;

 fseek(romfp, 0, SEEK_END);  		// get size of ROM
 len = ftell(romfp);
 fseek(romfp, 0, SEEK_SET);		// back to start

 if (len != 0)
    {
     if (len > size)
        {
         printf("memmap_loadrom: ROM size is too large (continuing): %s\n", name);
         return 0;
        }
     else
        if (len < size)
           size = len;
     len = fread(dest, 1, len, romfp);
     fclose(romfp);
    } 

 if (len == 0)
    printf("memmap_loadrom: Failed to load ROM (continuing): %s\n", name);

 return len;
}

//==============================================================================
// Initialize RAM to look like typical 4164 DRAM Chips
//
// Pattern: 128 x 00s - 128 x FFs - repeat
//
//   pass: void
// return: void
//==============================================================================
static void memmap_init4164 (void)
{
 int i, x, z, a;
 char *dram;

 for (i = 0; i < 16; i++)
    {
     dram = block_ptrs[i];
     a = 0;
     for (x = 0; x < 128; x++)
        {
         for (z = 0; z < 128; z++)
            *(dram + a++) = 0x00;
         for (z = 0; z < 128; z++)
            *(dram + a++) = 0xFF;
        }
    }
}

//==============================================================================
// Initialize RAM to look like typical 4256 DRAM Chips
//
// Pattern: 128 x 00:FF - 128 x FF:00 - repeat
//
//   pass: void
// return: void
//==============================================================================
static void memmap_init4256 (void)
{
 int i, x, z, a;
 char *dram;

 for (i = 0; i < 16; i++)
    {
     dram = block_ptrs[i];
     a = 0;
     for (x = 0; x < 64; x++)
        {
         for (z = 0; z < 128; z++)
            {
             *(dram + a++) = 0x00;
             *(dram + a++) = 0xFF;
            }
         for (z = 0; z < 128; z++)
            {
             *(dram + a++) = 0xFF;
             *(dram + a++) = 0x00;
            }
        }
    }
}    

//==============================================================================
// Initialize RAM (in 2K chunks) to look like typical 6264 static RAM chips
//
// Pattern: 64 x FF:FF:00:00 - 64 x 00:00:FF:FF - repeat
//
//   pass: char *mem  		pointer to the memory to be initialized
//   pass: int banks            number of 2K banks of consecutive memory
// return: void
//==============================================================================
void memmap_init6264 (char *mem, int banks)
{
 int x, z, a;

 a = 0;
 while (banks--)
    {
     for (x = 0; x < 4; x++)
        {
         for (z = 0; z < 64; z++)
            {
             *(mem + a++) = 0xFF;
             *(mem + a++) = 0xFF;
             *(mem + a++) = 0x00;
             *(mem + a++) = 0x00;                 
            }
         for (z = 0; z < 64; z++)
            {
             *(mem + a++) = 0x00;
             *(mem + a++) = 0x00;                 
             *(mem + a++) = 0xFF;
             *(mem + a++) = 0xFF;
            }
        }
    }    
}     

//==============================================================================
// Initialize RAM (in 2K chunks) to look like typical 6116 static RAM chips.
//
// For now this just uses the 6264 pattern.
//
//   pass: char *mem  		pointer to the memory to be initialized
//   pass: int banks            number of 2K banks of consecutive memory
// return: void
//==============================================================================
void memmap_init6116 (char *mem, int banks)
{
 int x, z, a;

 a = 0;
 while (banks--)
    {
     for (x = 0; x < 4; x++)
        {
         for (z = 0; z < 64; z++)
            {
             *(mem + a++) = 0xFF;
             *(mem + a++) = 0xFF;
             *(mem + a++) = 0x00;
             *(mem + a++) = 0x00;                 
            }
         for (z = 0; z < 64; z++)
            {
             *(mem + a++) = 0x00;
             *(mem + a++) = 0x00;                 
             *(mem + a++) = 0xFF;
             *(mem + a++) = 0xFF;
            }
        }
    }    
}     

//==============================================================================
// Set memory map mode - Port 0x50 write handler
//
//   pass: UINT16 port
//	   UINT8 data
//         struct z80PortWrite *port_struct
// return: void
//==============================================================================
void memmap_mode (UINT16 port, UINT8 data, struct z80PortWrite *port_struct)
{
 if (modio.mem)
    {
     printf("memmap_mode: port=0x%02x, data=0x%02x\n", port, data);
     if (modio.level)
        fprintf(modio.log, "memmap_mode: port=0x%02x, data=0x%02x\n", port, data);
    }    

 if (modelx.ram < 64)
    return;

 if (data != port50h)
    {
     switchout();
     switchin(data);
     port50h = data;
    } 
}

//==============================================================================
// Switch in the lower or upper half of the Premium ROM basic if emulating a
// Premium PC85 ROM machine.
//
//   pass: int upper			1 for upper, else lower
// return: void
//==============================================================================
void memmap_basicrom (int upper)
{
 if ((model == MOD_PPC85) && (upper != x_upper))
    {
     if (upper)
        memcpy(z80mem + BASIC1_ADDR, basic + 0x2000, 0x2000);
     else   
        memcpy(z80mem + BASIC1_ADDR, basic, 0x2000);
     x_upper = upper;
    }
}        

//==============================================================================
// Switch in DRAM blocks, Video memory locations, ROMs 1,2 and 3, etc
//
// Requires the following:
// - Create a blocksel value depending on the model emulated. At most there is
//   16 x 32K memory blocks.  Copy this block of memory to 0000-7FFF.
// - Copy the high DRAM block to 8000-FFFF, will be overlayed with ROM and VDU
//   RAM as needed.
// - If VDU is enabled then set the VDU memory handler for 8000-8FFF or F000-
//   FFFF.
// - If ROMs are enabled then set the ROM write memory handler for ROMs
//   1,2 and 3 (to do nothing) and copy the ROMs to upper memory.
//
// Notes:
// - VDU RAM takes precedence over ROM (tech86 8-10) and DRAM if port 50
//   bit 3=0 (VDU enabled)
// - Memory handlers are used in the order they are inserted,  the higher
//   precedence handler should be inserted before a lower precendence
//   handler.
// - ROM 1 is located at 8000-BFFF.
// - ROM 2 is located at C000-FFFF.
// - ROM 3 is located at E000-FFFF. (C000-DFFF is filled with FFs)
// - BANK_ROMS = 00000100 (enable ROMs bit=0)
// - BANK_VRAM = 00001000 (enable VDU RAM bit=0)
// - BANK_VADD = 00010000 (VDU RAM at F000 bit=0, 8000 bit=1)
// - BANK_ROM3 = 00100000 (enable ROM 2 bit=0, else ROM 3,  2nd 128K on 256TC)
//
//   pass: int m
// return: void
//==============================================================================
static void switchin (UINT8 m)
{
 int rindex = 0;
 int windex = 0;
 int blocksel;

// set mem_mode and blocksel values depending on the model emulated
 switch (modelx.ram)
    {
     case 512 : mem_mode = m & B8(11111111);
                blocksel = ((mem_mode & B8(11000000)) >> 3) | (mem_mode & B8(00000111));
                break;
     case 256 : if (model == MOD_256TC)
                   {
                    mem_mode = m & B8(00111111);
                    blocksel = ((mem_mode & B8(00100000)) >> 2) | (mem_mode & B8(00000111));
                   }
                else
                   {   
                    mem_mode = m & B8(01111111);
                    blocksel = ((mem_mode & B8(01000000)) >> 3) | (mem_mode & B8(00000111));
                   } 
                break;
     case 128 : mem_mode = m & B8(00111111);
                blocksel = mem_mode & B8(00000111);
                break;
     case  64 : mem_mode = m & B8(00111101);
                blocksel = mem_mode & B8(00000101);
                break;
      default : blocksel = 0;
    }

// save these values for when we switch out
 blocksel_x = blocksel;
 mem_mode_x = mem_mode;

// copy selected DRAM block to memory at 0x0000-0x7FFF
 memcpy(z80mem, block_ptrs[blocksel], BLOCK_SIZE);

// copy the high memory block to 8000-FFFF then further on we overlay it with
// ROMs and VDU depending on the mem_mode.
 memcpy(z80mem+0x8000, BLOCK_HIGH, BLOCK_SIZE);
 strcpy(maphigh, "DDDDDDDD");

// insert video memory handler if video RAM is enabled.
 if (! (mem_mode & BANK_VRAM))	// if video RAM enabled
    {
     z80_mem_r[rindex].memoryCall = vdu_vidmem_r;
     z80_mem_r[rindex].pUserArea  = NULL;

     z80_mem_w[windex].memoryCall = vdu_vidmem_w;
     z80_mem_w[windex].pUserArea  = NULL;

     if (mem_mode & BANK_VADD)   // if video RAM at 8000-8FFF
        {
         z80_mem_r[rindex].lowAddr =  0x8000;
         z80_mem_r[rindex].highAddr = 0x8FFF;        

         z80_mem_w[windex].lowAddr =  0x8000;
         z80_mem_w[windex].highAddr = 0x8FFF;
         maphigh[0] = 'V';
        }
     else
        {
         z80_mem_r[rindex].lowAddr =  0xF000;
         z80_mem_r[rindex].highAddr = 0xFFFF;

         z80_mem_w[windex].lowAddr =  0xF000;
         z80_mem_w[windex].highAddr = 0xFFFF;
         maphigh[7] = 'V';
        }
     rindex++;
     windex++;
    }

// insert ROM memory handlers if any ROMs are enabled.
 if (! (mem_mode & BANK_ROMS))     		// if ROMs enabled
    {
// set writes to any ROMs to be ignored.
     z80_mem_w[windex].lowAddr = 0x8000;
     z80_mem_w[windex].highAddr = 0xFFFF;
     z80_mem_w[windex].memoryCall = romxwrite;
     z80_mem_w[windex].pUserArea = NULL;
     windex++;    

// copy the 16K ROM to 8000-BFFF
     memcpy(z80mem+0x8000, rom1, ROM1_SIZE);
     if (maphigh[0] != 'V')
        maphigh[0] = 'r';
     maphigh[1] = 'r';
     maphigh[2] = 'r';
     maphigh[3] = 'r';
     maphigh[4] = 'r';
     maphigh[5] = 'r';
     maphigh[6] = 'r';
     if (maphigh[7] != 'V')
        maphigh[7] = 'r';

// if a 256TC then the ROM will also appear at C000-FFFF
     if (model == MOD_256TC)
        memcpy(z80mem+0xC000, rom1, ROM1_SIZE);
     else
// the 256TC model does not have ROMs 2 and 3
        {
         if (mem_mode & BANK_ROM3)   // if ROM3 selected
            {
             memset(z80mem+0xC000, 0xFF, 0x2000);
             memcpy(z80mem+0xE000, rom3, ROM3_SIZE);
            } 
         else
            memcpy(z80mem+0xC000, rom2, ROM2_SIZE);
        }
    }

// if the same DRAM block is mapped into upper and lower memory we need to
// echo any writes to one half to the other.  If any memory handlers have
// already been inserted for the same address locations then those shall
// take precedence.
 if (block_ptrs[blocksel] == BLOCK_HIGH) 
    {
     z80_mem_w[windex].lowAddr = 0x0000;
     z80_mem_w[windex].highAddr = 0x7FFF;
     z80_mem_w[windex].memoryCall = dupwrite_lo;
     z80_mem_w[windex].pUserArea = NULL;
     windex++;
     z80_mem_w[windex].lowAddr = 0x8000;
     z80_mem_w[windex].highAddr = 0xFFFF;
     z80_mem_w[windex].memoryCall = dupwrite_hi;
     z80_mem_w[windex].pUserArea = NULL;
     windex++;
    }

// terminate the memory handlers
 z80_mem_r[rindex].lowAddr = -1;
 z80_mem_r[rindex].highAddr = -1;
 z80_mem_r[rindex].memoryCall = NULL;
 z80_mem_r[rindex].pUserArea = NULL;

 z80_mem_w[windex].lowAddr = -1;
 z80_mem_w[windex].highAddr = -1;
 z80_mem_w[windex].memoryCall = NULL;
 z80_mem_w[windex].pUserArea = NULL;

// copy video memory into the main memory if video is enabled, done last so it
// overwrites anything already there.
 if (! (mem_mode & BANK_VRAM))	// if video RAM enabled
    vdu_switchvideo_in();
}

//==============================================================================
// Switch out DRAM blocks
//
// This function's only requirement is to copy back lower DRAM blocks and also
// the upper DRAM block.
//
// The memory boxes used below helps to make the switching out process easier
// to understand.  a [........] box containing periods are replaced by one of
// the following legends to represent what is in memory at each 1000 hex memory
// location starting at 8000 hex:
//
// V=VDU, D=DRAM, r=ROM
//
//   pass: void
// return: void
//==============================================================================
static void switchout (void)
{
 char *block;

// switch out current VDU RAM back to banked VDU
 if (! (mem_mode & BANK_VRAM))	// if video RAM enabled
    vdu_switchvideo_out();

// copy back the lower block at 0000-7FFF to the storage area
 block = (char *)block_ptrs[blocksel_x];
 memcpy(block, z80mem, BLOCK_SIZE);

// if the lower block was in then we have already switched out the higher block
 if (block_ptrs[blocksel_x] == BLOCK_HIGH)
   return;

 if (mem_mode_x & BANK_ROMS)	              // if ROMs disabled
    {
     if ((mem_mode_x & BANK_VRAM) == 0)
        {
         if (mem_mode_x & BANK_VADD)          // if video at 8000-8FFF
            memcpy(BLOCK_HIGH+0x1000, z80mem+0x9000, 0x7000); // [VDDDDDDD]
         else
            memcpy(BLOCK_HIGH+0x0000, z80mem+0x8000, 0x7000); // [DDDDDDDV]
        }
     else
        memcpy(BLOCK_HIGH, z80mem+0x8000, BLOCK_SIZE);        // [DDDDDDDD]
    }
#if 0
 else
    if (model == MOD_256TC)
       {
        if ((mem_mode_x & BANK_VRAM) == 0)
           {
            if (mem_mode_x & BANK_VADD)      // if video at 8000-8FFF
               memcpy(BLOCK_HIGH+0x4000, z80mem+0xC000, 0x4000); // [VrrrDDDD]
            else
               memcpy(BLOCK_HIGH+0x4000, z80mem+0xC000, 0x3000); // [rrrrDDDV]
           }
        else
           memcpy(BLOCK_HIGH+0x4000, z80mem+0xC000, 0x4000);     // [rrrrDDDD]
       }
#endif
}

//==============================================================================
// ROM x write (0x8000-0xEFFF)
//
// This handler is installed for any addresses to ROMs.  This prevents writing
// to ROM image in the z80mem memory.
//
//   pass: UINT32 addr
//	   UINT8 data
//         struct MemoryWriteByte *mem_s
// return: void
//==============================================================================
static void romxwrite (UINT32 addr, UINT8 data, struct MemoryWriteByte *mem_s)
{
}

//==============================================================================
// duplicates writes from low half of DRAM to the higher half.  This handler is
// installed when DRAM bank 0, block 1 is in the lower half of memory as it is
// also in the upper half of memory.  Mapping characters ensure that duplicate
// writes will not occur if VDU RAM or ROM are selected.
//
//   pass: UINT32 addr
//         UINT8 data
//         struct MemoryWriteByte *mem_s
// return: void
//==============================================================================
static void dupwrite_lo (UINT32 addr, UINT8 data, struct MemoryWriteByte *mem_s)
{
 z80mem[addr] = data;
 if (maphigh[(addr & 0x7000) >> 12] == 'D')
    z80mem[addr | 0x8000] = data;
}

//==============================================================================
// duplicates writes from high half of DRAM to the lower half.  This handler is
// installed when DRAM bank 0, block 1 is in the lower half of memory as it is
// also in the upper half of memory.
//
//   pass: UINT32 addr
//         UINT8 data
//         struct MemoryWriteByte *mem_s
// return: void
//==============================================================================
static void dupwrite_hi (UINT32 addr, UINT8 data, struct MemoryWriteByte *mem_s)
{
 z80mem[addr & 0x7FFF] = data;
 z80mem[addr] = data;
}
