//******************************************************************************
//*                                  uBee512                                   *
//*         An emulator for the Microbee Z80 FDD and ROM based models          *
//*                                                                            *
//*                              Z80 PIO module                                *
//*                                                                            *
//*                   Copyright (C) 2007-2008 Stewart Kay                      *
//******************************************************************************
//
// Emulate the Z80 PIO functionality and interrupts.
//
// All PIO peripheral devices are maintained in seperate source files:
//
// printer.c		Parallel printer port
// tape.c		Tape in/out port
// serial.c		Serial in/out port
// sound.c		Speaker sound
// rtc.c		RTC (interrupt)
// keytc.c		256TC keyboard (interrupt)
//
//==============================================================================
// ChangeLog (most recent entries are at top)
//==============================================================================
// v2.3.0 - 24 December 2007, S.J.Kay
// - Added modio_t structure.
//
// v2.2.0 - 15 November 2007, S.J.Kay
// - The printer_w() function is now always called even if no --print options
//   are specified.
//
// v2.1.0 - 23 October 2007, S.J.Kay
// - Added PIO port B bit 7 code to handle alternative signal sources in the
//   pio_polling() and pio_r() functions.
// - Added a fast poll persist counter,  this makes the vsync clock keep time
//   correctly by not missing interrupts.
// - Implement the modelx information structure.
//
// v2.0.0 - 15 October 2007, S.J.Kay
// - Added RTC polling and interrupt code, and PIO RTC IRQ bit 7.
// - Added 256TC polling and interrupt code, and PIO port B bit 1 (key ready)
// - PIO port B in/out now uses the direction bits.
// - PIO port B interrupt mask word now used.
//
// v1.4.0 - 20 September 2007, S.J.Kay
// - Added sound on/off conditional testing.
//
// v1.3.0 - 1 September 2007, S.J.Kay
// - Emulate PIO internal registers and interrupts.
// - Added RS232 serial TX, RX, DTR and CTS.
// - Added printer (parallel) out on port A.
//
// v1.2.0 - 22 August 2007, S.J.Kay
// - Added tape out and tape in.
//
// v1.0.0 - 23 June 2007, S.J.Kay
// - Created a new file to implement the Z80 PIO emulation.

//==============================================================================
/*
 *  uBee512 - An emulator for the Microbee Z80 FDD and ROM based models. 
 *  Copyright (C) 2007-2008 Stewart Kay
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/ 
//==============================================================================

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <inttypes.h>
#include <string.h>
#include <time.h>
#include <SDL.h>

#include "pio.h"
#include "sound.h"
#include "tape.h"
#include "serial.h"
#include "printer.h"
#include "keytc.h"
#include "rtc.h"
#include "crtc.h"
#include "ubee512.h"
#include "mz80.h"

#include "macros.h"

pio_t pio_a;
pio_t pio_b;

static int execratio_porta;
static int execratio_serial;

static int piob7_force;

static int fast_poll_persist;

extern int model;
extern model_t modelx;

extern int z80_ratio;
extern char key_tape;
extern int serial_byte_rx;

extern int sound;

extern char *tape_i;
extern char *tape_o;

extern char *print_a;
extern char *print_b;

extern char *coms1_dev;
extern int coms1;

extern modio_t modio;

//==============================================================================
// PIO - Control register
//
//   pass: pio_t *pio
//         UINT8 data
//         char port
// return: void
//==============================================================================
static void pio_control (pio_t *pio, UINT8 data, char port)
{
 pio->cont = data;
                 
 if (pio->action)
    {
     if (pio->action == 1)
        {     
         pio->direction = data; // set data direction (1=input, 0=output)
         if (modio.piocont)
            {
             printf("pio_control: port %c: set data direction: 0x%02x\n", port, data);
             if (modio.level)
                fprintf(modio.log, "pio_control: port %c: set data direction: 0x%02x\n", port, data);
            }
        }
     if (pio->action == 2)
        {
         pio->maskword = data;
         if (modio.piocont)
            {
             printf("pio_control: port %c: mask word: 0x%02x\n", port, data);
             if (modio.level)
                fprintf(modio.log, "pio_control: port %c: mask word: 0x%02x\n", port, data);
            }
        }
     pio->action = 0;   
    }
 else
    {
     if ((data & B8(00000001)) == 0) // if interrupt vector
       {
        pio->vector = data;
        if (modio.piocont)
           {
            printf("pio_control: port %c: interrupt vector: 0x%02x\n", port, data);
            if (modio.level)
               fprintf(modio.log, "pio_control: port %c: interrupt vector: 0x%02x\n", port, data);
           }    
       } 
     else   
        if ((data & B8(00001111)) == B8(00000111)) // if interrupt control word
           {
            pio->ienable = data & B8(10000000);
            if (pio->ienable)
               pio->ienableff = B8(10000000);               
            pio->andor = data & B8(01000000);
            pio->hilo = data & B8(00100000);
            if (data & B8(00010000)) // if mask word follows
               {
                pio->action = 2;
                pio->pending = 0;
               }
            if (modio.piocont)
               {
                printf("pio_control: port %c: interrupt control word: 0x%02x\n", port, data);
                if (modio.level)
                   fprintf(modio.log, "pio_control: port %c: interrupt control word: 0x%02x\n", port, data);
               }    
           }    
        else
           if ((data & B8(00001111)) == B8(00001111)) // if mode select
              {
               pio->mode = (data & B8(11000000)) >> 6;
               if (pio->mode == 3) // if mode 3 (bit control)
                  pio->action = 1;
               if (modio.piocont)
                  {
                   printf("pio_control: port %c: mode select: 0x%02x\n", port, data);
                   if (modio.level)
                      fprintf(modio.log, "pio_control: port %c: mode select: 0x%02x\n", port, data);
                  }    
              }
           else   
              if ((data & B8(00001111)) == B8(00000011)) // if interrupt flipflop set
                 {
                  pio->ienableff = data & B8(10000000);
                  if (modio.piocont)
                     {
                      printf("pio_control: port %c: interrupt flipflop set: 0x%02x\n", port, data);
                      if (modio.level)
                         fprintf(modio.log, "pio_control: port %c: interrupt flipflop set: 0x%02x\n", port, data);
                     }
                 }
    }    
}

//==============================================================================
// PIO Initialise
//
//   pass: void
// return: int				0 if success, -1 if error
//==============================================================================
int pio_init (void)
{
 int x=0;
#ifdef use_audio
 if (sound)
    x = sound_init();
#endif
 return (x | tape_init() | serial_init() | printer_init());
}

//==============================================================================
// PIO de-initialise
//
//   pass: void
// return: int				0 if success, -1 if error
//==============================================================================
int pio_deinit (void)
{
 int x=0;
#ifdef use_audio
 if (sound)
    x = sound_deinit();
#endif
 return (x | tape_deinit() | serial_deinit() | printer_deinit()); 
}

//==============================================================================
// PIO reset
//
//   pass: void
// return: int				0 if success, -1 if error
//==============================================================================
int pio_reset (void)
{
 int x=0;

 pio_a.mode = 1;
 pio_a.maskword = 0; 
 pio_a.ienableff = 0;
 pio_a.data = 0; 
 pio_a.action = 0; 

 pio_b.mode = 1;
 pio_b.maskword = 0;
 pio_b.ienableff = 0;
 pio_b.data = 0; 
 pio_b.action = 0; 

 piob7_force = 0;
 
#ifdef use_audio
 if (sound)
    x = sound_reset();
#endif
 return (x | tape_reset() | serial_reset() | printer_reset());
}

//==============================================================================
// PIO - poll for interrupt events.
//
// Checks to see if a maskable interrupt should be generated,  if so, set the
// interrupt to occur for the next mz80exec() block.
//
// A special z80_exec_ratio() function is used to allow faster responses to
// multiple interrupts.
//
// A fast poll persist counter is used to persist with the exec_ratio current
// value.  This can be set by any interrupt that needs to stay in fast polling
// mode for a certain time period after the interrupt is set.
//
// Tape input interrupt on port B bit has not been implemted at this stage.
//
// Interrupts that call the z80_exec_ratio() function are followed by a
// "return" to prevent resetting the execution ratio to normal.  This occurs
// if the code reaches the end of the polling function.
//
// Interrupts supported:
//
// RS232 Serial interrupts
// -----------------------
// If a byte has been received and ready in a buffer and the last RX PIO bit
// was a stop bit (serial_byte_rx == -1 in our case) an interrupt will be
// generated.
//
// See the pio_configure() function for information about the exec_ratio
// value used here.
//
// 256TC Key interrupts
// --------------------
// Used when emulating the Teleterm 256TC keyboard.  The 256TC boot ROM
// tested used only polling methods.
//
// Clock interrupts
// ----------------
// Used when emulating a source interrupt on bit port B bit 7.  the signal
// source can be one of the following:
//
//       RTC : Used if the RTC is being emulated and the source has been
//             selected to be taken from the RTC.
//     VSYNC : Used when emulating the CRTC VSYNC into port B bit 7.
//       NET : Not implemented.
//             D15 parallel port connector (X2 pin 14) used as a network
//             direction bit.
//       PUP : Pullup resistor,  no interrupt is generated for this.
//
// Port A LPT
// ----------
// Generates an interrupt after each write to port A.
//
//   pass: void
// return: void
//==============================================================================
void pio_polling (void)
{
 static int piob7_vsync;
 int piob7_irq; 

 int key_irq;
 
// poll the 256TC keyboard for interrupts
 if (model == MOD_256TC)
    key_irq = keytc_poll();

// poll the PIO port B bit 7 check for interrupts
 switch (modelx.piob7)
    {
     case MODPB7_PUP: piob7_irq = 0;
                      break;
     case  MODPB7_VS: piob7_irq = crtc_vblank();
                      if (piob7_irq != piob7_vsync)
                         {
                          piob7_vsync = piob7_irq;
                          if (piob7_irq == 1)
                             piob7_force = piob7_irq;
                         } 
                      else
                         piob7_irq = 0;    
                      break; 
     case MODPB7_RTC: piob7_irq = rtc_poll();
                      break;
     case MODPB7_NET: piob7_irq = 0;
                      break;
    }

// if (pio_a.ienable)
 if (pio_a.ienableff)
    {
     if (printer_strobe_get() == 0)
        {
         printer_strobe_set();
         z80_exec_ratio(execratio_porta);
         mz80int(pio_a.vector);
         if (modio.piocont)
            {
             printf("pio_interrupt (A): vector=0x%02x\n", pio_a.vector);
             if (modio.level)
                fprintf(modio.log, "pio_interrupt (A): vector=0x%02x\n", pio_a.vector);
            } 
         return;
        }
    }     

 if (pio_b.ienableff)
    {
     if (serial_byte_rx == -1)
        {
         if ((serial_readpoll() != -1) && ((pio_b.maskword & PIO_B_RS232_RX) == 0))
            {
             z80_exec_ratio(execratio_serial);
             mz80int(pio_b.vector);
             if (modio.piocont)
                {
                 printf("pio_interrupt (B): vector=0x%02x\n", pio_b.vector);
                 if (modio.level)
                    fprintf(modio.log, "pio_interrupt (B): vector=0x%02x\n", pio_b.vector);
                }
             return;
            }
        }
     else
        serial_r();	// see explanation in this function why we call

     if ((model == MOD_256TC) && (key_irq) && ((pio_b.maskword & PIO_B_KEY256TC) == 0))
        mz80int(pio_b.vector);
     else
        if ((piob7_irq) && ((pio_b.maskword & PIO_B_CLOCK) == 0))
           {
            switch (modelx.piob7)
               {
                case MODPB7_PUP: break;
                case  MODPB7_VS: mz80int(pio_b.vector);
                                 if (model != MOD_2MHZ)
                                    {
                                     z80_exec_ratio(5);
                                     fast_poll_persist = 200;
                                    } 
                                 else
                                    {
                                     z80_exec_ratio(5);   // ???
                                     fast_poll_persist = 200;
                                    } 
                                 if (modio.piocont)
                                    {
                                     printf("pio_interrupt (B) VSYNC : vector=0x%02x\n", pio_b.vector);
                                     if (modio.level)
                                        fprintf(modio.log, "pio_interrupt (B) VSYNC : vector=0x%02x\n", pio_b.vector);
                                    }
                                 return;                                      
                case MODPB7_RTC: mz80int(pio_b.vector);
                                 if (modio.piocont)
                                    {
                                     printf("pio_interrupt (B) RTC: vector=0x%02x\n", pio_b.vector);
                                     if (modio.level)
                                        fprintf(modio.log, "pio_interrupt (B) RTC: vector=0x%02x\n", pio_b.vector);
                                    }
                                 break;
                case MODPB7_NET: break;
               }
           }      
    } 

// set the Z80 execution ratio back to normal,  i.e 30 interrupts/sec @
// 3.375 MHz if no interrupt events, and not in fast_poll_persist mode.
 if (fast_poll_persist)
    fast_poll_persist--;
 else
    if (z80_ratio != 1)
       z80_exec_ratio(1);
} 

//==============================================================================
// PIO - configure.
//
// RS232 Serial interrupts
// -----------------------
// A z80_exec_ratio(40) ratio was found to be about right at fast baud rates
// values of 60 did not work any faster, 70 did not work in some circumstances.
//
// To get a true through put rate of 19200 baud (1920 bytes/sec) would require
// 1920 interrupts per second.  Some more experimentation is needed with
// the ratio value but for now a value of 40 is being used as maximum.
//
// At slower baud rates having too high a ratio will slow response time down
// so the ratio needs to be determined by the RX baud rate:
//
// These calculations are based on default values for the frame rate and XTAL:
//
// the time for 1 Z80 frame is: tz = (3375000 / 30) / 3375000
// the time for 1 serial bit is: tb = (1 / baudrx)
//
// ratio = tz / tb
//
//   pass: int cpuclock
//         int frame_rate
//         int serial_rx_baud
// return: void
//==============================================================================
void pio_configure (int cpuclock, int frame_rate, int serial_rx_baud)
{
 float tz;
 float tb;

// set port A to about 600 interrupts/sec @ 3.375 MHz
 execratio_porta = 20;

// calculate the execution ratio for RS232 serial
 tz = ((float)cpuclock / frame_rate) / (float)cpuclock;
 tb = ((float)1.0 / serial_rx_baud);

 execratio_serial = (int)(tz/tb);

// set the maximum allowed ratio for serial
 if (execratio_serial > 40)
    execratio_serial = 40;
}

//==============================================================================
// PIO read - Port function
//
//   pass: UINT16 port
//         struct z80PortRead *port_s
// return: UINT16		register data
//==============================================================================
UINT16 pio_r (UINT16 port, struct z80PortRead *port_s)
{
 switch (port & 0x03)
    {
     case 0x00 : // PIO port A data port
                 if (modio.pioa)
                    {
                     printf("pio_r (A): port=0x%02x pio_a.data=%x\n", port, pio_a.data);
                     if (modio.level)
                         fprintf(modio.log, "pio_r (A): port=0x%02x pio_a.data=%x\n", port, pio_a.data);
                    }
                 return pio_a.data;
     case 0x01 : // PIO port A control port
                 if (modio.piocont)
                    {
                     printf("pio_r (A): port=0x%02x control=0x%02x\n", port, pio_a.cont);
                     if (modio.level)
                        fprintf(modio.log, "pio_r (A): port=0x%02x control=0x%02x\n", port, pio_a.cont);
                    }
                 return pio_a.cont;
     case 0x02 : // PIO port B data port
                 if ((key_tape) && (tape_i != NULL) && (pio_b.direction & PIO_B_CASIN))
                    pio_b.data = (pio_b.data & B8(11111110)) | tape_r();
                 if ((model == MOD_256TC) && (pio_b.direction & PIO_B_KEY256TC))
                    pio_b.data = (pio_b.data & B8(11111101)) | keytc_poll();
                 if (pio_b.direction & PIO_B_RS232_RX)
                    {
                     if (coms1 != -1)
                        pio_b.data = (pio_b.data & B8(11100111)) | serial_r(); // + CTS
                     else
                        pio_b.data = (pio_b.data & B8(11100111)) | B8(00001000); // stop bit, CTS
                    }
                 switch (modelx.piob7)
                    {
                     case MODPB7_PUP: pio_b.data |= B8(10000000);  // pull up resistor
                                      break;                    
                     case  MODPB7_VS: if (pio_b.direction & PIO_B_CLOCK)
                                         {
                                          pio_b.data = (pio_b.data & B8(01111111)) | crtc_vblank() | piob7_force;
                                          piob7_force = 0;
//                                        printf("pio_r: vsync pio_b.data=0x%02X\n", pio_b.data);
                                         } 
                                      break;
                     case MODPB7_RTC: if ((modelx.rtc) && (pio_b.direction & PIO_B_CLOCK))
                                         pio_b.data = (pio_b.data & B8(01111111)) | rtc_poll();
                                      break;
                     case MODPB7_NET: break;
                    }
                                           
                 if (modio.piob)
                    {
                     printf("pio_r (B): port=0x%02x pio_b.data=%x\n", port, pio_b.data);
                     if (modio.level)
                         fprintf(modio.log, "pio_r (B): port=0x%02x pio_b.data=%x\n", port, pio_b.data);
                    }    

                 return pio_b.data;
     case 0x03 : // PIO port B control port
                 if (modio.piocont)
                    {
                     printf("pio_r (B): port=0x%02x control=0x%02x\n", port, pio_b.cont);
                     if (modio.level)
                        fprintf(modio.log, "pio_r (B): port=0x%02x control=0x%02x\n", port, pio_b.cont);
                    }
                 return pio_b.cont;
    } 
 return 0xFFFF;
}

//==============================================================================
// PIO write - Port function
//
//   pass: UINT16 port
//         UINT8 data
//         struct z80PortWrite *port_s
// return: void
//==============================================================================
void pio_w (UINT16 port, UINT8 data, struct z80PortWrite *port_s)
{
 switch (port & 0x03)
    {
     case 0x00 : // PIO port A data port
                 if (modio.pioa)
                    {
                     printf("pio_w (A): port=0x%02x data=0x%02x\n", port, data);
                     if (modio.level)
                        fprintf(modio.log, "pio_w: port=0x%02x data=0x%02x\n", port, data);
                    }
                 pio_a.data = data;
                 printer_w(data);
                 break;
     case 0x01 : // PIO port A control port
                 if (modio.piocont)
                    {
                     printf("pio_w (A): port=0x%02x data=0x%02x\n", port, data);
                     if (modio.level)
                        fprintf(modio.log, "pio_w: port=0x%02x data=0x%02x\n", port, data);
                    }
                 pio_control(&pio_a, data, 'A');
                 break;
     case 0x02 : // PIO port B data port
                 if (modio.piob)
                    {
                     printf("pio_w (B): port=0x%02x data=0x%02x\n", port, data);
                     if (modio.level)
                        fprintf(modio.log, "pio_w: port=0x%02x data=0x%02x\n", port, data);
                    }
                 pio_b.data = data & (0xff ^ pio_b.direction);
#ifdef use_audio
                 if ((sound) && ((pio_b.direction & PIO_B_SPEAKER) == 0))
                    sound_w(data);
#endif
                 if ((tape_o != NULL) && ((pio_b.direction & PIO_B_CASOUT) == 0))
                    tape_w(data);
                 if ((coms1 != -1) && ((pio_b.direction & PIO_B_RS232_TX) == 0))
                    serial_w(data);   
                 break;
     case 0x03 : // PIO port B control port
                 if (modio.piocont)
                    {
                     printf("pio_w (B): port=0x%02x data=0x%02x\n", port, data);
                     if (modio.level)
                        fprintf(modio.log, "pio_w: port=0x%02x data=0x%02x\n", port, data);
                    }
                 pio_control(&pio_b, data, 'B');
                 break;
    }             
}
