//******************************************************************************
//*                                  uBee512                                   *
//*         An emulator for the Microbee Z80 FDD and ROM based models          *
//*                                                                            *
//*                              Printer module                                *
//*                                                                            *
//*                   Copyright (C) 2007-2008 Stewart Kay                      *
//******************************************************************************
//
// This module is used to emulate the parallel printer port,  it sends data to
// an ASCII decimal and/or direct without any changes to files.
//
//==============================================================================
// ChangeLog (most recent entries are at top)
//==============================================================================
// v2.2.0 - 15 November 2007, S.J.Kay
// - Make changes so that printer_strobe is still cleared even if no --print
//   options are specified.
//
// v1.3.0 - 29 August 2007, S.J.Kay
// - Created a new file and implement the printer emulation.

//==============================================================================
/*
 *  uBee512 - An emulator for the Microbee Z80 FDD and ROM based models. 
 *  Copyright (C) 2007-2008 Stewart Kay
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/ 
//==============================================================================

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>

#include "printer.h"
#include "ubee512.h"
#include "mz80.h"

static int count;

static int printer_strobe;

static FILE *print_a_file;
static FILE *print_b_file;

extern char *print_a;
extern char *print_b;

extern char printa[];
extern char printb[];

//==============================================================================
// Printer reset.
//
//   pass: void
// return: int				0
//==============================================================================
int printer_reset (void)
{
 printer_strobe = 0;
 return 0;
}

//==============================================================================
// Printer Initialise. 
//
// Open the text and binary printer files for writing
//
//   pass: void
// return: int				0 if success, -1 if error
//==============================================================================
int printer_init (void)
{
 if (print_a != NULL)
    {
     print_a_file = fopen(printa, "w");
     if (print_a_file == NULL)
        return -1;
     count = 0;
    }
        
 if (print_b != NULL)
    {
     print_b_file = fopen(printb, "wb");
     if (print_b_file == NULL)
        return -1;
    }    

 return 0;
}

//==============================================================================
// Printer de-initialise.
//
// Close all printer files
//
//   pass: void
// return: int				0
//==============================================================================
int printer_deinit (void)
{
 if ((print_a != NULL) && (print_a_file != NULL))
    {
     if (count != 0)
        fprintf(print_a_file, "\n");     
     fclose(print_a_file);
    }     

 if ((print_b != NULL) && (print_b_file != NULL))
    fclose(print_b_file);

 return 0;
}

//==============================================================================
// Printer strobe get.
//
//   pass: void
// return: int				status of printer strobe
//==============================================================================
int printer_strobe_get (void)
{
 return printer_strobe;
} 

//==============================================================================
// Printer strobe set.
//
//   pass: void
// return: void
//==============================================================================
void printer_strobe_set (void)
{
 printer_strobe = 1;
} 

//==============================================================================
// Printer write.
//
//   pass: int data
// return: void
//==============================================================================
void printer_w (uint8_t data)
{
 if ((print_a != NULL) && (print_a_file != NULL))
    {
     fprintf(print_a_file, "%3d ", data);
     if (++count == 16)
        {
         fprintf(print_a_file, "\n");
         count = 0;
        }
    }    

 if ((print_b != NULL) && (print_b_file != NULL))
    fwrite(&data, 1, 1, print_b_file);

 printer_strobe = 0;
}
