//******************************************************************************
//*                                  uBee512                                   *
//*         An emulator for the Microbee Z80 FDD and ROM based models          *
//*                                                                            *
//*                              uBee512 module                                *
//*                                                                            *
//*                   Copyright (C) 2007-2008 Stewart Kay                      *
//******************************************************************************
//
// Provides init, application loop, and deinit.
//
//==============================================================================
// ChangeLog (most recent entries are at top)
//==============================================================================
// v2.3.0 - 23 January 2008, S.J.Kay
// - Removed the vdu_configure() function and changed the way the --pcg
//   option is interpreted.
// - The --vdu option no longer does anything. It remains but does nothing.
// - Added --hint half intensity monochrome option for alpha+ models.
// - Removed the "invalid Z80 code" workaround during initial start up as
//   the problem causing it has now been fixed (see z80.c)
// - Removed the --standard (-s) option and replaced with --col and --mono
//   options that enables/disables colour for standard models.
// - Removed the 'standard' member from the model_t structure.
// - Added new members 'alphap' and 'pcg' to the model_t structure.
// - Added dissasembly options to debug commands (+on+trace+step+alt+count)
// - Added --dump option to set the dump address.
// - Added --bp and --bpc options to set break points for debugging.
// - Added --regs options and regdump_t structure.
// - Added modio_t and debug_t structures,  debugging methods have been
//   completely overhauled with additional functionality and flexbility.
//
// v2.2.0 - 13 December 2007, S.J.Kay
// - Added 'ubee512_ver.id' file to home account to allow later versions to
//   upgrade all the required account files for Unix systems.
// - Changed the default model values to use old colour circuitry where needed.
// - Added a --hwflash option for Premium (alpha+) models and changed the
//   default setting for hardware flashing for the standard Premium model.
// - Create directories in users home accounts for the placement of tools.
// - Changes to the reset section to optionally confirm the reset action. This
//   was required for the reset in the function module.
//
// v2.1.0 - 29 October 2007, S.J.Kay
// - M+RESET keys will not request a confirmation (for jumping into Monitor)
// - Setting of CPU clock speed now handled by set_clock_speed() function. this
//   can be called any time the CPU speed needs changing.
// - Added --lpen option to enable 6545 light pen key emulation for 256tc.
// - Added CPU clock speed functions and --speedsel option.
// - Added PIO port B bit 7 source option --piob7.
// - Added mouse button down and up event detection.
// - Added SDL version to the --version output option.
// - xtal is now cpuclock,  added new option name of --clock, options -x and
//   --xtal will be still be available for now.
// - model not found exits with segmentation fault fixed.
// - *model_args had incorrect ordering of p256k, 256k, and p128k, 128k models,
//   moved the 256k models above the 128k models.
// - The help option output is now organised into sections.
// - Better account creation for Unices when upgrading to later versions. Any
//   missing directories will be created.
// - Start some code re-organisation.
//
// v2.0.0 - 13 October 2007, S.J.Kay
// - Confirmation windows added for Unices builds.
// - Added --nodisk option.
// - Added --rtc option and rtc related functions.
// - Rearranged ordering from most recent to oldest for FDD models and enabled
//   the PJB P256K, 256K and 256TC models.
// - APC model name changed to 56K.
//
// v1.4.0 - 6 October 2007, S.J.Kay
// - Added --mmode option to force the FDD ROM's monitor mode to start.
// - Added --pcg option to select amount of PCG RAM for emulation (Premium).
// - Added --vdu option to select amount of VDU RAM for emulation (Premium).
// - Added --aspect option to select 1:1 or 2:1 Y scaling.
// - Changes for console input/output for the windows port, adding -mwindows
//   to the linking stopped the background console prompt appearing.
// - Added (Windows) --conio option to allow console to be output at all
//   times, else only outputs for --help and --version options or when an
//   error causes the emulator to exit.  If any console output was made a
//   message box appears requiring confirmation to allow the console output
//   to be read before exiting.  By default all stdout and stderr output
//   will redirected to files stdout.txt and stderr.txt files (see SDL).
// - Close window mouse click on [X] is now functional with a confirmation
//   request before exiting in Windows build.
// - Confirmation request for exit and reset keys added for Windows build,
//   ESC+RESET will not request a confirmation.
// - Added --model option.  This allows other model Microbees to be emulated.
// - Changes for win32 build to allow installation to any directory.  This now
//   allows an installer to be used instead of a ZIP file.
// - Main event handler moved out of keyb.c module an used here to dispatch to
//   each event handler as needed. 
// - string_search() function now returns matching index value and -1 if no
//   match found.
// - option --sound=off now does not init sound in SDL. 
// - Start some code re-organisation.
//
// v1.3.0 - 1 September 2007, S.J.Kay
// - Added PIO polling and PIO configure (for interrupts) with execution
//   ratio change feature when interrupts are made active.
// - Added RS232 serial TX and RX options to set communication port,  baud
//   rates for each direction, data and stop bits: --coms, --baud, --baudtx,
//   --baudrx, --datab and --stopb.
// - Added printer output options --print and --printa.
// - Additional directory 'printer' created in the user's home account in the
//   .ubee512 directory.
// - Added --tapevol option for setting volume level of tape wave file.
// - Changed some error messages to put out function name first.
//
// v1.2.0 - 21 August 2007, S.J.Kay
// - Added --tapeo, --tapei, and --tapesamp options to support tape functions.
// - Additional directory 'tapes' created in the user's home account in the
//   .ubee512 directory.
// - Added function module calls for init, deinit and reset.
// - disk_and_roms() function name changed to set_account_paths()
//
// v1.1.0 - 18 August 2007, S.J.Kay
//   Code has been modified to allow the XTAL frequency to be specified.  In
//   theory if the host PC can emulate the Z80 code at that speed then that
//   is what you will get with a frame rate of (n) FPS.
//
//   See the README file for an explanation of how the speed emulation works.
//
// - stdout and stderr on windows now directed back to the console instead
//   of stdout.txt and stderr.txt files.
// - Modified the cursor blinking and Premium hardware inverse/flashing
//   video timer methods depending on if turbo mode is used.  This keeps the
//   flash rate correct.
// - Added frate command line option, mainly intended for hacking.
// - Added vblank command line option, mainly intended for use with the
//   turbo option to make the host timer determine the 50Hz VBLANK rate.
// - Added XTAL command line option (-x) where the xtal rate in MHz can be
//   specified in floating point format.
// - Added sound command line options: off, prop and normal.
// - Added volume setting command line option.
// - Changed sound hold-off to use a timer instead of Z80 cycle counter.
// - Changed from getopt to getopt_long command arguments. Uses same single
//   options as before.
// - Drive B not responding if a drive D was specified.  Problem caused by
//   floppyd[0] incorrectly labeled as floppyb[0] in the disk_and_roms()
//   function clearing any values already set for floppyb.
//
// v1.0.1 - 11 August 2007, S.J.Kay
// - Wrong default disk image type set for boot. Was boot.dip,  changed to
//   boot.dsk
//
// v1.0.0 - 8 August 2007, S.J.Kay
// - A major recoding of most of the original source to improve speed
//   regulation and for the addition of new features.
// - Create directories in users home accounts for the placement of disk and
//   ROM images for first time use.
// - Z80_cycles and associated variables now use unsigned 64 bit variables.
// - Added signal call back handler to allow control from other applications.
// - Added video timers and function calls in application loop for blinking
//   cursor and alpha+ hardware video flashing.
// - Added some code in the application loop to support the new sound emulation
//   feature.
// - Recoded the Z80 CPU speed regulation for the clock target.  The code now
//   uses the gettimeofday() function for unices and clock() function in mingw
//   builds to determine delay periods between mz80exec(n) function calls.
// - Added command line option for standard and standard+colour emulation.
// - Added command line monitor type options.
// - Added command line parameters for disk images for drives A-D.
// - Added turbo command line option.
// - Added fullscreen command line option.
// - Report init failures, not successes.
//
// v0.0.0 - 16 June 2007, S.J.Kay
// Start with "nanowasp" source distribution version 0.22. An emulator for the
// microbee 128k. Copyright (C) 2000-2003  David G. Churchill.

//============================================================================== 
/*
 *  uBee512 - An emulator for the Microbee Z80 FDD and ROM based models. 
 *  Copyright (C) 2007-2008 Stewart Kay
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
//==============================================================================   

#include <stdlib.h>
#include <unistd.h>
#include <getopt.h>
#include <string.h>
#include <time.h>
#include <SDL.h>

#ifdef MINGW
#include <windows.h>
#include <conio.h>
#else
#include <sys/types.h> 		// various type definitions, like pid_t
#include <sys/time.h>
#include <sys/stat.h>
#include <signal.h>		// signal name macros, and the signal() prototype
#endif

#include "ubee512.h"
#include "z80.h"
#include "vdu.h"
#include "crtc.h"
#include "memmap.h"
#include "keyb.h"
#include "fdc.h"
#include "mz80.h"
#include "gui.h"
#include "paknet.h"
#include "pio.h"
#include "sound.h"
#include "tape.h"
#include "serial.h"
#include "rtc.h"
#include "clock.h"
#include "function.h"
#include "z80debug.h"

#include "macros.h"

//==============================================================================
// structures and variables
//==============================================================================
debug_t	debug;
modio_t	modio;
regdump_t regdump;

uint64_t z80_cycles;

int key_esc;
int key_m;
int key_quit;
int key_reset;

int fullscreen;

int flashvideo;

int done;
int loop;

#ifdef MINGW
char execpath[512];
clock_t ticks1, ticks2;
int consoleoutput;
#else
static struct timeval tod;
uint64_t ticks_tod;

char *shared_images[] = {"ubee512-logo.bmp", NULL};

char *shared_disks[] = {
                        "ubee512_cpm_tools.ss80_",
                        "ubee512_cpm_tools.ds40_",
                        "ubee512_cpm_tools.ds80_",
                        "ubee512_cpm_tools.ds82_",
                        "ubee512_cpm_tools.ds84_",
                        NULL
                       };
#endif

int delay;

char userfile[512];
char destfile[512];

char userhome[512];
char userhome_docspath[512];
char userhome_diskpath[512];
char userhome_romspath[512];
char userhome_tapepath[512];
char userhome_prnapath[512];
char userhome_prnbpath[512];
char userhome_imagepath[512];
char userhome_rtcpath[512];
char userhome_toolspath[512];

char floppya[512];
char floppyb[512];
char floppyc[512];
char floppyd[512];

char tapei[512];
char tapeo[512];

char printa[512];
char printb[512];

char coms1dev[512];

#ifdef use_audio
clock_t ticksound;
#endif

SDL_Event event;

volatile char gui_signal;

int sound = 1;			// sound on by default
static int turbo;

char *floppy_a = NULL;
char *floppy_b = NULL;
char *floppy_c = NULL;
char *floppy_d = NULL;

char *tape_i = NULL;
char *tape_o = NULL;
static int tape_samples=TAPE_SAMPLE_FREQ;
static int tape_volume=TAPE_VOLUME;

char *print_a = NULL;
char *print_b = NULL;

char *coms1_dev = NULL;
int serial_tx_baud = SERIAL_TX_BAUD;
int serial_rx_baud = SERIAL_RX_BAUD;
int serial_databits = SERIAL_DATABITS;
int serial_stopbits = SERIAL_STOPBITS;

int frame_rate = FRAMERATE;	// default frame rate
extern int audio_volume;
int vblank_method = 0;

int model = MOD_DEFAULT;	// default model

int cpuclock;
int z80ms;

int monitor = 0;		// 0=colour monitor
int z80_ratio;			// current z80 execution ratio (default = 1)

char *model_args[] =
{
 "256tc",
 "p512k",
 "512k",
 "p256k",
 "256k",
 "p128k",
 "128k",
 "p64k",  
 "64k",
 "56k",
 "ppc85",
 "pc85",  
 "pc",
 "ic",
 "2mhz",
 ""
 };

const model_t model_data[MOD_TOTAL] =
{
//               ALPHAP     ROM    boot  RAM  PCG   COLOUR   HWF  MHI LPEN    SPEED       PIOB7     RTC  CLOCK
 {/* 256tc */         1,      0, 0x8000, 256,   8, MODCOL2,    1,   1,   0,  MODSPD, MODPB7_RTC, MODRTC, 3.375},
 {/* p512k */         1,      0, 0x8000, 512,  16, MODCOL2,    1,   1,   1,  MODSPD, MODPB7_RTC, MODRTC, 3.375},
 {/*  512k */         0,      0, 0x8000, 512,   1, MODCOL1,    0,   0,   1,  MODSPD, MODPB7_RTC, MODRTC, 3.375},
 {/* p256k */         1,      0, 0x8000, 256,  16, MODCOL2,    1,   1,   1,  MODSPD, MODPB7_RTC, MODRTC, 3.375},
 {/*  256k */         0,      0, 0x8000, 256,   1, MODCOL1,    0,   0,   1,  MODSPD, MODPB7_RTC, MODRTC, 3.375},
 {/* p128k */         1,      0, 0x8000, 128,   8, MODCOL2,    0,   0,   1,       0, MODPB7_PUP,      0, 3.375},
 {/*  128k */         0,      0, 0x8000, 128,   1,       0,    0,   0,   1,       0, MODPB7_PUP,      0, 3.375},
 {/*  p64k */         1,      0, 0x8000,  64,   8, MODCOL2,    0,   0,   1,       0, MODPB7_PUP,      0, 3.375},
 {/*   64k */         0,      0, 0x8000,  64,   1,       0,    0,   0,   1,       0, MODPB7_PUP,      0, 3.375},
 {/*   56k */         0,      0, 0xE000,  56,   1,       0,    0,   0,   1,       0, MODPB7_PUP,      0, 3.375},
 {/* ppc85 */         1, MODROM, 0x8000,  32,   8, MODCOL2,    0,   0,   1,       0,  MODPB7_VS,      0, 3.375},
 {/*  pc85 */         0, MODROM, 0x8000,  32,   1, MODCOL1,    0,   0,   1,       0,  MODPB7_VS,      0, 3.375},
 {/*    pc */         0, MODROM, 0x8000,  32,   1, MODCOL1,    0,   0,   1,       0,  MODPB7_VS,      0, 3.375},
 {/*    ic */         0, MODROM, 0x8000,  32,   1,       0,    0,   0,   1,       0, MODPB7_PUP,      0, 3.375},
 {/*  2mhz */         0, MODROM, 0x8000,  32,   1,       0,    0,   0,   1,       0, MODPB7_PUP,      0, 2.000}
};

model_t modelx;

static int opt_aspect=2;

static int z80_block_cycles_def; // default number of Z80 cycles in 1 block
static int z80_blocks_def;	 // default number of Z80 blocks

static int z80_block_cycles_cur; // current number of Z80 cycles in 1 block
static int z80_blocks_cur;	// current number of Z80 blocks

int z80_block_cycles;		// working number of Z80 cycles in 1 block
int z80_blocks;			// working number of Z80 blocks

extern int fdc_status0;

extern int extendram;
extern int crtcresized;

extern int cur_blink;
extern int cur_mode;
extern int cur_pos;

extern int audio_silence;
extern int dobuffering;

extern int dump_addr;

//==============================================================================
// Search an array of strings for the first occurence of the passed search
// string.  The string array must be be terminated by an empty string. 
//
//   pass: char *strg_array	pointer to a string array to be checked
//         char *strg_find	pointer to a string to search for
// return: int			index if found, else -1
//==============================================================================
static int string_search (char *strg_array[], char *strg_find)
{
 int i;
 
 for (i=0; strg_array[i][0] != 0; i++)
    {
     if (strcmp(strg_find, strg_array[i]) == 0)
        return i;
    }
 return -1;
}        

#ifdef MINGW
#else
//==============================================================================
// External GUI signal handler.
//
// This allows an optional external GUI program to change values during the
// running of the emulator.  The code here is kept as simple as possible and
// simply just sets a flag to indicate that a signal was received.  The main
// loop checks the flag value to see if any action is required.
//
// Can be tested in unices by using the following to cause a reset():
// $kill -s 30 pid
//
// To find the pid use:
// $ps ax | grep ubee512
//
//   pass: int sig_num
// return: void
//==============================================================================
static void signal_handler (int sig_num)
{
 gui_signal = 1;
}
#endif

//==============================================================================
// Event checking.
//
//   pass: void
// return: void
//==============================================================================
void event_handler (void)
{
 while (SDL_PollEvent(&event))
    {
     switch (event.type)
        {
         case SDL_KEYDOWN:
            keyb_keydown_event();
            break; 
         case SDL_KEYUP:
            keyb_keyup_event();
            break;
         case SDL_QUIT:
            if (gui_ask_exit_emulator())
               done = 1;
            break;
         case SDL_MOUSEBUTTONDOWN:
            gui_mousebuttondown_event();
            break;
         case SDL_MOUSEBUTTONUP:
            gui_mousebuttonup_event();            
            break;
        }                    
    }
}

//==============================================================================
// Create user home account directories and determine paths to be for disks,
// ROMs, tapes, and printers.
//
//   pass: void
// return: int				0 if no errors, else -1
//==============================================================================
static int set_account_paths (void)
{
 int i;

#ifdef MINGW

// TEMPORARY - This will be changed in a later version
#define execpath userhome

// get the path and executable filename
 if (GetModuleFileName(NULL, execpath, sizeof(execpath)) == 0)
    {
     console_output();    
     printf("ubee512: Unable to find the executable path\n");
     return -1;
    }

// delete the executable name part and last '\' character
 i = strlen(execpath);
 while (execpath[--i] != SLASHCHAR)
    {}
 execpath[i] = 0;

#if 0
// get the environment path for application data
 if (GetEnvironmentVariable("APPDATA", userhome, sizeof(userhome)) == 0)
    {
     console_output();    
     printf("ubee512: Unable to obtain the APPDATA environment\n");
     return -1;
    }
 strcat(userhome, "/ubee512"); 

// printf("env var=%s\n", userhome);
#endif

#else

 char *s;

 int dir_created;
 int install_files;
 char install_version[80];
 int x;

 mode_t mode;
 FILE *textfp;
 
 s = getenv("HOME");
 if (s == NULL)
    {
     console_output();
     printf("ubee512: Unable to find the user's home path\n");
     return -1;
    }
 else
    strcat(userhome, s);

 strcat(userhome, "/.ubee512"); 
#endif

 snprintf(userhome_docspath, sizeof(userhome_docspath), "%s%s", userhome, DIR_DOCS);
 snprintf(userhome_diskpath, sizeof(userhome_diskpath), "%s%s", userhome, DIR_DISKS); 
 snprintf(userhome_romspath, sizeof(userhome_romspath), "%s%s", userhome, DIR_ROMS);
 snprintf(userhome_tapepath, sizeof(userhome_tapepath), "%s%s", userhome, DIR_TAPES);
 snprintf(userhome_prnapath, sizeof(userhome_prnapath), "%s%s", userhome, DIR_PRINTER);
 snprintf(userhome_prnbpath, sizeof(userhome_prnbpath), "%s%s", userhome, DIR_PRINTER);
 snprintf(userhome_imagepath, sizeof(userhome_imagepath), "%s%s", userhome, DIR_IMAGES);
 snprintf(userhome_rtcpath, sizeof(userhome_rtcpath), "%s%s", userhome, DIR_RTC);
 snprintf(userhome_toolspath, sizeof(userhome_toolspath), "%s%s", userhome, DIR_TOOLS); 
 
#ifdef MINGW
#else
 mode = S_IRWXG | S_IRWXU | S_IROTH | S_IXOTH;

// create the directory account if it does not already exist.
 dir_created = mkdir(userhome, mode) == 0;

 snprintf(userfile, sizeof(userfile), "%s/ubee512_ver.id", userhome); 
 
 if (! dir_created)  // if not created then already have a .ubee512 directory
    {
     textfp = fopen(userfile, "r");
     if (textfp == NULL)
        install_files = 1;
     else
        {   
         x = fscanf(textfp, "%s", install_version);
         fclose(textfp);
         install_files = (strcmp(install_version, APPVER) != 0) || (x != 1);
        } 
    }
 else
    install_files = 1;   
    
 if (install_files)
    {
     textfp = fopen(userfile, "w");
     if (textfp != NULL)
        {
         fprintf(textfp, APPVER"\n");
         fclose(textfp);
        } 

// create the sub-directories if not already present     
     mkdir(userhome_romspath, mode);
     mkdir(userhome_tapepath, mode);
     mkdir(userhome_prnapath, mode);
     mkdir(userhome_prnbpath, mode);
     mkdir(userhome_rtcpath, mode);
     mkdir(userhome_diskpath, mode);
     mkdir(userhome_docspath, mode);
     mkdir(userhome_imagepath, mode);
     mkdir(userhome_toolspath, mode);
     
// copy files from the shared images directory to the user's home account
     for (i = 0; shared_images[i] != NULL; i++)
        {
         snprintf(userfile, sizeof(userfile), PATH_SHARED_IMAGES"%s", shared_images[i]);
         snprintf(destfile, sizeof(destfile), "%s%s", userhome_imagepath, shared_images[i]);
         copy_file(destfile, userfile);
        }

// copy files from the shared disks directory to the user's home account
     for (i = 0; shared_disks[i] != NULL; i++)
        {
         snprintf(userfile, sizeof(userfile), PATH_SHARED_DISKS"%s", shared_disks[i]);
         snprintf(destfile, sizeof(destfile), "%s%s", userhome_diskpath, shared_disks[i]);
         copy_file(destfile, userfile);
        }

// create readme.txt file telling the user where to find documentation
     snprintf(userfile, sizeof(userfile), "%sreadme.txt", userhome_docspath);
     textfp = fopen(userfile, "w");
     if (textfp != NULL)
        {
         fprintf(textfp, "Documentation including licensing information is located in:\n");
         fprintf(textfp, PATH_SHARED_DOCS"\n");
         fclose(textfp);
        } 

// create readme.txt file telling the user where to find tools
     snprintf(userfile, sizeof(userfile), "%sreadme.txt", userhome_toolspath);
     textfp = fopen(userfile, "w");
     if (textfp != NULL)
        {
         fprintf(textfp, "CP/M Z80 emulator tools sources are located in:\n");
         fprintf(textfp, PATH_SHARED_TOOLS"\n");
         fclose(textfp);
        } 

     console_output();
     printf("\n");
     printf("==========================================================================\n");
     if (dir_created)
        printf("This is the first time you have run the ubee512 emulator in this account.\n");
     else
        {
         printf("A different version of the emulator has been detected.  Additional\n");
         printf("directories and files may have been created in your account.\n");
        }
     printf("\n");
     printf("You now need to copy ROMs: charrom.bin, and either disk and/or BASIC ROMs\n");
     printf("to the ROMs directory if these don't already exist:\n");
     printf("%s\n", userhome_romspath);
     printf("\n");
     printf("Copy any required disk image(s) and make sure these are writable to:\n");
     printf("%s\n", userhome_diskpath);     
     printf("\n");     
     printf("When you have done this run the ubee512 emulator again.\n");          
     printf("==========================================================================\n");
     exit(0);		// exit so user can copy the required files
    } 
#endif

 if (! modelx.rom)
    {    
// Image A
     if (floppy_a != NULL)
        {
         if (strchr(floppy_a, SLASHCHAR) == NULL)
            {
             floppya[0] = 0;
             strcat(floppya, userhome_diskpath);
             strcat(floppya, floppy_a); 
            }
         else    
            strcat(floppya, floppy_a);
        }

// Image B
     if (floppy_b != NULL)
        {
         if (strchr(floppy_b, SLASHCHAR) == NULL)
            {
             floppyb[0] = 0;
             strcat(floppyb, userhome_diskpath);      
             strcat(floppyb, floppy_b); 
            }
         else    
            strcat(floppyb, floppy_b);
        }

// Image C
     if (floppy_c != NULL)
        {
         if (strchr(floppy_c, SLASHCHAR) == NULL)
            {
             floppyc[0] = 0;
             strcat(floppyc, userhome_diskpath);      
             strcat(floppyc, floppy_c); 
            }
         else    
            strcat(floppyc, floppy_c);
        }

// Image D
     if (floppy_d != NULL)
        {
         if (strchr(floppy_d, SLASHCHAR) == NULL)
            {
             floppyd[0] = 0;
             strcat(floppyd, userhome_diskpath);      
             strcat(floppyd, floppy_d); 
            }
         else    
            strcat(floppyd, floppy_d);
        }
    }

// Tape input file
 if (tape_i != NULL)
    {
     if (strchr(tape_i, SLASHCHAR) == NULL)
        {
         tapei[0] = 0;
         strcat(tapei, userhome_tapepath);      
         strcat(tapei, tape_i); 
        }
     else    
        strcat(tapei, tape_i);
     printf("Tape in file: %s\n", tapei);
    }

// Tape output file
 if (tape_o != NULL)
    {
     if (strchr(tape_o, SLASHCHAR) == NULL)
        {
         tapeo[0] = 0;
         strcat(tapeo, userhome_tapepath);      
         strcat(tapeo, tape_o); 
        }
     else    
        strcat(tapeo, tape_o);
     printf("Tape out file: %s\n", tapeo);
    }

 if ((tape_o != NULL) && (tape_i != NULL))
    if (strcmp(tapeo, tapei) == 0)
       {
        console_output();
        printf("ubee512: tape in and out can not be the same file\n");
        return -1;
       }

// Printer (ASCII) output file
 if (print_a != NULL)
    {
     if (strchr(print_a, SLASHCHAR) == NULL)
        {
         printa[0] = 0;
         strcat(printa, userhome_prnapath);      
         strcat(printa, print_a); 
        }
     else    
        strcat(printa, print_a);
     printf("Printer (ASCII) file: %s\n", printa);
    }

// Printer (BINARY) output file
 if (print_b != NULL)
    {
     if (strchr(print_b, SLASHCHAR) == NULL)
        {
         printb[0] = 0;
         strcat(printb, userhome_prnapath);      
         strcat(printb, print_b); 
        }
     else    
        strcat(printb, print_b);
     printf("Printer (BINARY) file: %s\n", printb);
    }

// Serial communications device #1
 if (coms1_dev != NULL)
    {
     strcat(coms1dev, coms1_dev);
     printf("Serial communications dev: %s\n", coms1dev);
    }

 return 0;
}

//==============================================================================
// Initialise.
//
//   pass: void
// return: int				0 if no errors, else -1
//==============================================================================
static int init (void)
{
 SDL_Surface *icon;
 uint32_t sdl_init_properties;

 printf(TITLESTRING"\n");

 if (set_account_paths() != 0)
    return -1;
 
#ifdef MINGW
#else
// set the GUI signal handler
 signal(30, signal_handler);
#endif

 if (modio.level > 0)
    {
     modio.log = fopen(LOGFILE, "w"); 
     if (modio.log == NULL)
        {
         console_output();
         printf("init: Unable to write to log file\n");
         return -1;
        }
     fprintf(modio.log, "modio level = %i\n", modio.level);
    }

 sdl_init_properties = SDL_INIT_VIDEO | SDL_INIT_TIMER;
 if (sound)
    sdl_init_properties |= SDL_INIT_AUDIO;

 if (SDL_Init(sdl_init_properties) != 0)
    {
     console_output();
     printf("init: Failed SDL_Init - %s\n", SDL_GetError());
     return -1;
    }

 SDL_EventState(SDL_ACTIVEEVENT | SDL_KEYUP | SDL_KEYDOWN |
 SDL_MOUSEBUTTONDOWN | SDL_SYSWMEVENT, SDL_IGNORE);

 SDL_WM_SetCaption(TITLESTRING, ICONSTRING);

 userfile[0] = 0;
 strcat(userfile, userhome_imagepath);
 strcat(userfile, "ubee512-logo.bmp"); 

 if ((icon = SDL_LoadBMP(userfile)) != NULL)
    {
     SDL_SetColorKey(icon, SDL_SRCCOLORKEY, SDL_MapRGB(icon->format, 0, 0, 0));
     SDL_WM_SetIcon(icon, NULL);
     SDL_FreeSurface(icon);
    }

 if (gui_init() != 0)
    {
     console_output();
     printf("init: Failed gui_init\n");    
     return -1;
    } 
 else
    {
    }   

 if (z80_init() != 0)
    {
     console_output();
     printf("init: Failed z80_init\n");    
     return -1;
    } 

 if (clock_init() != 0)
    {
     console_output();
     printf("init: Failed clock_init\n");
     return -1;
    } 

 if (vdu_init() != 0)
    {
     console_output();
     printf("init: Failed vdu_init\n");
     return -1;
    } 

 if (memmap_init() != 0)
    {
     console_output();
     printf("init: Failed memmap_init\n");    
     return -1;
    } 

 if (crtc_init() != 0)
    {
     console_output();
     printf("init: Failed crtc_init\n");
     return -1;
    }

 if (keyb_init() != 0)
    {
     console_output();
     printf("init: Failed keyb_init\n");
     return -1;
    } 

 if (fdc_init() != 0)
    {
     console_output();
     printf("init: Failed fdc_init\n");
     return -1;
    } 

 if (pio_init() != 0)
    {
     console_output();
     printf("init: Failed pio_init\n");
     return -1;
    } 

 if (paknet_init() != 0)
    {
     console_output();
     printf("init: Failed paknet_init\n");
     return -1;
    } 

 if (rtc_init() != 0)
    {
     console_output();
     printf("init: Failed rtc_init\n");
     return -1;
    } 

 if (function_init() != 0)
    {
     console_output();
     printf("init: Failed function_init\n");
     return -1;
    } 

 return 0;
}

//==============================================================================
// De-initialise.
//
//   pass: void
// return: int			result of de-init functions
//==============================================================================
static int deinit(void)
{
 int res;

 if (modio.level > 0)
    fclose(modio.log);

 res = function_deinit() || pio_deinit() || rtc_deinit() || paknet_deinit() ||
       fdc_deinit() || vdu_deinit() || keyb_deinit() || crtc_deinit() ||
       memmap_deinit() || clock_deinit() ||z80_deinit() || gui_deinit();

 SDL_Quit();

 return res;
}

//==============================================================================
// Reset the virtual Microbee.
//
// Make sure all reset functions clear variables that are constantly active
// values.
//
//   pass: void
// return: int			result of reset functions
//==============================================================================
static int reset (void)
{
 int res;
 
 z80_cycles = 0;
 done = 0;
 loop = 0;

#ifdef use_resetinfo
 printf("ubee512: emulation reset\n");
#endif

#ifdef MINGW
 ticks1 = clock();
#else
 gettimeofday(&tod, NULL);          
 ticks_tod = (uint64_t)((tod.tv_sec * 1000) + (tod.tv_usec / 1000));
#endif
 
 res =  z80_reset() || clock_reset() || memmap_reset() || crtc_reset() ||
        keyb_reset() || vdu_reset() || fdc_reset() || pio_reset() ||
        paknet_reset() || rtc_reset() || function_reset();

 return res;
}

//==============================================================================
// Z80 execution ratio set.
//
// The z80 execute method may be changed to execute fewer Z80 cycles in one Z80
// block but with more blocks by changing the ratio of these two values. 
//
// This is intended to allow PIO interrupts to be serviced faster as each
// interrupt is only serviced once on each call to mz80exec().  The ratio
// used can be set back to default after interrupts have completed.
//
//   pass: int divider		use 1 for default values
// return: void
//==============================================================================
void z80_exec_ratio (int divider)
{
 if (divider == 1)
    {
     z80_block_cycles_cur = z80_block_cycles_def;
     z80_blocks_cur = z80_blocks_def;
    }
 else
    {
     z80_block_cycles_cur = z80_block_cycles_def / divider;
     z80_blocks_cur = divider;
    } 
 z80_ratio = divider;
}

//==============================================================================
// Set clock speed setting
//
// All modules dependent on the CPU clock speed are set up from here.  This
// function may be called at any time to change the execution speed.  It is
// used by the "port 0x0a read" speed change.
//
//   pass: float clock		CPU clock speed in MHz
// return: void
//==============================================================================
void set_clock_speed (float clock)
{
 float speedadj;
 float z80ms_r;

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// CPU clock configuration
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// convert clock to an integer in Hz
 cpuclock = (int)(clock * 1E6);

// calculate the number of Z80 cycles executed in 1 frame 
 z80_block_cycles_def = cpuclock / frame_rate;

// execution time in mS needed for one lot of z80_block_cycles_def
 z80ms_r = ((float)z80_block_cycles_def / cpuclock) * 1000.0;

// calculate a speed adjustment based on a constant percentage
 speedadj = (float)z80ms * (SPEED_PCT_ADJ / 100.0);

// apply the adjustment value
 z80ms = (int)(z80ms_r - speedadj);
 if (z80ms < 0)
    z80ms = 0;

 z80_blocks_def = 1;		// default blocks

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// Z80 PIO configuration
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 pio_configure(cpuclock, frame_rate, serial_rx_baud); // set exec ratios prota/serial

 z80_exec_ratio(1);	// use the values calculated here

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// Tape configuration
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 tape_config_out(cpuclock, tape_samples, tape_volume);
 tape_config_in(cpuclock);

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// RS232 Serial configuration
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 serial_config(cpuclock, serial_tx_baud, serial_rx_baud, serial_databits, serial_stopbits);

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// Sound configuration
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 switch (sound)
    {
     case 0 : break;
     case 1 : sound_clock(cpuclock, SOUND_PROP); // sound proportional to clock
              break;
     case 2 : sound_clock(cpuclock, SOUND_3375); // sound set to normal rate
              break; 
    }             

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// CRTC configuration
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 crtc_clock(cpuclock, vblank_method);  // set VBLANK method
}

//==============================================================================
// Application setup.
//
//   pass: void
// return: void
//==============================================================================
void ApplicationSetup (void)
{
 set_clock_speed(modelx.cpuclock);
 crtc_configure(opt_aspect);

 reset();			// resets the virtual Microbee

#ifdef MINGW
 ticks1 = clock();   
#else
 uint64_t ticks_tod;
 gettimeofday(&tod, NULL); 
 ticks_tod = (uint64_t)((tod.tv_sec * 1000) + (tod.tv_usec / 1000));
#endif 
}

//==============================================================================
// Application loop.
//
// Video timers
// ------------
// The CRTC blinking cursor and Alpha+ hardware inverse and flashing video
// timers use the accumulated Z80 cycle count,  this works well as the Z80 CPU
// execution is regulated with delays and so the instruction count is accurate
// over a period of time,  the rate will increase in turbo mode.
//
//   pass: void
// return: void
//==============================================================================
static void ApplicationLoop (void)
{
 int dwResult;
 int tstates;
 struct mz80context z80; 

 int cur_blink_rate;
 int cur_blink_last=0;
 int flashvideo_last=0;
 
 while (! done)
    {
#ifdef MINGW
     clock_t ticks_s; 		// for sound hold-off period
     ticks1 = clock();
#else
     uint64_t ticks_tod_snd;    // for sound hold-off period
     struct timeval tod_snd;
     struct timeval tod_x;     
     gettimeofday(&tod, NULL);          
     ticks_tod = (uint64_t)((tod.tv_sec * 1000) + (tod.tv_usec / 1000));
#endif

// Looping feature for speeding up PIO interrupt responses by having a
// smaller z80_block_cycles value.  Normally z80_blocks is set to 1 with
// z80_block_cycles equal to the default calculated rate.

     if (debug.debug)
        z80_blocks = 5000;
     else
        {   
         z80_blocks = z80_blocks_cur;
         z80_block_cycles = z80_block_cycles_cur;
        } 

     while (z80_blocks--)
        {
// Disassemble the next Z80 instruction
         if (debug.debug)
            {
             tstates = z80debug_before();

// Execute tstates cycles (determined by next instruction type)
             if (tstates != -1)
                {
                 dwResult = mz80exec(tstates);

// Show the state of the Z80 registers, etc
                 z80debug_after(dwResult);
                } 
            }
         else
            {
// Execute a block (Z80CYCLES) of Z80 instructions and return the result.
// The Z80CYCLES value used has been calculated for timing purposes.
             dwResult = mz80exec(z80_block_cycles);

// If the result is not 0x80000000, it's an address where an invalid
// instruction was hit.
             if (dwResult != 0x80000000)
                {
                 mz80GetContext(&z80);
                 console_output();
                 printf("Invalid instruction at: 0x%04x\n", (unsigned int)z80.z80pc);
                 printf("     Opcodes (first 4): 0x%02x 0x%02x 0x%02x 0x%02x\n",
                 (uint8_t)z80mem[z80.z80pc+0], (uint8_t)z80mem[z80.z80pc+1],
                 (uint8_t)z80mem[z80.z80pc+2], (uint8_t)z80mem[z80.z80pc+3]);
                 printf("           Clock ticks: %ld\n", z80.z80clockticks);
                 z80_blocks = 0;
                 done = 1;
                }
            }

// Accumulate the count of Z80 cycles
         z80_cycles += mz80GetElapsedTicks(1);

// Poll the PIO for interrupt events
         pio_polling();
        } 

// Resize the video when certain CRTC registers have been changed under Z80
// emulation. A wait period is introduced here to prevent the display being
// resized too often in a short time period.
     if (crtcresized)
        {
         if (crtcresized++ == 2)	// will be the 2nd time around loop
            crtc_videochange();		// resets crtcresized
        }

// Determine the current status for the CRTC blinking cursor and refresh it
// if this has changed.  The method used here depends on if turbo mode is
// used. if turbo mode is used then Z80 execution speed will not be known as
// no delays will be inserted, if not turbo then the rate must be determined
// by the Z80 cycle count to achieve smooth results.
     if (cur_mode > 1)			// only if blinking cursor
        {
         if (turbo)
            {
             if (cur_mode == 2)
                cur_blink_rate = 500;	// blinking at 1/32 field rate
             else   
                cur_blink_rate = 250;	// blinking at 1/16 field rate
#ifdef MINGW
             if ((clock() / cur_blink_rate) & 0x01)
#else
             gettimeofday(&tod_x, NULL);          
             if (((tod_x.tv_usec / 1000) / cur_blink_rate) & 0x01) // rolls over each sec
#endif
                cur_blink = 0xFF;
             else
                cur_blink = 0x00;
            }
         else
            {
             if (cur_mode == 2)
                cur_blink_rate = cpuclock / 2;	// blinking at 1/32 field rate
             else   
                cur_blink_rate = cpuclock / 4;	// blinking at 1/16 field rate        
             if ((z80_cycles / cur_blink_rate) & 0x01)
                cur_blink = 0xFF;
             else
                cur_blink = 0x00;
            }            
         if (cur_blink != cur_blink_last)
            {
             cur_blink_last = cur_blink;            
             crtc_redraw_char(cur_pos, 0);
            }
        }

// Determine the current state of the alpha+ flashing video and refresh it
// if this has changed.  the method used here is the same as that used by
// the blinking cursor timer.
     if (extendram)			// only if extended RAM selected
        {
         if (turbo)
            {
#ifdef MINGW
             if ((clock() / 250) & 0x01)
#else
             gettimeofday(&tod_x, NULL);          
             if (((tod_x.tv_usec / 1000) / 250) & 0x01) // rolls over each sec
#endif
                flashvideo = modelx.hwflash;
             else
                flashvideo = 0;
            }
         else
            {
             if ((z80_cycles / (cpuclock / 4)) & 0x01)
                flashvideo = modelx.hwflash;
             else
                flashvideo = 0;
            }
         if (flashvideo != flashvideo_last)
            {
             flashvideo_last = flashvideo;            
             crtc_set_redraw();	// force crtc to redraw screen
            }
        }    

#if use_audio
// Sound buffering.  This is required as CPU emulation speed is affected by
// other processes occuring at the same time, we need to try to ensure the
// sound is slightly in front so that it does not matter if the Z80 lags
// behind slightly in creating the sound.  A 100mS delay has been found
// to be about right.  If the buffer delay is too long the lag will be
// noticable by the user.  If too short then break up, and crackling will be
// evident.
     if (dobuffering)
        {
         if (ticksound == -1)		// if just starting the buffering
#ifdef MINGW
            {
             ticksound = 0;
             ticks_s = clock();
            } 
#else
            { 
             ticksound = 0;            
             gettimeofday(&tod_snd, NULL);          
             ticks_tod_snd = (uint64_t)((tod_snd.tv_sec * 1000) + (tod_snd.tv_usec / 1000));
            }
#endif
         else
            {
#ifdef MINGW
             if ((clock() - ticks_s) >= SOUND_HOLD_OFF)
#else
             gettimeofday(&tod_snd, NULL);          
             if ((((uint64_t)((tod_snd.tv_sec * 1000) + (tod_snd.tv_usec / 1000))) - ticks_tod_snd) >= SOUND_HOLD_OFF)
#endif
                sound_unsilence();	// allow sound to play
            }
        }    
#endif

// Update the display (only if it's update flag is set)           
     gui_redraw();

// Check and handle any pending events.
     event_handler();

// Handle external GUI signals
     if (gui_signal)
        {
         reset();
         gui_signal = 0;
        }

// Check for exit and reset key flags and get confirmation
     if (key_quit)
        {
         key_quit = 0;
         if (gui_ask_exit_emulator())
            done = 1;
        } 
     if (key_reset)
        {
         if (key_reset == 2)
            {
             key_reset = 0;
             reset();
            }
         else
            {    
             key_reset = 0;
             if ((! key_esc) && (! key_m))	// if ESC+ RESET or M then no prompting
                {
                 if (gui_ask_reset_emulator())
                    reset();
                }
             else
                reset();
             key_esc = 0;
            } 
        }

// Insert delays to achieve correct emulated Z80 CPU speed.  The Z80 speed
// may be made a little faster to help prevent audio break up (in ubee512.h).
//
// The MINGW build relies on a 1mS timer (1ms granularity) and the delay
// required is calculated from the last block (Z80CYCLES) of Z80
// instructions.
//
// The non MINGW builds (unices) uses the gettimeofday function and this is
// converted to a 64 bit variable with a 1mS resolution.
//
     if (turbo)
        SDL_Delay(0);
     else
        {
#ifdef MINGW
// Find the execution delay required for the last (1/frame_rate)
         ticks2 = clock() - ticks1;
         if (ticks2 >= 0)		// if clock() has not wrapped around
            {                   	// else use the last delay value
             delay = z80ms - ticks2;
            }
#else
         gettimeofday(&tod, NULL);
         delay = z80ms - ((uint64_t)((tod.tv_sec * 1000) + (tod.tv_usec / 1000)) - ticks_tod);
#endif
         
// Execute the delay period.  If a delay is not required then a delay of 0
// is used to give up the CPU for other processes.
         if (delay < 0)
            SDL_Delay(0);
         else
            SDL_Delay(delay);
        }
    }    
}

//==============================================================================
// Application usage
//
// The help information is arranged in sections, each section and members of
// each section are arranged in alphabetical order.
//
//   pass: void
// return: void
//==============================================================================
static void ApplicationUsage (void)
{
 char const usage[] =
TITLESTRING"\n"
"\n"
"Usage: ubee512 [options]\n"
"\n"
// +++++++++++++++++++++++++++ Control related +++++++++++++++++++++++++++++++++
" Control related:\n\n"
"  --mmode                 Forces the return of the 'M' key once when the\n"
"                          emulator is started.  This is used for jumping\n"
"                          directly into the ROM's Monitor mode.\n"
"\n"
"  --nodisk                Force no disks initially,  used to start some boot\n"
"                          ROMs in menu mode.\n"
"\n"
// ++++++++++++++++++++++++++++ Debugging tools ++++++++++++++++++++++++++++++++
" Debugging tools:\n\n"
"  --bp=addr               Set a Z80 address break point.  This option can be\n"
"                          specified as many times as is required.  A break\n"
"                          point can only be detected when in debug mode.\n"
"\n"
"  --bpc=count             Set a Z80 break point determined by the number of\n"
"                          instructions executed. Can only be specified once.\n"
"                          A break point can only be detected when in debug\n"
"                          mode.\n"
"\n"
"  -z, --debug=options     Debugging mode.  Arguments commence with a '+'\n"
"                          character,  additional arguments may be declared in\n"
"                          the same --debug option.\n"
"\n"
"                          The arguments supported are:\n"
"                          +alt - output the alternate and the I, R registers.\n"
"                          +all - output all options.\n"
"                          +count - use instruction counter in disassembly.\n"
"                          +index - output the index registers.\n"
"                          +on - turn debugging mode on without disassembly.\n"
"                          +step - turn on debugging mode and start stepping.\n"
"                          +trace - turn on debugging  mode and start tracing.\n"
"\n"
"  --dump=addr             Set the initial dump address value when using the dump\n"
"                          commands.  addr may be a valid Z80 address from 0 to\n"
"                          0xFFFF. If not specified then address 0 is used.\n"
"\n"
"  --modio=options         Module I/O debugging output.  Arguments commence with\n"
"                          a '+' character,  additional arguments may be\n"
"                          declared in the same --modio option.\n"
"\n"
"                          The arguments supported are:\n"
"                          +log - logs to ubee512_log.txt (makes big log files!)\n"
"                          +raminit - uses bank numbers as DRAM init values.\n"
"\n"
"                          These arguments turn on port debugging for modules:\n"
"                          +clock, +crtc, +fdc, +func, +keystd, +keytc, +mem,\n"
"                          +paknet, +pioa, +piob, +piocont, +rtc, +vdu, +vdumem,\n"
"                          +z80.\n"
"\n"
"  --regs=options          Register dump. Determines what registers will be\n"
"                          dumped when the  EMUKEY+R key is pressed.  Arguments\n"
"                          commence with a '+' character,  additional arguments\n"
"                          may be declared in the same --regs option.\n"
"                          The arguments supported are:\n"
"                          +crtc, +rtc, +z80.\n"
"\n"
// ++++++++++++++++++++++++++ Disk drive images ++++++++++++++++++++++++++++++++
" Disk drive images :\n\n"
"  -a, --image_a=file      Image file path & filename for floppy drive A\n"
"  -b, --image_b=file      Image file path & filename for floppy drive B\n"
"  -c, --image_c=file      Image file path & filename for floppy drive C\n"
"  -d, --image_d=file      Image file path & filename for floppy drive D\n"
"\n"
"                          If you specify a path as part of the image name then\n"
"                          it will be accessed from that area.  If only an\n"
"                          image name is specified the emulator will look for\n"
"                          it here:\n"
"\n"
"                          Windows:\n"
"                          installed location\\ubee512\\disks\\\n"
"\n"
"                          Unices:\n"
"                          your home path/.ubee512/disks/\n"
"\n"
"                          Hint: if you want to access an image in the current\n"
"                          directory you need to force it using the following\n"
"                          method:\n"
"\n"
"                          Windows:\n"
"                          ubee512 -a .\\myimage.dsk\n"
"\n"
"                          Unices:\n"
"                          ubee512 -a ./myimage.dsk\n"
"\n"
// +++++++++++++++++++++++++++ Display related +++++++++++++++++++++++++++++++++
" Display related:\n\n"
"  --aspect=n              Set the display aspect, n=1 is 1:1,  default aspect\n"
"                          2:1 (n=2),  n may be set to 1 or 2.  1:1 scaling\n"
"                          may be enforced for some CRTC6545 display sizes'.\n" 
"\n"
"  -f, --fullscreen        Start in full screen mode (use ALT+ENTER to toggle).\n"
"\n"
"  -m, --monitor=type      Monitor type, if this option is not specified a\n"
"                          colour monitor is the default. <type> may be one\n"
"                          of the following:\n"
"\n"
"                          a : amber monitor\n"
"                          g : green monitor\n"
"                          w : white monitor, white foreground on black\n"
"                              background\n"
"                          b : black monitor, black foreground on white\n"
"                              background\n"
"                          c : colour monitor\n"
"\n"
"                          Note: This option by itself does not force the\n"
"                          emulation into a standard model Microbee, it's use\n"
"                          simply determines what monitor type is connected\n"
"                          to the emulated Microbee.\n"
"\n"
// +++++++++++++++++++++++++++ Information output ++++++++++++++++++++++++++++++
" Information output:\n\n"
"  -h, --help, --usage     Display help information on command line usage.\n"
"\n"
"  --conio                 Switches on verbose console output for Windows port.\n"
"                          By default only fatal errors and some option's output\n"
"                          is sent to the console.\n"
"\n"
"  --version               Obtain the version number of the emulator and SDL.\n"
"\n"
// ++++++++++++++++++++++++++++ Model emulation ++++++++++++++++++++++++++++++++
" Model emulation:\n\n"
"  --col                   Enables early colour circuit emulation for standard\n"
"                          models.  This option has no affect when emulating a\n"
"                          Premium or 256TC model.\n"
"\n"
"  --hint=x                Half intensity monochrome emulation for Premium\n"
"                          (alpha+) models.  x=on to enable, x=off to disable.\n"
"                          This is set to 'on' by default for the 256TC and\n"
"                          upgraded Premium models.  This option has no affect\n"
"                          when emulating a standard model.\n"
"\n"
"  --hwflash=x             Hardware Inverse and flashing video emulation for\n"
"                          Premium (alpha+) models.  x=on to enable, x=off to\n"
"                          disable.  This is set to 'on' by default for the 256TC\n"
"                          and upgraded Premium models.  This option has no\n"
"                          affect when emulating a standard model.\n"
"\n"
"  --lpen                  Enables the use of the 6545 Light pen keys emulation.\n"
"                          This is enabled by default for all models except for\n"
"                          the 256tc model.\n"
"\n"
"  --model=type            Model type, if this option is not specified the p512k\n"
"                          model is emulated.  The 'p' denotes a Premium\n"
"                          variation in a model.  This option should be used\n"
"                          before any other options.  The following models are\n"
"                          supported: 256tc, p512k, 512k, p256k, 256k, p128k,\n"
"                          128k, p64k, 64k, 56k, ppc85, pc85, pc, ic, and 2mhz.\n"
"\n"
"  --mono                  Disables early colour circuit emulation for standard\n"
"                          models.  This option when emulating a Premium or 256TC\n"
"                          model has the same affect as --monitor=g.\n"
"\n"
"  --pcg=n                 Premium model option that sets size of PCG RAM to be\n"
"                          emulated.  n is the size of the PCG memory in\n"
"                          Kilobytes and can be any even value between 2 and 32.\n"
"                          256TC and Premium models are 16K,  upgraded Premium\n"
"                          models are 32K.  This option has no affect when\n"
"                          emulating a standard model.\n"
"\n"
"  --piob7=signal          Determines what signal is used for PIO port B bit 7.\n"
"                          The default value depends on the model emulated.\n"
"                          The source values are: rtc, vsync, net, and pup.\n"
"\n"
"  --vdu=n                 Deprecated option.  Will be ignored if specified.\n"
"\n"
// ++++++++++++++++++++ Parallel printer port emulation ++++++++++++++++++++++++
" Parallel printer port emulation:\n\n"
"  --print=file            Printer output to file,  the output is not modified.\n"
"                          The location of the file follows the same rule as\n"
"                          used for the drive image options except default\n"
"                          directory is 'printer'.\n"
"\n"
"  --printa=file           Printer output to file,  the output is converted to\n"
"                          ASCII decimal. The location of the file follows the\n"
"                          same rule as used for the drive image options except\n"
"                          default directory is 'printer'.\n"
"\n"
// ++++++++++++++++++++ Real Time Clock (RTC) emulation ++++++++++++++++++++++++
" Real Time Clock (RTC) emulation:\n\n"
"  --rtc=n                 Real Time Clock (RTC) emulation. n=1 to enable, n=0\n"
"                          to disable.  The following models use RTC by\n"
"                          default: 256tc, p512k, 512k, p256k, 256k.\n"
"\n"
// +++++++++++++++++++++++++ Serial port emulation +++++++++++++++++++++++++++++
" Serial port emulation:\n\n"
"  --baud=rate             Set serial communications baud rate for both TX and\n"
"                          RX.  A value from 1 to 19200 is allowed.  Default\n"
"                          rate is 300 baud.  If Individual baud rates are\n"
"                          required for TX and RX then use the --baudtx and\n"
"                          --baudrx options instead.  This value must match the\n"
"                          Microbee serial application's value.\n"
"\n"
"  --baudtx=rate           Set serial communications baud rate for TX only.  A\n"
"                          value from 1 to 19200 is allowed.  Default rate is\n"
"                          300 baud.  This value must match the Microbee serial\n"
"                          application's value.\n"
"\n"
"  --baudrx=rate           Set serial communications baud rate for RX only.  A\n"
"                          value from 1 to 19200 is allowed.  Default rate is\n"
"                          300 baud. This value must match the Microbee serial\n"
"                          application's value.\n"
"\n"
"  --coms=port             Serial communications port for emulation of RS232.\n"
"                          On Unices specify a device, on Windows specify the\n"
"                          com port number.  No serial communications will be\n"
"                          emulated if this option is not specified.\n"
"\n"
"  --datab=bits            Set serial communications number of data bits.  A\n"
"                          value from 5 to 8 is allowed.  Default value is 8\n"
"                          data bits.  This value must match the Microbee\n"
"                          serial application's value.\n"
"\n"
"  --stopb=bits            Set serial communications number of stop bits.  A\n"
"                          value from 1 to 2 is allowed.  Default value is 1\n"
"                          data bits.  This value must match the Microbee\n"
"                          serial application's value for TX.\n"
"\n"
// +++++++++++++++++++++++++++ Sound emulation +++++++++++++++++++++++++++++++++
" Sound emulation:\n\n"
"  --sound=method          Determine the method used for sound,  the default\n"
"                          method is 'prop':\n"
"\n"
"                             off : sound is turned off\n"
"                            prop : sound is proportional to CPU clock frequency\n"
"                          normal : sound rate forced as if 3.375 MHz CPU clock\n"
"\n"
"  --vol=level             Set the sound volume level.  A level of 0 to 127 is\n"
"                          allowed. (default level is 16)\n"
"\n"
// ++++++++++++++++++++++++++++ Speed related ++++++++++++++++++++++++++++++++++
" Speed related:\n\n"
"  --frate=fps             Frame rate,  the default is 30 FPS,  an integer value\n"
"                          between 1 and 1,000,000 is allowed.  This option is\n"
"                          intended for 'hacking' and code development use.\n"
"\n"
"  --speedsel=n            CPU speed selection emulation. n=1 to enable, n=0\n"
"                          to disable.  The following models are enabled by\n"
"                          default: 256tc, p512k, 512k, p256k, 256k.\n"
"\n"
"  -t, --turbo             Turbo mode, executes Z80 code as fast as possible.\n"
"                          Without this option the emulation attempts to keep\n"
"                          Z80 CPU execution to match the CPU clock value.  This\n"
"                          option is intended for 'hacking' and code development\n"
"                          use.  There are much faster methods if more speed is\n"
"                          required. (see the README file)\n"
"\n"
"  --vblank=method         Vertical blanking method to be employed. This is\n"
"                          only intended for 'hacking' when experimenting with\n"
"                          turbo mode and/or high CPU clock speeds.  It is not\n"
"                          required or even recommended to be used if 3.375 MHz\n"
"                          or 2 MHz emulation is desired.\n"
"\n"
"                          0 : 50 Hz VBLANK rate derived from Z80 cycles and is\n"
"                              proportional to the CPU clock frequency.\n"
"                          1 : 50 Hz VBLANK rate derived from the host timer.\n"
"\n"
"                          When running in turbo mode then setting the <method>\n"
"                          equal to 1 will ensure that a VBLANK rate of 50Hz\n"
"                          will be used.  Without this, key repeating may be\n"
"                          too fast.\n"
"\n"
"  -x, --clock=f --xtal=f  Set the Z80 clock frequency for emulation in MHz.\n"
"                          Standard emulation frequencies are 3.375 and 2.0\n"
"                          MHz.  All other frequencies are classed as 'hacking'.\n" 
"                          3.375 MHz is the default clock frequency.  Frequency\n"
"                          is entered as a floating point number.\n"
"\n"
"                          The highest frequency possible is determined by the\n"
"                          host platform.  As the value increases to the point\n"
"                          where uBee512 can no longer regulate the Z80\n"
"                          execution rate the frame rate will decrease (slower\n"
"                          screen update periods).\n"
"\n"
// +++++++++++++++++++++++++ Tape port emulation +++++++++++++++++++++++++++++++
" Tape port emulation:\n\n"
"  --tapei=file            Tape input from a WAV file.  The location of the file\n"
"                          follows the same rule as used for the drive image\n"
"                          options except default directory is 'tapes'.\n"
"\n"
"  --tapeo=file            Tape output to a WAV file.  The location of the file\n"
"                          follows the same rule as used for the drive image\n"
"                          options except default directory is 'tapes'.\n"
"\n"
"  --tapesamp=frequency    Tape output sample frequency in Hz.  Default is\n"
"                          22050 Hz.\n"
"\n"
"  --tapevol=level         Tape output wave file volume level.  A level of 0 to\n"
"                          127 is allowed. (default level is 16)\n"
"\n"
// +++++++++++++++++++++++++++++++++ other +++++++++++++++++++++++++++++++++++++
"If you have any new feature suggestions, bug reports, etc. then please send\n"
"an email to: ubee512@gmail.com\n";

 console_output();
 printf("%s", usage);
}

//==============================================================================
// console input/output.
//
// Windows:
// Turns on console input/output so that messages can be seen on a text
// console window.  If this is not called then console out will be
// redirected to stdout.txt and stderr.txt files This is set in the SDL main
// section.
//
// Unices:
// Output for now relies on the emulator started from a command line.
//
//   pass: void
// return: void
//==============================================================================
void console_output (void)
{
#ifdef MINGW
 if (! consoleoutput)
    {
     consoleoutput = 1;
     AllocConsole();
     freopen("conin$", "r", stdin);
     freopen("conout$", "w", stdout);
     freopen("conout$", "w", stderr);
    }
#else
#endif
}

//==============================================================================
// Parameter error message.
//
//   pass: char *parameter
//         char *option_arg
// return: void
//==============================================================================
static void param_error_mesg (char *parameter, char *option_arg)
{
#define PARMERR_MESG "ubee512: option %s argument of '%s' is not permitted\n" 
 console_output();
 printf(PARMERR_MESG, parameter, option_arg);
}

//==============================================================================
// Main program start point.
//
//   pass: int argc, char *argv[]	Command line arguments
// return: int				0 if no error, 1 if error
//==============================================================================
int main (int argc, char *argv[])
{
 static struct option long_options[] =
    {
// Control related
     {"mmode",      no_argument,       0,  23},
     {"nodisk",     no_argument,       0,  24},     

// Debugging tools
     {"bp",         required_argument, 0,  35},
     {"bpc",        required_argument, 0,  36},
     {"dump",       required_argument, 0,  32},
     {"debug",      required_argument, 0, 'z'},
     {"modio",      required_argument, 0,  37},     
     {"regs",       required_argument, 0,  34},     

// Disk drive images
     {"image_a",    required_argument, 0, 'a'},
     {"image_b",    required_argument, 0, 'b'},
     {"image_c",    required_argument, 0, 'c'},
     {"image_d",    required_argument, 0, 'd'},

// Display related
     {"aspect",     required_argument, 0, 19 },
     {"fullscreen", no_argument,       0, 'f'},
     {"monitor",    required_argument, 0, 'm'},

// Model emulation
     {"col",        no_argument,       0, 30 },
     {"hint",       required_argument, 0, 33 },
     {"hwflash",    required_argument, 0, 29 },
     {"model",      required_argument, 0, 17 },
     {"mono",       no_argument,       0, 31 },     
     {"pcg",        required_argument, 0, 20 },
     {"vdu",        required_argument, 0, 21 },
     {"piob7",      required_argument, 0, 27 },
     {"lpen",       no_argument,       0, 28 },

// Information output
     {"help",       no_argument,       0, 'h'},
     {"version",    no_argument,       0,  3 },
     {"usage",      no_argument,       0, 'h'},
     {"conio",      no_argument,       0, 18 },

// Parallel printer port emulation
     {"print",      required_argument, 0, 15 },
     {"printa",     required_argument, 0, 16 },          

// Serial port emulation
     {"coms",       required_argument, 0,  9 },
     {"baud",       required_argument, 0, 10 },
     {"baudtx",     required_argument, 0, 11 },
     {"baudrx",     required_argument, 0, 12 },
     {"datab",      required_argument, 0, 13 },
     {"stopb",      required_argument, 0, 14 },

// Sound emulation
     {"sound",      required_argument, 0,  1 },
     {"vol",        required_argument, 0, 'v'},

// Speed related
     {"frate",      required_argument, 0,  4 },
     {"turbo",      no_argument,       0, 't'},
     {"vblank",     required_argument, 0,  2 },
     {"xtal",       required_argument, 0, 'x'},
     {"clock",      required_argument, 0, 26 },     
     {"speedsel",   required_argument, 0, 25 },

// Tape port emulation
     {"tapei",      required_argument, 0,  5 },
     {"tapeo",      required_argument, 0,  6 },
     {"tapesamp",   required_argument, 0,  7 },
     {"tapevol",    required_argument, 0,  8 },

// Real Time Clock (RTC) emulation
     {"rtc",        required_argument, 0, 22 },

     {0, 	    0,                 0,  0 }
    };

// Do not re-arrange the ordering on any of these values
 char *monitor_args[] =
 {
  "c",
  "a",
  "g",
  "b",
  "w",
  ""
 };

 char *sound_args[] =
 {
  "off",
  "prop",
  "normal",
  ""
 };

 char *piob7_args[] =
 {
  "pup", 
  "vsync",
  "rtc",  
  "net",
  ""
 };

 char *offon_args[] =
 {
  "off",
  "on",
  ""
 };

 const SDL_version *sdlv;

 int int_arg;
 float float_arg;
 
 char temp_str[512];
 
#define TRY_MESG "%s: Try `%s --help' or `%s --usage' for more information.\n"

 int c;
 int exitstatus = 0;

// get a copy of the default model data to work with
 memcpy(&modelx, &model_data[model], sizeof(model_t));

 while (1)
    {
// getopt_long stores the option index here.
     int option_index = 0;
     
     c = getopt_long (argc, argv, "fhta:b:c:d:m:v:x:z:", long_options, &option_index);
     
// detect the end of the options.
     if (c == -1)
        break;
     
// do integer and floating point conversions
     if (optarg != NULL)
        {
         int_arg = -1;
         if ((strchr(optarg, 'x') == NULL) && (strchr(optarg, 'X') == NULL))
            sscanf(optarg, "%d", &int_arg);
         else   
            sscanf(optarg, "%x", &int_arg);         
         float_arg = -1.0;
         sscanf(optarg, "%f", &float_arg);
        } 
     
     switch (c)
        {
         case 'a': floppy_a = optarg;
                   break;
         case 'b': floppy_b = optarg;
                   break;
         case 'c': floppy_c = optarg;
                   break;
         case 'd': floppy_d = optarg;
                   break;
         case 'f': fullscreen = 1;
                   break;
         case 'h': ApplicationUsage();
                   exitstatus = -1;
                   break;
         case 'm': monitor = string_search(monitor_args, optarg);
                   if (monitor == -1)
                      {
                       param_error_mesg("'-m, --monitor'", optarg);
                       exitstatus = 1;
                       break;
                      } 
                   break;
         case 't': turbo = 1;
                   break;
         case 'v': int_arg = -1;
                   sscanf(optarg, "%d", &int_arg);
                   if ((int_arg < 0) || (int_arg > 127))
                      {
                       param_error_mesg("'--vol'", optarg);
                       exitstatus = 1;
                       break;
                      }
                   else
                      audio_volume = int_arg;
                   break;
         case 'x': if (float_arg <= 0.0)
                      {
                       param_error_mesg("'-x, --xtal'", optarg);
                       exitstatus = 1;
                       break;
                      }
                   else
                      modelx.cpuclock = float_arg;    
                   break;
         case 'z': snprintf(temp_str, sizeof(temp_str), "%s+", optarg);
                   debug.debug = 1;
                   if (strstr(temp_str, "+index+"))
                      debug.show |= B8(00000001);
                   if (strstr(temp_str, "+alt+"))
                      debug.show |= B8(00000010);
                   if (strstr(temp_str, "+all+"))
                      debug.show |= B8(00000011);                      
                   if (strstr(temp_str, "+count"))
                      debug.showcount = 1;
                   if (strstr(temp_str, "+step+"))
                      function_emu_command(EMU_CMD_DBGSTEP01);
                   if (strstr(temp_str, "+trace+"))
                      function_emu_command(EMU_CMD_DBGTRACE);
                   break;
         case  1 : sound = string_search(sound_args, optarg);
                   if (sound == -1) 
                      {
                       param_error_mesg("'--sound'", optarg);
                       exitstatus = 1;
                       break;
                      } 
                   break;
         case  2 : if ((int_arg < 0) || (int_arg > 1))
                      {
                       param_error_mesg("'--vblank'", optarg);
                       exitstatus = 1;
                       break;
                      }
                   else
                      vblank_method = int_arg;    
                   break;
         case  3 : console_output();
                   printf("%s\n", APPVER);
                   sdlv = SDL_Linked_Version();
                   printf("SDL Version: %u.%u.%u\n", sdlv->major, sdlv->minor, sdlv->patch);                   
                   exitstatus = -1;
                   break;                    
         case  4 : if ((int_arg < 1) || (int_arg > 1000000L))
                      {
                       param_error_mesg("'--frate'", optarg);
                       exitstatus = 1;
                       break;
                      }
                   else
                      frame_rate = int_arg;    
                   break;
         case  5 : tape_i = optarg;
                   break;
         case  6 : tape_o = optarg;
                   break;                   
         case  7 : if ((int_arg < 1) || (int_arg > 1000000L))
                      {
                       param_error_mesg("'--tapesamp'", optarg);
                       exitstatus = 1;
                       break;
                      }
                   else
                      tape_samples = int_arg;    
                   break;
         case  8 : if ((int_arg < 0) || (int_arg > 127))
                      {
                       param_error_mesg("'--tapevol'", optarg);
                       exitstatus = 1;
                       break;
                      }
                   else
                      tape_volume = int_arg;
                   break;   
         case  9 : coms1_dev = optarg;
                   break;
         case 10 : if ((int_arg < 1) || (int_arg > 19200))
                      {
                       param_error_mesg("'--baud'", optarg);
                       exitstatus = 1;
                       break;
                      }
                   else
                      {
                       serial_tx_baud = int_arg;
                       serial_rx_baud = int_arg;
                      } 
                   break;
         case 11 : if ((int_arg < 1) || (int_arg > 19200))
                      {
                       param_error_mesg("'--baudtx'", optarg);
                       exitstatus = 1;
                       break;
                      }
                   else
                      serial_tx_baud = int_arg;
                   break;
         case 12 : if ((int_arg < 1) || (int_arg > 19200))
                      {
                       param_error_mesg("'--baudrx'", optarg);
                       exitstatus = 1;
                       break;
                      }
                   else
                      serial_rx_baud = int_arg;
                   break;
         case 13 : if ((int_arg < 5) || (int_arg > 8))
                      {
                       param_error_mesg("'--datab'", optarg);
                       exitstatus = 1;
                       break;
                      }
                   else
                      serial_databits = int_arg;
                   break;
         case 14 : if ((int_arg < 1) || (int_arg > 2))
                      {
                       param_error_mesg("'--stopb'", optarg);
                       exitstatus = 1;
                       break;
                      }
                   else
                      serial_stopbits = int_arg;
                   break;
         case 15 : print_b = optarg;
                   break;
         case 16 : print_a = optarg;
                   break;
         case 17 : model = string_search(model_args, optarg);
                   if (model == -1)
                      {
                       param_error_mesg("'--model'", optarg);
                       exitstatus = 1;
                       break;
                      }
                   else
                      memcpy(&modelx, &model_data[model], sizeof(model_t));
                   break;
         case 18 : console_output();
                   break;
         case 19 : if ((int_arg < 1) || (int_arg > 2))
                      {
                       param_error_mesg("'--aspect'", optarg);
                       exitstatus = 1;
                       break;
                      }
                   else
                      opt_aspect = int_arg;
                   break;
         case 20 : if ((int_arg < 2) || (int_arg > 32) || (int_arg % 2))
                      {
                       param_error_mesg("'--pcg'", optarg);
                       exitstatus = 1;
                       break;
                      }
                   else
                      if (modelx.alphap)
                         modelx.pcg = int_arg / 2;
                   break;
         case 21 : break;
         case 22 : if ((int_arg != 0) && (int_arg != 1))
                      {
                       param_error_mesg("'--rtc'", optarg);
                       exitstatus = 1;
                       break;
                      }
                   else
                      modelx.rtc = int_arg;
                   break;
         case 23 : keyb_force(0x0d, 100);	// force into monitor mode
                   break;
         case 24 : fdc_status0 = 5;
                   break;
         case 25 : if ((int_arg < 0) || (int_arg > 1))
                      {
                       param_error_mesg("'--speedsel'", optarg);
                       exitstatus = 1;
                       break;
                      }
                   else
                      modelx.speed = int_arg;
                   break;
         case  26: if (float_arg <= 0.0)
                      {
                       param_error_mesg("'--clock'", optarg);
                       exitstatus = 1;
                       break;
                      }
                   else
                      modelx.cpuclock = float_arg;    
                   break;
         case  27: modelx.piob7 = string_search(piob7_args, optarg);
                   if (modelx.piob7 == -1) 
                      {
                       param_error_mesg("'--piob7'", optarg);
                       exitstatus = 1;
                       break;
                      } 
                   break;
         case  28: modelx.lpen = 1;
                   break;
         case  29: modelx.hwflash = string_search(offon_args, optarg);
                   if (modelx.hwflash == -1)
                      {
                       param_error_mesg("'--hwflash'", optarg);
                       exitstatus = 1;
                       break;
                      }
                   break;                   
         case 30 : if (! modelx.alphap)
                      modelx.colour = 1;
                   break;
         case 31 : if (modelx.alphap)
                      monitor = string_search(monitor_args, "g");
                   else   
                      modelx.colour = 0;
                   break;
         case 32 : if ((int_arg < 0) || (int_arg > 0xFFFF))
                      {
                       param_error_mesg("'--dump'", optarg);
                       exitstatus = 1;
                       break;
                      }
                   else
                      dump_addr = int_arg;    
                   break;
         case 33 : modelx.halfint = string_search(offon_args, optarg);
                   if (modelx.halfint == -1)
                      {
                       param_error_mesg("'--hint'", optarg);
                       exitstatus = 1;
                       break;
                      }
                   break;
         case 34 : snprintf(temp_str, sizeof(temp_str), "%s+", optarg);
                   if (strstr(temp_str, "+crtc+"))
                      regdump.crtc = 1;
                   if (strstr(temp_str, "+rtc+"))
                      regdump.rtc = 1;
                   if (strstr(temp_str, "+z80+"))
                      regdump.z80 = 1;
                   break;
         case 35 : if ((int_arg < 0) || (int_arg > 0xFFFF))
                      {
                       param_error_mesg("'--bp'", optarg);
                       exitstatus = 1;
                       break;
                      }
                   else
                      debug.break_point[int_arg] = 1;
                   break;
         case 36 : if (int_arg < 0)
                      {
                       param_error_mesg("'--bpc'", optarg);
                       exitstatus = 1;
                       break;
                      }
                   else
                      debug.break_point_count = int_arg;
                   break;                   
         case 37 : snprintf(temp_str, sizeof(temp_str), "%s+", optarg);
                   if (strstr(temp_str, "+log+"))
                      modio.level = 1;
                   if (strstr(temp_str, "+clock+"))
                      modio.clock = 1;
                   if (strstr(temp_str, "+crtc+"))
                      modio.crtc = 1;
                   if (strstr(temp_str, "+fdc+"))
                      modio.fdc = 1;
                   if (strstr(temp_str, "+func+"))
                      modio.func = 1;
                   if (strstr(temp_str, "+keystd+"))
                      modio.keystd = 1;
                   if (strstr(temp_str, "+keytc+"))
                      modio.keytc = 1;
                   if (strstr(temp_str, "+mem+"))
                      modio.mem = 1;
                   if (strstr(temp_str, "+paknet+"))
                      modio.paknet = 1;
                   if (strstr(temp_str, "+pioa+"))
                      modio.pioa = 1;
                   if (strstr(temp_str, "+piob+"))
                      modio.piob = 1;
                   if (strstr(temp_str, "+piocont+"))
                      modio.piocont = 1;
                   if (strstr(temp_str, "+raminit+"))
                      modio.raminit = 1;
                   if (strstr(temp_str, "+rtc+"))
                      modio.rtc = 1;
                   if (strstr(temp_str, "+vdu+"))
                      modio.vdu = 1;
                   if (strstr(temp_str, "+vdumem+"))
                      modio.vdumem = 1;
                   if (strstr(temp_str, "+z80+"))
                      modio.z80 = 1;
                   break;
         case '?': // getopt_long already printed an error message.
                   console_output();
#ifdef MINGW
                   // now do it again becuase it would only have gone
                   // to stderr.txt
                   printf("%s: invalid option -- %c\n", argv[0], c);
#endif
                   printf(TRY_MESG, argv[0], argv[0], argv[0]);
                   exitstatus = 1;
                   break;
         default : exitstatus = 1;
                   break;
        }
    }

// Check if any other arguments are incorrectly specified
 if (optind < argc)
    {
     console_output();
     exitstatus = 1;
     printf("%s: %d additional arguments were specified that are not recognised:\n", argv[0], argc - optind);         
     printf("%s: ", argv[0]);
     while (optind < argc)
        printf ("%s ", argv[optind++]);
     putchar('\n');
     printf(TRY_MESG, argv[0], argv[0], argv[0]);
    }

// if no option errors then initialise the emulator
 if (exitstatus == 0)
    {
     if (init() != 0)
        {
         console_output();
         exitstatus = 1;
         printf("main: Fatal error during initialisation.\n");
         SDL_Quit();
        }
    }

// if no errors then start the emulator
 if (exitstatus == 0)
    {
     ApplicationSetup();
     ApplicationLoop();
    }

// if no errors then de-initialise the emulator
 if (exitstatus == 0)
    {
     if (deinit() != 0)
        {
         console_output();
         exitstatus = 1;
         printf("main: Error while de-initialising.\n");
        }
    }    

// if running on Windows then get a confirmation before the console
// output window is closed.
 if (exitstatus)
    {
#ifdef MINGW
     gui_message_box(BUTTON_OK, "Read message(s) in console output window before closing.");
#endif
     if (exitstatus == -1)
        return 0;
     else   
        return exitstatus;
    }

 return 0;
}
