//******************************************************************************
//*                                  uBee512                                   *
//*         An emulator for the Microbee Z80 FDD and ROM based models          *
//*                                                                            *
//*                                 VDU module                                 *
//*                                                                            *
//*                   Copyright (C) 2007-2008 Stewart Kay                      *
//******************************************************************************
//
// Emulates the graphics hardware for Standard and Premium models (Alpha+)
//
// - Character ROM
// - Screen RAM
// - PCG RAM
// - Colour RAM
// - Attribute RAM
//
//==============================================================================
// ChangeLog (most recent entries are at top)
//==============================================================================
// v2.3.0 - 23 January 2008, S.J.Kay
// - VDU and PCG mask variables and vdu_configure() function removed.  The
//   number of PCG banks is now handled correctly.  The VDU mask is now a
//   constant as all alpha+ models had 8K screen, attribute and colour RAM.
//   The VDU_BANK0 define is set to 0 for screen and attribute RAM, a value
//   of 3 is not working, 256TC boot ROM has incorrect graphics and locking
//   is evident when it's 3. VDU_BANK3 define is used for extended colour
//   RAM and appears to work correctly.
// - ROM Basic 5.22e colour failure mystery solved and fixed by emulating
//   static memory pattern.  Removed VDU init code from memory vdu_reset()
//   function.
// - Many changes made to the colour and alpha+ graphics read/write and the
//   video switching in/out functions.
// - Made changes to the vdu_switchvideo_in() function to suit all model types.
// - Changes to the vdu_lvdat_r() and vdu_lvdat_w() functions due to changes
//   made in z80.c
// - Added modio_t structure.
//
// v2.2.0 - 16 November 2007, S.J.Kay
// - Make the colour_cont port variable public as is needed for old colour
//   circuit emulation.
// - Port 8 read now returns 0.
//
// v2.1.0 - 20 October 2007, S.J.Kay
// - Implement the modelx information structure.
//
// v1.4.0 - 30 September 2007, S.J.Kay
// - Added vdu_configure() function to set PCG and VDU values for Premium
//   models.
// - Added Premium PC85 ROM model basic upper/lower half select to vdu_lvdat_w
//   function.
// - Removed the non aplpha+ source build option.
// - Change all error output to stdout.
// - Character ROM loading now uses the memmap_loadrom() function.
// - Added changes to error reporting.
//
// v1.0.0 - 21 June 2007, S.J.Kay
// - Added standard and standard+colour machine options to vdu_colcont_w and
//   vdu_lvdat_w functions.
// - Added vdu_lvdat_r and vdu_lvdat_w port handlers.
// - Implemented Alpha+ (Premium) graphics.
//
// v0.0.0 - 8 June 2007, S.J.Kay
// Start with "nanowasp" source distribution version 0.22. An emulator for the
// microbee 128k. Copyright (C) 2000-2003  David G. Churchill.

//==============================================================================
/*
 *  uBee512 - An emulator for the Microbee Z80 FDD and ROM based models. 
 *  Copyright (C) 2007-2008 Stewart Kay
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/ 
//==============================================================================

#include <stdio.h>
#include <string.h>
#include <SDL.h>

#include "vdu.h"
#include "crtc.h"
#include "z80.h"
#include "ubee512.h"
#include "memmap.h"
#include "mz80.h"

#include "macros.h"

//==============================================================================
// structures and variables
//==============================================================================
#define VDU_MASK0	B8(00000000)
#define VDU_MASK3	B8(00000011)

int latchrom;
int attribram;			// LV DAT attribute RAM selected if bit 4=1
int colourram;			// colour RAM selected if bit 4=1 else PCG RAM
int extendram;	        	// LV DAT extended graphics selected if bit 7=1

static int videobank;		// LV DAT bank bits 0-3

int colour_cont;		// colour control port (0x08) value
static int x_colour_cont;

static int lv_dat;		// LV DAT port (0x1c) value
static int x_lv_dat;

static char romimage[512];
char charrom[CHARROMSIZE];

// Alpha+ (Premium) variables, 8K for Screen, Colour and attribute, 32K for PCG
char screen_ram[4 * 0x0800];
char colour_ram[4 * 0x0800];
char attrib_ram[4 * 0x0800];
char pcg_ram[16 * 0x0800];

extern char userhome_romspath[512];
extern int mem_mode;

extern int model;
extern model_t modelx;

extern modio_t modio;

//==============================================================================
// VDU initialise
//
// Make VDU RAM location look like how real static RAM would.  This is needed
// by Basic 5.22e (possibly 6.22e) to detect if hardware supports colour.
// It appears as if 5.22e Basic does not set 0xFFFF to 0xFF in the PCG RAM and
// just assumes it will be non zero from a cold start because of the typical
// static RAM pattern value when starting up the Microbee ?  Running TBASICC.COM
// works when loaded from CP/M because the system already set 0xFFFF to 0xFF
// when it made the inverse character set.
//
// Any unused alpha+ PCG banks will contain random data, testing on a Premium
// Microbee shows the values as mostly zerobut can vary greatly between tests.
// The emulation here just uses all zeros.
//
//   pass: void
// return: int				0 if success else -1
//==============================================================================
int vdu_init (void)
{
 int i;
 
// Alpha+ (Premium) variables, 8K for Screen, Colour and attribute, 32K for PCG
 if (modelx.alphap)
    {
     memmap_init6264(screen_ram, 4);
     memmap_init6264(colour_ram, 4);
     memmap_init6264(attrib_ram, 4);
     for (i = 0; i < 16; i++)
        {
         if (i < modelx.pcg)
            memmap_init6264(pcg_ram + i * 0x0800, 1);
         else   
            memset(pcg_ram + i * 0x0800, 0x00, 0x0800);
        }    
    }
 else
// Standard model 2K Screen, 2K Colour, 2K PCG
    {
     memmap_init6116(screen_ram, 1);
     memmap_init6116(colour_ram, 1);
     memmap_init6116(pcg_ram, 1);
    } 

 if (modelx.alphap)
    {
     lv_dat = 0;			// port (0x1c) value
     videobank = 0;                     // bank bits 0-3
     attribram = 0;                     // attribute RAM select
     extendram = 0;                     // extended graphics select
     memmap_basicrom(lv_dat & B8(00100000));
    } 
 
 if (modelx.colour)
    {
     colour_cont = 0;
     x_colour_cont = 0;
     colourram = 0;			// set to PCG
    } 

 vdu_switchvideo_in();			// switch in VDU
 
 romimage[0] = 0;
 strcat(romimage, userhome_romspath); 
 strcat(romimage, "charrom.bin"); 

 if (memmap_loadrom(romimage, charrom, CHARROMSIZE) != CHARROMSIZE)
    {
     console_output();
     printf("vdu_init: unable to read %s\n", romimage);
     return -1;
    }
 
 return 0;
}

//==============================================================================
// VDU de-initialise
//
//   pass: void
// return: int				0
//==============================================================================
int vdu_deinit (void)
{
 return 0;
}

//==============================================================================
// VDU reset
//
//   pass: void
// return: int				0
//==============================================================================
int vdu_reset (void)
{
 latchrom = 0;

 vdu_switchvideo_out();			// switch out current VDU

 if (modelx.alphap)
    {
     lv_dat = 0;			// port (0x1c) value
     x_lv_dat = 0;
     videobank = 0;                     // bank bits 0-3
     attribram = 0;                     // attribute RAM select
     extendram = 0;                     // extended graphics select
     memmap_basicrom(lv_dat & B8(00100000));
    } 
 
 if (modelx.colour)
    {
     colour_cont = 0;
     x_colour_cont = 0;
     colourram = 0;			// set to PCG
    } 

 vdu_switchvideo_in();			// switch in VDU

 return 0;
}

//==============================================================================
// Switch video in.
//
// Switch in the following video banks to 8000/8800 or F000/F800:
// 
// 1. Copy attribute/screen RAM (bit 4=1 if attribute) bank (bits 0-3) to
//    main memory.
// 2. Copy PCG/Colour RAM bank (bits 0-3) to main memory. The state of port
//    0x08 will determine if PCG or Colour RAM 
//
// NOTES:
// - if bit 7=0 then bank 0 still must be put back in to place for the above.
// - character ROM is never in main memory so no check is needed for it.
//
//   pass: void
// return: void
//==============================================================================
void vdu_switchvideo_in (void)
{
 int vb;
 int video;

// DRAM models, port 50h determines video enabled and location
 if (modelx.ram >= 64)
    {
     if (mem_mode & BANK_VRAM)		// if video RAM disabled
        return;
     else
        {
         if (mem_mode & BANK_VADD)	// if video at 8000-8FFF
            video = 0x8000;
         else
            video = 0xF000;
        }
    }
 else
    video = 0xF000;   			// all other non banked models

// determine the video bank to use
 if (extendram)		// if the alpha+ graphics is enabled
    vb = videobank;
 else
    vb = 0;   

// VDU switching required for DRAM and alpha+ models
 if ((modelx.alphap) || (modelx.ram >= 64))
    {
     if (attribram)
        memcpy(z80mem+video, &attrib_ram[(vb & VDU_MASK0) * 0x0800], 0x800);
     else
        memcpy(z80mem+video, &screen_ram[(vb & VDU_MASK0) * 0x0800], 0x800);
    }

// VDU switching required for DRAM and colour models
 if ((modelx.colour) || (modelx.ram >= 64))
    {    
     if (colourram)
        memcpy(z80mem+video+0x0800, &colour_ram[(vb & VDU_MASK3) * 0x0800], 0x800);
     else
        memcpy(z80mem+video+0x0800, &pcg_ram[vb * 0x0800], 0x800);
    }

 if (modio.vdumem)
    {
     printf("vdu_switchvideo_in: video addr=0x%04x vb=0x%02x\n", video, vb);
     if (modio.level)
        fprintf(modio.log, "vdu_switchvideo_in: video addr=0x%04x vb=0x%02x\n", video, vb);
    }    
}

//==============================================================================
// Switch video out.
//
// This should only be called after video has been switched in.  Use the
// current video settings when calling!
//
// Switch out memory 8000/8800 or F000/F800 to video banks:
// 
// 1. Copy attribute/screen RAM (bit 4=1 if attribute) bank (bits 0-3) to
//    banked memory.
// 2. Copy PCG/Colour RAM to banked memory (bits 0-3). The state of port
//    0x08 will determine if PCG or Colour RAM 
//
// NOTES:
// - if bit 7=0 then bank 0 still must be put back in to place for the above.
// - character ROM is never in main memory so no check is needed for it.
//
//   pass: void
// return: void
//==============================================================================
void vdu_switchvideo_out (void)
{
 int vb;
 int video;

// DRAM models, port 50h determines video enabled and location
 if (modelx.ram >= 64)
    {
     if (mem_mode & BANK_VRAM)		// if video RAM disabled
        return;
     else
        {
         if (mem_mode & BANK_VADD)	// if video at 8000-8FFF
            video = 0x8000;
         else
            video = 0xF000;
        }
    }
 else
    video = 0xF000;   			// all other non banked models

// determine the video bank to use
 if (extendram)		// if the alpha+ graphics is enabled
    vb = videobank;
 else
    vb = 0;   

// VDU switching required for DRAM and alpha+ models
 if ((modelx.alphap) || (modelx.ram >= 64))
    {
     if (attribram)
        memcpy(&attrib_ram[(vb & VDU_MASK0) * 0x0800], z80mem+video, 0x800);
     else
        memcpy(&screen_ram[(vb & VDU_MASK0) * 0x0800], z80mem+video, 0x800);
    }    

// VDU switching required for DRAM and colour models
 if ((modelx.colour) || (modelx.ram >= 64))
    {
     if (colourram)
        memcpy(&colour_ram[(vb & VDU_MASK3) * 0x0800], z80mem+video+0x0800, 0x800);
     else
        memcpy(&pcg_ram[vb * 0x0800], z80mem+video+0x0800, 0x800);
    }    

 if (modio.vdumem)
    {
     printf("vdu_switchvideo_out: video addr=0x%04x vb=0x%02x\n", video, vb);
     if (modio.level)
        fprintf(modio.log, "vdu_switchvideo_out: video addr=0x%04x vb=0x%02x\n", video, vb);
    }    
}

//==============================================================================
// Video memory read.
//
//   pass: UINT32 addr
//         struct MemoryReadByte *mem_s
// return: UINT8
//==============================================================================
UINT8 vdu_vidmem_r (UINT32 addr, struct MemoryReadByte *mem_s)
{
 if (modio.vdumem)
    {
     printf("vdu_vidmem_r: addr=0x%04x\n", (unsigned int)addr);
     if (modio.level)
        fprintf(modio.log, "vdu_vidmem_r: addr=0x%04x\n", (unsigned int)addr);
    }    

 if (addr & 0x0800)    // if PCG or Colour RAM
    return z80mem[addr & 0xFFFF];
 else
    {
     if (! latchrom)
        return z80mem[addr & 0xFFFF];
     else       
/* FIXME this will not work if disp_start is set to just before 0x2000, and the
   memory addresses would overlap 0x2000 */
        return charrom[((disp_start & 0x2000) >> 2) + (addr & 0x07FF)];
    }
}

//==============================================================================
// Video memory write.
//
// Up to version 3.4 of mz80 (asm) PUSH instructions will use the
// alternative memory structures (i.e stack SP set to video RAM) but CALL
// and RET will not!  So switching out video is stilled required for those
// cases as it is required for VDU refresh emulation.
//
//   pass: UINT32 addr
//         UINT8 data
//         struct MemoryWriteByte *mem_s
// return: void
//==============================================================================
void vdu_vidmem_w (UINT32 addr, UINT8 data, struct MemoryWriteByte *mem_s)
{
 int vb;

 if (extendram)		// if the alpha+ graphics is enabled
    vb = videobank;
 else
    vb = 0;

 if (modio.vdumem)
    {
     printf("vdu_vidmem_w: addr=0x%04x data=0x%02x vbank=%d\n", (unsigned int)addr, data, vb);
     if (modio.level)
        fprintf(modio.log, "vdu_vidmem_w: addr=0x%04x data=0x%02x vbank=%d\n", (unsigned int)addr, data, vb);
    }    

 if (addr & 0x0800)    
    {
     if (! colourram)
        {
         if (vb < modelx.pcg)
            {
             pcg_ram[vb * 0x0800 + (addr & 0x07FF)] = data;
             z80mem[addr & 0xFFFF] = data;
            }
        }    
     else
        {
         colour_ram[(vb & VDU_MASK3) * 0x0800 + (addr & 0x07FF)] = data; 
         z80mem[addr & 0xFFFF] = data;
        } 
     crtc_set_redraw();        
    }
 else
    {
     if (! latchrom)
        {
         if (attribram)
            attrib_ram[(vb & VDU_MASK0) * 0x0800 + (addr & 0x07FF)] = data;
         else   
            screen_ram[(vb & VDU_MASK0) * 0x0800 + (addr & 0x07FF)] = data;
         z80mem[addr & 0xFFFF] = data;
         crtc_redraw_char(addr, 1);
        }
    }
}

//==============================================================================
// Read Port 0x08 - colour control
//
//   pass: UINT16 port
//         struct z80PortRead *port_s
// return: UINT16			colour_cont
//==============================================================================
UINT16 vdu_colcont_r (UINT16 port, struct z80PortRead *port_s)
{
 if (modio.vdu)
    {
     printf("vdu_colcont_r: port=0x%02x (no data)\n", port); 
     if (modio.level)
        fprintf(modio.log, "vdu_colcont_r: port=0x%02x (no data)\n", port);
    }    

 return 0;
}

//==============================================================================
// Write Port 0x08 - colour control
//
// Must switch in either the PCG or Colour RAM banks to 8000/8800 or F000/F800
// depending on the port data.
//
//   pass: UINT16 port
//         UINT8 data
//         struct z80PortWrite *port_s
// return: void
//==============================================================================
void vdu_colcont_w (UINT16 port, UINT8 data, struct z80PortWrite *port_s)
{
 if (modelx.colour)
    {
     if (modio.vdu)
        {
         printf("vdu_colcont_w: port=0x%02x colour_cont=0x%02x colourram=0x%02x\n", port, colour_cont, colourram);
         if (modio.level)
            fprintf(modio.log, "vdu_colcont_w: port=0x%02x colour_cont=0x%02x colourram=0x%02x\n", port, colour_cont, colourram);
        }    

     colour_cont = data & B8(11111110);        

     if (x_colour_cont != colour_cont)
       {
        vdu_switchvideo_out();			// switch out current VDU
        colourram = colour_cont & B8(01000000);
        vdu_switchvideo_in();
        x_colour_cont = colour_cont;        
       } 
    } 
}

//==============================================================================
// Read Port 0x09 - Colour wait off
//
//   pass: UINT16 port
//         struct z80PortRead *port_s
// return: UNINT16			0
//==============================================================================
UINT16 vdu_colwait_r (UINT16 port, struct z80PortRead *port_s)
{
 if (modio.vdu)
    {
     printf("vdu_colwait_r: port=0x%02x (no data)\n", port);
     if (modio.level)
        fprintf(modio.log, "vdu_colwait_r: port=0x%02x (no data)\n", port);
    }    
     
 return 0;
}

//==============================================================================
// Write Port 0x09 - Colour wait off
//
//   pass: UINT16 port
//         UINT8 data
//         struct z80PortWrite *port_s
// return: void
//==============================================================================
void vdu_colwait_w (UINT16 port, UINT8 data, struct z80PortWrite *port_s)
{
 if (modio.vdu)
    {
     printf("vdu_colwait_w: port=0x%02x data=0x%02x\n", port, data);
     if (modio.level)
        fprintf(modio.log, "vdu_colwait_w: port=0x%02x data=0x%02x\n", port, data);
    }    
}

//==============================================================================
// Read Port 0x1C - LV DATA 
//
// Will only be called if emulating an alpha+ model.
//
//   pass: UINT16 port
//         struct z80PortRead *port_s
// return: UINT16			lv_dat
//==============================================================================
UINT16 vdu_lvdat_r (UINT16 port, struct z80PortRead *port_s)
{
 if (modio.vdu)
    {
     printf("vdu_lvdat_r: port=0x%02x lv_dat=0x%02x\n", port, lv_dat);
     if (modio.level)
        fprintf(modio.log, "vdu_lvdat_r: port=0x%02x lv_dat=0x%02x\n", port, lv_dat);
    }    

 if (port == 0x1C)
    return lv_dat;
 else
    return 0;   
}

//==============================================================================
// Write Port 0x1C - LV DATA
//
// Will only be called if emulating an alpha+ model.
//
//   pass: UINT16 port
//         UINT8 data
//         struct z80PortWrite *port_s
// return: void
//==============================================================================
void vdu_lvdat_w (UINT16 port, UINT8 data, struct z80PortWrite *port_s)
{
 if (modio.vdu)
    {
     printf("vdu_lvdat_w: port=0x%02x data=0x%02x\n", port, data);
     if (modio.level)
        fprintf(modio.log, "vdu_lvdat_w: port=0x%02x data=0x%02x\n", port, data);
    }    
 
 if (port != 0x1C)
    return;

 lv_dat = data & B8(10111111);			// port (0x1c) value

 if (x_lv_dat != lv_dat)
    {
     memmap_basicrom(lv_dat & B8(00100000));
     vdu_switchvideo_out();			// switch out current VDU
     videobank = lv_dat & B8(00001111);		// bank bits 0-3
     attribram = lv_dat & B8(00010000);		// attribute RAM select
     extendram = lv_dat & B8(10000000); 	// extended graphics select
     vdu_switchvideo_in();
     x_lv_dat = lv_dat;				// port (0x1c) value
    }
}

//==============================================================================
// Port 0x0B - Character ROM select
//
//   pass: UINT16 port
//         UINT8 data
//         struct z80PortWrite *port_s
// return: void
//==============================================================================
void vdu_latchrom (UINT16 port, UINT8 data, struct z80PortWrite *port_s)
{
 latchrom = data & B8(00000001);
 if (modio.vdu)
    {
     printf("vdu_latchrom: port=0x%02x latchrom=0x%02x\n", port, latchrom);
     if (modio.level)
        fprintf(modio.log, "vdu_latchrom: port=0x%02x latchrom=0x%02x\n", port, latchrom);
    }    
}
