//******************************************************************************
//*                                  uBee512                                   *
//*         An emulator for the Microbee Z80 FDD and ROM based models          *
//*                                                                            *
//*                                 z80 module                                 *
//*                                                                            *
//*                   Copyright (C) 2007-2008 Stewart Kay                      *
//******************************************************************************
//
// Provides an init/deinit and an interface to the MZ80 routines.
//
//==============================================================================
// ChangeLog (most recent entries are at top)
//==============================================================================
// v2.3.0 - 9 January 2008, S.J.Kay
// - Added z80_regdump() function to dump mz80context and Z80 CPU registers.
// - Moved 'struct mz80context z80' from z80_init() function to global area
//   so that the z80 structure is initialized with 0s.  Removed all the other
//   'struct mz80context z80' in other functions and use the local one instead.
// - Added Serial Comminications Controller (SCC) ports 0x68 and 0x69.
// - Removed the alpha+ conditional compilations as model type is determined
//   at run time using the new alphap member in the model_t structure.
// - Added Port interface to call the appropriate port handler code depending
//   on the model emulated.
// - Port 0x18-0x1B write had incorrect handler and was fixed with the other
//   port changes.
// - Added modio_t structure.
//
// v2.1.0 - 28 October 2007, S.J.Kay
// - Implement the modelx information structure.
// - Add clock speed port 0x09 for read.
//
// v2.0.0 - 13 October 2007, S.J.Kay
// - Ports enabled for the Teleterm model and RTC emulation.
//
// v1.4.0 - 26 September 2007, S.J.Kay
// - Boot address (boot_addr) is now dependent on the model being emulated.
// - Added mz80init() to the z80_init() function.  This does not affect 386
//   ASM code builds but is required for 'C' builds of MZ80.  The 'C' build
//   is not currently working on i386 systems and possibly others.
// - Added changes to error reporting.
//
// v1.2.0 - 19 August 2007, S.J.Kay
// - Added a special emulator port to access built in emulator commands.
//
// v1.0.0 - 1 August 2007, S.J.Kay
// - Added alpha+ (Premimum) and other ports with compile time options.
//
// v0.0.0 - 5 June 2007, S.J.Kay
// Start with "nanowasp" source distribution version 0.22. An emulator for the
// microbee 128k. Copyright (C) 2000-2003  David G. Churchill.

//==============================================================================
/*
 *  uBee512 - An emulator for the Microbee Z80 FDD and ROM based models. 
 *  Copyright (C) 2007-2008 Stewart Kay
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/ 
//==============================================================================

#include <stdio.h>
#include <stdlib.h>

#include "z80.h"
#include "ubee512.h"
#include "mz80.h"
#include "crtc.h"
#include "memmap.h"
#include "vdu.h"
#include "fdc.h"
#include "pio.h"
#include "paknet.h"
#include "rtc.h"
#include "keytc.h"
#include "clock.h"
#include "scc.h"
#include "function.h"

UINT16 z80_port_10_11_12_13_r (UINT16 port, struct z80PortRead *port_s);
UINT16 z80_port_14_16_17_r (UINT16 port, struct z80PortRead *port_s);
UINT16 z80_port_18_r (UINT16 port, struct z80PortRead *port_s);
UINT16 z80_port_19_r (UINT16 port, struct z80PortRead *port_s);
UINT16 z80_port_1A_r (UINT16 port, struct z80PortRead *port_s);
UINT16 z80_port_1B_r (UINT16 port, struct z80PortRead *port_s);
UINT16 z80_port_1C_1E_r (UINT16 port, struct z80PortRead *port_s);
UINT16 z80_port_1D_1F_r (UINT16 port, struct z80PortRead *port_s);
UINT16 z80_port_68_69_r (UINT16 port, struct z80PortRead *port_s);
UINT16 z80_unhandled_r (UINT16 port, struct z80PortRead *port_s);

void z80_port_10_11_12_13_w (UINT16 port, UINT8 data, struct z80PortWrite *port_s);
void z80_port_14_16_17_w (UINT16 port, UINT8 data, struct z80PortWrite *port_s);
void z80_port_18_w (UINT16 port, UINT8 data, struct z80PortWrite *port_s);
void z80_port_19_w (UINT16 port, UINT8 data, struct z80PortWrite *port_s);
void z80_port_1A_w (UINT16 port, UINT8 data, struct z80PortWrite *port_s);
void z80_port_1B_w (UINT16 port, UINT8 data, struct z80PortWrite *port_s);
void z80_port_1C_1E_w (UINT16 port, UINT8 data, struct z80PortWrite *port_s);
void z80_port_1D_1F_w (UINT16 port, UINT8 data, struct z80PortWrite *port_s);
void z80_port_68_69_w (UINT16 port, UINT8 data, struct z80PortWrite *port_s);
void z80_unhandled_w (UINT16 port, UINT8 data, struct z80PortWrite *port_s);

// The Z80 memory
char z80mem[Z80MEMSIZE];

struct mz80context z80;

extern int model;
extern model_t modelx;

extern modio_t modio;

//==============================================================================
// Define the read ports handlers structure.
//
// Read MZ80 text file documentation for an explanation of this structure.
//
// Any port if used requires an appropriate C module to be compiled in.  The
// first part of the function name indicates which C file the function can
// found in.  The C modules are:
//
// pio.c	Z80 PIO (PC speaker, serial, tape and parallel port).
// rtc.c	Real Time Clock.
// vdu.c	Colour control & waitoff, latch ROM, lvdat (premium graphics).
// paknet.c	NET and PAK selection.
// crtc.c	CRT 6845 emulation.
// soundchip.c	76489AN Sound chip emulation.
// keytc.c	Teleterm keyboard.
// fdc.c	Floppy Disk Controller - WD2793 emulation.
// scc.c	Serial SCC emulation.
// memmap.c	Memory map handler.
//==============================================================================
static struct z80PortRead z80_ports_r[] =
{
// Z80 PIO (speaker, serial, tape, and parallel port)
 { 0x00, 0x03, pio_r, NULL },
 { 0x10, 0x13, z80_port_10_11_12_13_r, NULL },

// RTC
 { 0x04, 0x04, rtc_r, NULL },
 { 0x07, 0x07, rtc_r, NULL },
 { 0x14, 0x14, z80_port_14_16_17_r, NULL },
 { 0x16, 0x17, z80_port_14_16_17_r, NULL },

// Colour control
 { 0x08, 0x08, vdu_colcont_r, NULL },
 { 0x18, 0x18, z80_port_18_r, NULL },

// CPU clock speed or Viatel video mode
 { 0x09, 0x09, clock_r, NULL },
 { 0x19, 0x19, z80_port_19_r, NULL },

// NET selection
 { 0x0A, 0x0A, paknet_nsel_r, NULL },
 { 0x1A, 0x1A, z80_port_1A_r, NULL },

// Teleterm keyboard
 { 0x1B, 0x1B, keytc_r, NULL },

// Video control
 { 0x0C, 0x0C, crtc_status, NULL },
 { 0x0D, 0x0D, crtc_data_r, NULL },
 { 0x0E, 0x0E, crtc_status, NULL },
 { 0x0F, 0x0F, crtc_data_r, NULL },  

 { 0x1C, 0x1C, z80_port_1C_1E_r, NULL },
 { 0x1D, 0x1D, z80_port_1D_1F_r, NULL },
 { 0x1E, 0x1E, z80_port_1C_1E_r, NULL }, 
 { 0x1F, 0x1F, z80_port_1D_1F_r, NULL },

// Floppy Disk Controller (FDC)
 { 0x40, 0x40, fdc_status, NULL },
 { 0x41, 0x41, fdc_track_r, NULL },
 { 0x42, 0x42, fdc_sect_r, NULL },
 { 0x43, 0x43, fdc_data_r, NULL },
 { 0x44, 0x44, fdc_status, NULL },
 { 0x45, 0x45, fdc_track_r, NULL },
 { 0x46, 0x46, fdc_sect_r, NULL },
 { 0x47, 0x47, fdc_data_r, NULL },
 { 0x48, 0x4B, fdc_ext_r, NULL },

// SCC
 { 0x68, 0x69, z80_port_68_69_r, NULL },

// Special uBee512 built in function commands
 { 0xFF, 0xFF, function_ubee_r, NULL },
 
 { 0x00, 0xFF, z80_unhandled_r, NULL }, 
 
 { -1, -1, NULL, NULL }
};

//==============================================================================
// Define the write ports handlers structure.
//
// Read MZ80 text file documentation for an explanation of this structure.
//
// Any port if used requires an appropriate C module to be compiled in.  The
// first part of the function name indicates which C file the function can
// found in.  The C modules are:
//
// pio.c	Z80 PIO (PC speaker, tape, serial, and parallel port).
// rtc.c	Real Time Clock.
// vdu.c	Colour control & waitoff, latch ROM, lvdat (premium graphics).
// paknet.c	NET and PAK selection.
// crtc.c	CRT 6845 emulation.
// soundchip.c	76489AN Sound chip emulation.
// keytc.c	Teleterm keyboard.
// fdc.c	Floppy Disk Controller - WD2793 emulation.
// memmap.c	Memory map handler.
// clock.c      Clock speed
// scc.c	Serial SCC emulation.
// function.c   Emulator functions.
//==============================================================================
static struct z80PortWrite z80_ports_w[] =
{
// Z80 PIO (speaker, serial, tape, and parallel port)
 { 0x00, 0x03, pio_w, NULL },
 { 0x10, 0x13, z80_port_10_11_12_13_w, NULL },

// RTC
 { 0x04, 0x04, rtc_w, NULL },
 { 0x06, 0x06, rtc_w, NULL },
 { 0x14, 0x14, z80_port_14_16_17_w, NULL },
 { 0x16, 0x17, z80_port_14_16_17_w, NULL },

// Colour control
 { 0x08, 0x08, vdu_colcont_w, NULL },
 { 0x18, 0x18, z80_port_18_w, NULL },

 // Colour waitoff
 { 0x09, 0x09, vdu_colwait_w, NULL },
 { 0x19, 0x19, z80_port_19_w, NULL },

// PAK selection
 { 0x0A, 0x0A, paknet_psel_w, NULL },
 { 0x1A, 0x1A, z80_port_1A_w, NULL },

// Teleterm keyboard
 { 0x1B, 0x1B, z80_port_1B_w, NULL },

// Video control 
 { 0x0B, 0x0B, vdu_latchrom, NULL },
 { 0x0C, 0x0C, crtc_address, NULL },
 { 0x0D, 0x0D, crtc_data_w, NULL },
 { 0x0E, 0x0E, crtc_address, NULL }, 
 { 0x0F, 0x0F, crtc_data_w, NULL },

 { 0x1C, 0x1C, z80_port_1C_1E_w, NULL },
 { 0x1D, 0x1D, z80_port_1D_1F_w, NULL },
 { 0x1E, 0x1E, z80_port_1C_1E_w, NULL }, 
 { 0x1F, 0x1F, z80_port_1D_1F_w, NULL },

// Floppy Disk Controller (FDC)
 { 0x40, 0x40, fdc_cmd, NULL },
 { 0x41, 0x41, fdc_track_w, NULL },
 { 0x42, 0x42, fdc_sect_w, NULL },
 { 0x43, 0x43, fdc_data_w, NULL },
 { 0x44, 0x44, fdc_cmd, NULL },
 { 0x45, 0x45, fdc_track_w, NULL },
 { 0x46, 0x46, fdc_sect_w, NULL },
 { 0x47, 0x47, fdc_data_w, NULL },
 { 0x48, 0x4B, fdc_ext_w, NULL },

// Memory map control
 { 0x50, 0x57, memmap_mode, NULL },

// SCC
 { 0x68, 0x69, z80_port_68_69_w, NULL },

// Special uBee512 built in function commands 
 { 0xFF, 0xFF, function_ubee_w, NULL },

 { 0x00, 0xFF, z80_unhandled_w, NULL },

 { -1, -1, NULL, NULL }
};

//==============================================================================
// Z80 Initilization
//
//   pass: void
// return: int				0
//==============================================================================
int z80_init (void)
{
 mz80init();
  
 z80.z80Base = (UINT8 *)z80mem;
 z80.z80IoRead = z80_ports_r;
 z80.z80IoWrite = z80_ports_w;

 z80.z80MemRead = NULL;
 z80.z80MemWrite = NULL;

// put interrupt setup stuff here
 mz80SetContext(&z80);

 return 0;
}

//==============================================================================
// Z80 De-Initilization
//
//   pass: void
// return: int				0
//==============================================================================
int z80_deinit (void)
{
 return 0;
}

//==============================================================================
// Z80 Reset
//
//   pass: void
// return: int				0
//==============================================================================
int z80_reset (void)
{
 mz80reset();
 mz80GetContext(&z80);
 z80.z80pc = modelx.bootaddr;			// start executing code here
 mz80SetContext(&z80);
 
 return 0;
}

//==============================================================================
// Z80 Get Program Counter (PC)
//
//   pass: void
// return: int				PC
//==============================================================================
int z80_getpc (void)
{
 mz80GetContext(&z80);
 return z80.z80pc;
}

//==============================================================================
// Read port: 10, 11, 12, 13
//
// decides what port action is required depending on the model emulated.
//
//   pass: UINT16 port
//         struct z80PortRead *port_s
// return: UINT16			0
//==============================================================================
UINT16 z80_port_10_11_12_13_r (UINT16 port, struct z80PortRead *port_s)
{
 if (! modelx.alphap)
    return pio_r(port, port_s);
 else   
#ifdef use_soundchip
    return soundchip_r(port, port_s);
#else
    return 0;
#endif    
}

//==============================================================================
// Read port: 14, 16, 17
//
// decides what port action is required depending on the model emulated.
//
//   pass: UINT16 port
//         struct z80PortRead *port_s
// return: UINT16			0
//==============================================================================
UINT16 z80_port_14_16_17_r (UINT16 port, struct z80PortRead *port_s)
{
 if (modelx.alphap)
    return rtc_r(port, port_s);
 else
    return rtc_r(port, port_s);
}

//==============================================================================
// Read port: 18
//
// decides what port action is required depending on the model emulated.
//
//   pass: UINT16 port
//         struct z80PortRead *port_s
// return: UINT16			0
//==============================================================================
UINT16 z80_port_18_r (UINT16 port, struct z80PortRead *port_s)
{
 if (modelx.alphap)
    {
     if (model != MOD_256TC)
        return vdu_colcont_r(port, port_s);
     else   
        return keytc_r(port, port_s);
    }   
 else
    return vdu_colcont_r(port, port_s);
}

//==============================================================================
// Read port: 19
//
// decides what port action is required depending on the model emulated.
//
//   pass: UINT16 port
//         struct z80PortRead *port_s
// return: UINT16			0
//==============================================================================
UINT16 z80_port_19_r (UINT16 port, struct z80PortRead *port_s)
{
 if (modelx.alphap)
    {
     if (model != MOD_256TC)
        return clock_r(port, port_s);
     else   
        return keytc_r(port, port_s);
    }   
 else
    return clock_r(port, port_s);
}

//==============================================================================
// Read port: 1A
//
// decides what port action is required depending on the model emulated.
//
//   pass: UINT16 port
//         struct z80PortRead *port_s
// return: UINT16			0
//==============================================================================
UINT16 z80_port_1A_r (UINT16 port, struct z80PortRead *port_s)
{
 if (modelx.rom)
    return paknet_nsel_r(port, port_s);
 else
    if (model == MOD_256TC)
       return keytc_r(port, port_s);
    else
       return 0;   
}   

//==============================================================================
// Read port: 1B
//
// decides what port action is required depending on the model emulated.
//
//   pass: UINT16 port
//         struct z80PortRead *port_s
// return: UINT16			0
//==============================================================================
UINT16 z80_port_1B_r (UINT16 port, struct z80PortRead *port_s)
{
 if (model == MOD_256TC)
    return keytc_r(port, port_s);
 else
    return 0;
}

//==============================================================================
// Read port: 1C 1E
//
// decides what port action is required depending on the model emulated.
//
//   pass: UINT16 port
//         struct z80PortRead *port_s
// return: UINT16			0
//==============================================================================
UINT16 z80_port_1C_1E_r (UINT16 port, struct z80PortRead *port_s)
{
 if (modelx.alphap)
    return vdu_lvdat_r(port, port_s);
 else
    return crtc_status(port, port_s);
}

//==============================================================================
// Read port: 1D 1F
//
// decides what port action is required depending on the model emulated.
//
//   pass: UINT16 port
//         struct z80PortRead *port_s
// return: UINT16			0
//==============================================================================
UINT16 z80_port_1D_1F_r (UINT16 port, struct z80PortRead *port_s)
{
 if (modelx.alphap)
    return vdu_lvdat_r(port, port_s);
 else
    return crtc_data_r(port, port_s);
}

//==============================================================================
// Read port: 68 69
//
// decides what port action is required depending on the model emulated.
//
//   pass: UINT16 port
//         struct z80PortRead *port_s
// return: UINT16			0
//==============================================================================
UINT16 z80_port_68_69_r (UINT16 port, struct z80PortRead *port_s)
{
 if (modelx.ram >= 64)
    return scc_r(port, port_s);
 else
    return z80_unhandled_r(port, port_s);
}

//==============================================================================
// Z80 Unhandled read port
//
// Not reported normally as this may be deliberate.  Reporting may affect the
// emulation speed of the program.
//
//   pass: UINT16 port
//         struct z80PortRead *port_s
// return: UINT16			0
//==============================================================================
UINT16 z80_unhandled_r (UINT16 port, struct z80PortRead *port_s)
{
 if (modio.z80)
    {
     printf("z80_unhandled_r: port=0x%02x\n", port);
     if (modio.level)
        fprintf(modio.log, "z80_unhandled_r: port=0x%02x\n", port);
    }    
 return 0;
}

//==============================================================================
// Write port: 10, 11, 12, 13
//
// decides what port action is required depending on the model emulated.
//
//   pass: UINT16 port
//         UINT8 data
//         struct z80PortWrite *port_s
// return: void
//==============================================================================
void z80_port_10_11_12_13_w (UINT16 port, UINT8 data, struct z80PortWrite *port_s)
{
 if (! modelx.alphap)
    pio_w(port, data, port_s); 
#ifdef use_soundchip
 else
    soundchip_w(port, data, port_s);
#endif
}

//==============================================================================
// Write port: 14, 16, 17
//
// decides what port action is required depending on the model emulated.
//
//   pass: UINT16 port
//         UINT8 data
//         struct z80PortWrite *port_s
// return: void
//==============================================================================
void z80_port_14_16_17_w (UINT16 port, UINT8 data, struct z80PortWrite *port_s)
{
 if (! modelx.alphap)
    rtc_w(port, data, port_s); 
 else
    rtc_w(port, data, port_s);
}

//==============================================================================
// Write port: 18
//
// decides what port action is required depending on the model emulated.
//
//   pass: UINT16 port
//         UINT8 data
//         struct z80PortWrite *port_s
// return: void
//==============================================================================
void z80_port_18_w (UINT16 port, UINT8 data, struct z80PortWrite *port_s)
{
 if (modelx.alphap)
    {
     if (model == MOD_256TC)
        keytc_w(port, data, port_s);
     else   
        vdu_colcont_w(port, data, port_s);
    }   
 else
    vdu_colcont_w(port, data, port_s);
}

//==============================================================================
// Write port: 19
//
// decides what port action is required depending on the model emulated.
//
//   pass: UINT16 port
//         UINT8 data
//         struct z80PortWrite *port_s
// return: void
//==============================================================================
void z80_port_19_w (UINT16 port, UINT8 data, struct z80PortWrite *port_s)
{
 if (modelx.alphap)
    {
     if (model == MOD_256TC)
        keytc_w(port, data, port_s);
     else    
        vdu_colwait_w(port, data, port_s);
    }   
 else
    vdu_colwait_w(port, data, port_s);
}

//==============================================================================
// Write port: 1A
//
// decides what port action is required depending on the model emulated.
//
//   pass: UINT16 port
//         UINT8 data
//         struct z80PortWrite *port_s
// return: void
//==============================================================================
void z80_port_1A_w (UINT16 port, UINT8 data, struct z80PortWrite *port_s)
{
 if (modelx.rom)
    paknet_psel_w(port, data, port_s);
 else
    if (model == MOD_256TC)
       keytc_w(port, data, port_s);
}   

//==============================================================================
// Write port: 1B
//
// decides what port action is required depending on the model emulated.
//
//   pass: UINT16 port
//         UINT8 data
//         struct z80PortWrite *port_s
// return: void
//==============================================================================
void z80_port_1B_w (UINT16 port, UINT8 data, struct z80PortWrite *port_s)
{
 if (modelx.alphap)
    {
     if (model == MOD_256TC)
        keytc_w(port, data, port_s);
     else    
        vdu_latchrom(port, data, port_s);
    }
 else
    vdu_latchrom(port, data, port_s);
}

//==============================================================================
// Write port: 1C, 1E
//
// decides what port action is required depending on the model emulated.
//
//   pass: UINT16 port
//         UINT8 data
//         struct z80PortWrite *port_s
// return: void
//==============================================================================
void z80_port_1C_1E_w (UINT16 port, UINT8 data, struct z80PortWrite *port_s)
{
 if (modelx.alphap)
    vdu_lvdat_w(port, data, port_s); 
 else
    crtc_address(port, data, port_s);
}

//==============================================================================
// Write port: 1D, 1F
//
// decides what port action is required depending on the model emulated.
//
//   pass: UINT16 port
//         UINT8 data
//         struct z80PortWrite *port_s
// return: void
//==============================================================================
void z80_port_1D_1F_w (UINT16 port, UINT8 data, struct z80PortWrite *port_s)
{
 if (modelx.alphap)
    vdu_lvdat_w(port, data, port_s); 
 else
    crtc_data_w(port, data, port_s);
}

//==============================================================================
// Write port: 68, 69
//
// decides what port action is required depending on the model emulated.
//
//   pass: UINT16 port
//         UINT8 data
//         struct z80PortWrite *port_s
// return: void
//==============================================================================
void z80_port_68_69_w (UINT16 port, UINT8 data, struct z80PortWrite *port_s)
{
 if (modelx.ram >= 64)
    scc_w(port, data, port_s); 
 else
    z80_unhandled_w(port, data, port_s);
}

//==============================================================================
// Z80 Unhandled write port
//
// Not reported normally as this may be deliberate.  Reporting may affect the
// emulation speed of the program.
//
// Applications                         port (hex)
// Astoroids Plus (tape conv to disk)   FD
//
//   pass: UINT16 port
//         UINT8 data
//         struct z80PortWrite *port_s
// return: void
//==============================================================================
void z80_unhandled_w (UINT16 port, UINT8 data, struct z80PortWrite *port_s)
{
 if (modio.z80)
    {
     printf("z80_unhandled_w: port=0x%02x, data=0x%02x\n", port, data);
     if (modio.level)
        fprintf(modio.log, "z80_unhandled_w: port=0x%02x, data=0x%02x\n", port, data);
    }    
}

//==============================================================================
// Z80 register dump
//
// Dump the contents of the Z80 registers and the current mz80context.
//
//   pass: void
// return: void
//==============================================================================
void z80_regdump (void)
{
 mz80GetContext(&z80);

 printf("\n");
 printf("mz80context and Z80 Registers       Contents\n");
 printf("----------------------------------------------\n");
 printf("%-22s 0x%08x (%dd)\n", "*z80Base", (uint32_t)z80.z80Base, (uint32_t)z80.z80Base);
 printf("%-22s 0x%08x (%dd)\n", "*z80MemRead", (uint32_t)z80.z80MemRead, (uint32_t)z80.z80MemRead);
 printf("%-22s 0x%08x (%dd)\n", "*z80MemWrite", (uint32_t)z80.z80MemWrite, (uint32_t)z80.z80MemWrite);
 printf("%-22s 0x%08x (%dd)\n", "*z80IoRead", (uint32_t)z80.z80IoRead, (uint32_t)z80.z80IoRead);
 printf("%-22s 0x%08x (%dd)\n", "*z80IoWrite", (uint32_t)z80.z80IoWrite, (uint32_t)z80.z80IoWrite);
 printf("\n");
 printf("%-22s 0x%08x (%dd)\n", "z80clockticks", (uint32_t)z80.z80clockticks, (uint32_t)z80.z80clockticks);
 printf("%-22s 0x%08x (%dd)\n", "z80iff", (uint32_t)z80.z80iff, (uint32_t)z80.z80iff); 
 printf("%-22s 0x%08x (%dd)\n", "z80interruptMode", (uint32_t)z80.z80interruptMode, (uint32_t)z80.z80interruptMode);
 printf("%-22s 0x%08x (%dd)\n", "z80halted", (uint32_t)z80.z80halted, (uint32_t)z80.z80halted);
 printf("%-22s 0x%08x (%dd)\n", "z80af.af", (uint32_t)z80.z80af.af, (uint32_t)z80.z80af.af);
 printf("%-22s 0x%08x (%dd)\n", "z80bc.bc", (uint32_t)z80.z80bc.bc, (uint32_t)z80.z80bc.bc); 
 printf("%-22s 0x%08x (%dd)\n", "z80de.de", (uint32_t)z80.z80de.de, (uint32_t)z80.z80de.de);
 printf("%-22s 0x%08x (%dd)\n", "z80hl.hl", (uint32_t)z80.z80hl.hl, (uint32_t)z80.z80hl.hl);
 printf("%-22s 0x%08x (%dd)\n", "z80afprime", (uint32_t)z80.z80afprime, (uint32_t)z80.z80afprime);
 printf("%-22s 0x%08x (%dd)\n", "z80bcprime", (uint32_t)z80.z80bcprime, (uint32_t)z80.z80bcprime);
 printf("%-22s 0x%08x (%dd)\n", "z80deprime", (uint32_t)z80.z80deprime, (uint32_t)z80.z80deprime); 
 printf("%-22s 0x%08x (%dd)\n", "z80hlprime", (uint32_t)z80.z80hlprime, (uint32_t)z80.z80hlprime);
 printf("%-22s 0x%08x (%dd)\n", "z80ix.ix", (uint32_t)z80.z80ix.ix, (uint32_t)z80.z80ix.ix);
 printf("%-22s 0x%08x (%dd)\n", "z80iy.iy", (uint32_t)z80.z80iy.iy, (uint32_t)z80.z80iy.iy);
 printf("%-22s 0x%08x (%dd)\n", "z80sp", (uint32_t)z80.z80sp, (uint32_t)z80.z80sp); 
 printf("%-22s 0x%08x (%dd)\n", "z80pc", (uint32_t)z80.z80pc, (uint32_t)z80.z80pc); 
 printf("%-22s 0x%08x (%dd)\n", "z80nmiAddr", (uint32_t)z80.z80nmiAddr, (uint32_t)z80.z80nmiAddr);
 printf("%-22s 0x%08x (%dd)\n", "z80intAddr", (uint32_t)z80.z80intAddr, (uint32_t)z80.z80intAddr);
 printf("%-22s 0x%08x (%dd)\n", "z80rCounter", (uint32_t)z80.z80rCounter, (uint32_t)z80.z80rCounter);
 printf("%-28s 0x%02x (%dd)\n", "z80i", z80.z80i, z80.z80i);
 printf("%-28s 0x%02x (%dd)\n", "z80r", z80.z80r, z80.z80r);
 printf("%-28s 0x%02x (%dd)\n", "z80intPending", z80.z80intPending, z80.z80intPending);
 printf("\n");
 printf("%-28s 0x%02x (%dd)\n", "z80af.half.a (reg A)", z80.z80af.half.a, z80.z80af.half.a);
 printf("%-28s 0x%02x (%dd)\n", "z80af.half.f (reg F)", z80.z80af.half.f, z80.z80af.half.f);
 printf("%-28s 0x%02x (%dd)\n", "z80bc.half.b (reg B)", z80.z80bc.half.b, z80.z80bc.half.b);
 printf("%-28s 0x%02x (%dd)\n", "z80bc.half.c (reg C)", z80.z80bc.half.c, z80.z80bc.half.c);
 printf("%-28s 0x%02x (%dd)\n", "z80de.half.d (reg D)", z80.z80de.half.d, z80.z80de.half.d);
 printf("%-28s 0x%02x (%dd)\n", "z80de.half.e (reg E)", z80.z80de.half.e, z80.z80de.half.e);
 printf("%-28s 0x%02x (%dd)\n", "z80hl.half.h (reg H)", z80.z80hl.half.h, z80.z80hl.half.h);
 printf("%-28s 0x%02x (%dd)\n", "z80hl.half.l (reg L)", z80.z80hl.half.l, z80.z80hl.half.l);
}
