//******************************************************************************
//*                                  uBee512                                   *
//*         An emulator for the Microbee Z80 FDD and ROM based models          *
//*                                                                            *
//*                           Z80 debugging module                             *
//*                                                                            *
//*                   Copyright (C) 2007-2008 Stewart Kay                      *
//******************************************************************************
//
// Provides debugging tools to aid in the development of the emulator.
//
//==============================================================================
// ChangeLog (most recent entries are at top)
//==============================================================================
// v2.3.0 - 23 January 2008, S.J.Kay
// - Created a new file to implement the Z80 debugging tools.

//==============================================================================
/*
 *  uBee512 - An emulator for the Microbee Z80 FDD and ROM based models. 
 *  Copyright (C) 2007-2008 Stewart Kay
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/ 
//==============================================================================

#include <stdio.h>
#include <stdlib.h>

#include "z80.h"
#include "mz80.h"
#include "z80debug.h"
#include "ubee512.h"

#include "macros.h"

//==============================================================================
// structures and variables
//==============================================================================
struct mz80context z80before;
struct mz80context z80;

static char z80dis[300];

typedef char *str_p;

static int z80pc_before;
static int z80pc_same;

extern int done;
extern debug_t debug;

//==============================================================================
// z80debug initialise
//
//   pass: void
// return: int 				0
//==============================================================================
int z80debug_init (void)
{
 debug.debugcount = -1;
 return 0;
}

//==============================================================================
// z80debug de-initialise
//
//   pass: void
// return: int 				0
//==============================================================================
int z80debug_deinit (void)
{
 return 0;
}

//==============================================================================
// z80debug reset
//
//   pass: void
// return: int 				0
//==============================================================================
int z80debug_reset (void)
{
 z80pc_same = 0;;
 return 0;
}

//==============================================================================
// z80debug disassembler.
//
// The original source code was by (c)1992 Markus Fritz.  I have made a lot
// of changes to it for this application. The source has been recoded to
// make it more compact and broken up into functional parts.
//
//   pass: void
// return: int 				0
//==============================================================================
#define mem16u op_1+(op_2 << 8)
#define memrel adr+2+(int8_t)z80mem[adr+1]

 char t[100];

 uint8_t op_a;
 uint8_t op_b;
 uint8_t op_c;
 
 uint8_t op_1;
 uint8_t op_2;

 static str_p single_reg[] = {"b","c","d","e","h","l","(hl)","a"};
 static str_p double_reg[] = {"bc","de","hl","sp"};
 static str_p cond_flags[] = {"nz","z","nc","c","po","pe","p","m"};
 static str_p arith_ops[] = {"add~a,","adc~a,","sub~","sbc~a,","and~","xor~","or~","cp~"};

 static str_p str_a[] = {"rlca","rrca","rla","rra","daa","cpl","scf","ccf"};
 static str_p str_b[] = {"rlc","rrc","rl","rr","sla","sra","???","srl"};
 static str_p str_c[] = {"neg","???","???","???","???","???","???","???"}; 
 static str_p str_d[] = {"retn","reti","???","???","???","???","???","???"};
 static str_p str_e[] = {"ld~i,a","???","ld~a,i","???","rrd","rld","???","???"};
 static str_p str_f[] = {"ldi","cpi","ini","outi","???","???","???","???",
                             "ldd","cpd","ind","outd","???","???","???","???",
                             "ldir","cpir","inir","otir","???","???","???","???",
                             "lddr","cpdr","indr","otdr","???","???","???","???"
                        }; 
 static str_p str_g[] = {"rlc","rrc","rl","rr","sla","sra","???","srl"};
 
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// void z80debug_dis_Op1andC0_x00 (int adr, str_p t)
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
void z80debug_dis_Op1andC0_x00 (int adr, str_p t)
{
 switch (op_c)
    {
     case 0x00: switch (op_b)
                   {
                    case 0x00: strcpy(t, "nop");
                               break;
                    case 0x01: strcpy(t, "ex~af,af'");
                               break;
                    case 0x02: sprintf(t, "djnz~%04xh", memrel);
                               break;
                    case 0x03: sprintf(t, "jr~%04xh", memrel);
                               break;
                      default: sprintf(t, "jr~%s,%04xh", cond_flags[op_b & B8(00000011)], memrel);
                               break;
                   }
                break;
     case 0x01: if (op_a & B8(00001000))
                   sprintf(t, "add~hl,%s", double_reg[op_b >> 1]);
                else
                   sprintf(t, "ld~%s,%04xh", double_reg[op_b >> 1], mem16u);
                break;
                               
     case 0x02: switch (op_b)
                   {
                    case 0x00: strcpy(t,"ld~(bc),a");
                               break;
                    case 0x01: strcpy(t,"ld~a,(bc)");
                               break;
                    case 0x02: strcpy(t,"ld~(de),a");
                               break;
                    case 0x03: strcpy(t,"ld~a,(de)");
                               break;
                    case 0x04: sprintf(t, "ld~(%04xh),hl", mem16u);
                               break;
                    case 0x05: sprintf(t, "ld~hl,(%04xh)", mem16u);
                               break;
                    case 0x06: sprintf(t, "ld~(%04xh),a", mem16u);
                               break;
                    case 0x07: sprintf(t, "ld~a,(%04xh)", mem16u);
                               break;
                   }
                break;

     case 0x03: if (op_a & B8(00001000))
                   sprintf(t, "dec~%s", double_reg[op_b >> 1]);
                else
                   sprintf(t, "inc~%s", double_reg[op_b >> 1]);                               
                break;
     case 0x04: sprintf(t, "inc~%s", single_reg[op_b]);
                break;
     case 0x05: sprintf(t, "dec~%s", single_reg[op_b]);
                break;
     case 0x06: sprintf(t, "ld~%s,%02xh", single_reg[op_b], op_1);
                break;
     case 0x07: strcpy(t, str_a[op_b]);
                break;
    }
}

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// void z80debug_dis_Op1andC0_x40 (int adr, str_p t)
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
void z80debug_dis_Op1andC0_x40 (int adr, str_p t)
{
 if (op_b == op_c)
    strcpy(t,"halt");
 else
    sprintf(t, "ld~%s,%s", single_reg[op_b], single_reg[op_c]);
}

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// void z80debug_dis_Op1andC0_x80 (int adr, str_p t)
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
void z80debug_dis_Op1andC0_x80 (int adr, str_p t)
{
 sprintf(t, "%s%s", arith_ops[op_b], single_reg[op_c]);
}

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// void z80debug_dis_Op1andC0_xC0_e_x03 (int adr, str_p t)
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
void z80debug_dis_Op1andC0_xC0_e_x03 (int adr, str_p t)
{
 switch (op_b)
    {
     case 0x00: sprintf(t, "jp~%04xh", mem16u);
                break;
     case 0x01: op_a = (uint8_t)z80mem[++adr];
                op_b = (op_a >> 3) & B8(00000111);
                op_c = op_a & B8(00000111);
                switch (op_a & B8(11000000))
                   {
                    case 0x00: sprintf(t, "%s~%s", str_b[op_b], single_reg[op_c]);
                               break;
                    case 0x40: sprintf(t, "bit~%d,%s", op_b, single_reg[op_c]);
                               break;
                    case 0x80: sprintf(t, "res~%d,%s", op_b, single_reg[op_c]);
                               break;
                    case 0xC0: sprintf(t, "set~%d,%s", op_b, single_reg[op_c]);
                               break;
                   }
                break;
     case 0x02: sprintf(t, "out~(%02xh),a", op_1);
                break;
     case 0x03: sprintf(t, "in~a,(%02xh)", op_1);
                break;
     case 0x04: strcpy(t,"ex~(sp),hl");
                break;
     case 0x05: strcpy(t,"ex~de,hl");
                break;
     case 0x06: strcpy(t,"di");
                break;
     case 0x07: strcpy(t,"ei");
                break;
    }
}

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// void z80debug_dis_Op1andC0_xC0_e_x05 (int adr, str_p t)
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
void z80debug_dis_Op1andC0_xC0_e_x05 (int adr, str_p t)
{
 char index_reg[3];

 if (op_b & B8(00000001))
    {
     switch (op_b >> 1)
        {
         case 0x00: sprintf(t, "call~%04xh", mem16u);
                    break;
         case 0x02: op_a = (uint8_t)z80mem[++adr];
                    op_b = (op_a >> 3) & B8(00000111);
                    op_c = op_a & B8(00000111);
                    switch (op_a & B8(11000000))
                       {
                        case 0x40: switch (op_c)
                                      {
                                       case 0x00: sprintf(t, "in~%s,(c)", single_reg[op_b]);
                                                  break;
                                       case 0x01: sprintf(t, "out~(c),%s", single_reg[op_b]);
                                                  break;
                                       case 0x02: if (op_b & B8(00000001))
                                                     sprintf(t, "adc~hl,%s", double_reg[op_b >> 1]);
                                                  else
                                                     sprintf(t, "sbc~hl,%s", double_reg[op_b >> 1]);                                                     
                                                  break;
                                       case 0x03: if (op_b & B8(00000001))
                                                     sprintf(t, "ld~%s,(%04xh)", double_reg[op_b >> 1], mem16u);
                                                  else
                                                     sprintf(t, "ld~(%04xh),%s", mem16u, double_reg[op_b >> 1]);
                                                  break;
                                       case 0x04: strcpy(t, str_c[op_b]);
                                                  break;
                                       case 0x05: strcpy(t, str_d[op_b]);
                                                  break;
                                       case 0x06: sprintf(t, "im~%d", op_b-1);
                                                  break;
                                       case 0x07: strcpy(t, str_e[op_b]);
                                                  break;
                                      }
                                   break;
                        case 0x80: strcpy(t, str_f[op_a & B8(00011111)]);
                                   break;
                       }
                    break;
           default: strcpy(index_reg, (op_a & B8(00100000))?"iy":"ix");
                    op_a = (uint8_t)z80mem[++adr];
                    switch (op_a)
                       {
                        case 0x09: sprintf(t, "add~%s,bc", index_reg);
                                   break;
                        case 0x19: sprintf(t, "add~%s,de", index_reg);
                                   break;
                        case 0x21: sprintf(t, "ld~%s,%04xh", index_reg, mem16u);
                                   break;
                        case 0x22: sprintf(t, "ld~(%04xh),%s", mem16u, index_reg);
                                   break;
                        case 0x23: sprintf(t, "inc~%s", index_reg);
                                   break;
                        case 0x29: sprintf(t, "add~%s,%s", index_reg, index_reg);
                                   break;
                        case 0x2A: sprintf(t, "ld~%s,(%04xh)", index_reg, mem16u);
                                   break;
                        case 0x2B: sprintf(t, "dec~%s", index_reg);
                                   break;
                        case 0x34: sprintf(t, "inc~(%s+%02xh)", index_reg, op_1);
                                   break;
                        case 0x35: sprintf(t, "dec~(%s+%02xh)", index_reg, op_1);
                                   break;
                        case 0x36: sprintf(t, "ld~(%s+%02xh),%02xh", index_reg, op_1, op_2);
                                   break;
                        case 0x39: sprintf(t, "add~%s,sp", index_reg);
                                   break;
                        case 0x46:
                        case 0x4E:
                        case 0x56:
                        case 0x5E:
                        case 0x66:
                        case 0x6E: sprintf(t, "ld~%s,(%s+%02xh)", single_reg[(op_a >> 3) & B8(00000111)], index_reg, op_1);
                                   break;
                        case 0x70:
                        case 0x71:
                        case 0x72:
                        case 0x73:
                        case 0x74:
                        case 0x75:
                        case 0x77: sprintf(t, "ld~(%s+%02xh),%s", index_reg, op_1, single_reg[op_a & B8(00000111)]);
                                   break;
                        case 0x7E: sprintf(t, "ld~a,(%s+%02xh)", index_reg, op_1);
                                   break;
                        case 0x86: sprintf(t, "add~a,(%s+%02xh)", index_reg, op_1);
                                   break;
                        case 0x8E: sprintf(t, "adc~a,(%s+%02xh)", index_reg, op_1);
                                   break;
                        case 0x96: sprintf(t, "sub~,(%s+%02xh)", index_reg, op_1);
                                   break;
                        case 0x9E: sprintf(t, "sbc~a,(%s+%02xh)", index_reg, op_1);
                                   break;
                        case 0xA6: sprintf(t, "and~a,(%s+%02xh)", index_reg, op_1);
                                   break;
                        case 0xAE: sprintf(t, "xor~a,(%s+%02xh)", index_reg, op_1);
                                   break;
                        case 0xB6: sprintf(t, "or~a,(%s+%02xh)", index_reg, op_1);
                                   break;
                        case 0xBE: sprintf(t, "cp~a,(%s+%02xh)", index_reg, op_1);
                                   break;
                        case 0xE1: sprintf(t, "pop~%s", index_reg);
                                   break;
                        case 0xE3: sprintf(t, "ex~(sp),%s", index_reg);
                                   break;
                        case 0xE5: sprintf(t, "push~%s", index_reg);
                                   break;
                        case 0xE9: sprintf(t, "jp~(%s)", index_reg);
                                   break;
                        case 0xF9: sprintf(t, "ld~sp,%s", index_reg);
                                   break;
                        case 0xCB: op_a = op_2;
                                   op_b = (op_a >> 3) & B8(00000111);
                                   switch (op_a & B8(11000000))
                                      {
                                       case 0x00: sprintf(t, "%s~(%s+%02xh)", str_g[op_b], index_reg, op_1);
                                                  break;
                                       case 0x40: sprintf(t, "bit~%d,(%s+%02xh)", op_b, index_reg, op_1);
                                                  break;
                                       case 0x80: sprintf(t, "res~%d,(%s+%02xh)", op_b, index_reg, op_1);
                                                  break;
                                       case 0xC0: sprintf(t, "set~%d,(%s+%02xh)", op_b, index_reg, op_1);
                                                  break;
                                      }
                                   break;
                       }
                    break;
        }
    }
 else
    {
     if ((op_b >> 1) == B8(00000011))
        sprintf(t, "push~af");
     else
        sprintf(t, "push~%s", double_reg[op_b >> 1]);     
    }
}

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// void z80debug_dis_Op1andC0_xC0 (int adr, str_p t)
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
void z80debug_dis_Op1andC0_xC0 (int adr, str_p t)
{
 switch (op_c)
    {
     case 0x00: sprintf(t, "ret~%s", cond_flags[op_b]);
                break;
     case 0x01: if (op_b & B8(00000001))
                   {
                    switch (op_b >> 1)
                       {
                        case 0x00: strcpy(t, "ret");
                                   break;
                        case 0x01: strcpy(t, "exx");
                                   break;
                        case 0x02: strcpy(t, "jp~(hl)");
                                   break;
                        case 0x03: strcpy(t, "ld~sp,hl");
                                   break;
                       }
                   }
                else
                   {
                    strcpy(t, "pop~");
                    if ((op_b >> 1) == B8(00000011))
                       strcat(t, "af");
                    else
                       strcat(t, double_reg[op_b >> 1]);
                   }
                break;
     case 0x02: sprintf(t, "jp~%s,%04xh", cond_flags[op_b], mem16u);
                break;
     case 0x03: z80debug_dis_Op1andC0_xC0_e_x03(adr, t);
                break;
     case 0x04: sprintf(t, "call~%s,%04xh", cond_flags[op_b], mem16u);
                break;
     case 0x05: z80debug_dis_Op1andC0_xC0_e_x05(adr, t);
                break;
     case 0x06: sprintf(t, "%s%02xh", arith_ops[op_b], op_1);
                break;
     case 0x07: sprintf(t, "rst~%02xh", op_a & B8(00111000));
                break;
    }
}

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// void z80debug_disassemble (int adr, str_p dis)
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
void z80debug_disassemble (int adr, str_p dis)
{
 static char t[40];
 
 op_a = (uint8_t)z80mem[adr];
 op_b = (op_a >> 3) & B8(00000111);
 op_c = op_a & B8(00000111);

 op_1 = (uint8_t)z80mem[adr+1]; 
 op_2 = (uint8_t)z80mem[adr+2];

 switch (op_a & B8(11000000))
    {
     case 0x00: z80debug_dis_Op1andC0_x00(adr, t);
                break;
     case 0x40: z80debug_dis_Op1andC0_x40(adr, t);
                break;
     case 0x80: z80debug_dis_Op1andC0_x80(adr, t);
                break;
     case 0xC0: z80debug_dis_Op1andC0_xC0(adr, t);
                break;
    }

 sprintf(dis, "%04x:~%s", (uint16_t)adr, t);    
}

//==============================================================================
// z80debug before instruction execution
//
// - dissasemble instruction at current Z80 PC.  This is placed into a string
//   and is output in the z80debug_after function.
//
//   pass: void
// return: int				minimum cycle count needed
//==============================================================================
int z80debug_before (void)
{
 mz80GetContext(&z80before);

// test and do the disassembly only if needed
 if ((debug.mode == 2) || ((debug.mode == 1) && (debug.step)))
    z80debug_disassemble(z80before.z80pc, z80dis);

// if step mode and nothing to do then return (no code execution)
 if ((debug.mode == 1) && (! debug.step))
    return -1;

// check for code break points in code execution
 if (debug.mode != 1)	// if not in step mode
    {
     if (debug.break_point[z80before.z80pc])
        {
         printf("Z80 Debugging break point at PC: 0x%04x\n", (uint32_t)z80before.z80pc);
         debug.break_point[z80before.z80pc] = 0;
         debug.mode = 1;  // change to step mode
         debug.step = 0;
         return -1;
        }

     if ((debug.debugcount == debug.break_point_count) && debug.debugcount)
        {
         printf("Z80 Debugging break point (counter=%d) at PC: 0x%04x\n", debug.debugcount, (uint32_t)z80before.z80pc);
         debug.mode = 1;  // change to step mode
         debug.step = 0;
         return -1;
        }
    }    

 z80pc_before = z80before.z80pc;
 z80pc_same++;
 debug.debugcount++; 

 return 0;
}

//==============================================================================
// z80debug after instruction execution
//
// - show state of all registers.
//
//   pass: void
// return: void
//==============================================================================
void z80debug_after (int dwResult)
{
 int x;

 int i;
 int count;
 int begin;
 int l;

 uint32_t SP_C;
 
 static char dis[20];
 static char flags[9];

 if ((debug.mode == 2) || ((debug.mode == 1) && (debug.step)))
    {
     mz80GetContext(&z80); 

// don't keep showing same address unless bigger than max BC value
     if ((z80pc_before == z80.z80pc) && (z80pc_same < 66000L))
        return;
     z80pc_same = 0;   
     
     if (debug.step)
        debug.step--;

     uint32_t AF = z80.z80af.af;
     uint32_t BC = z80.z80bc.bc;
     uint32_t DE = z80.z80de.de;
     uint32_t HL = z80.z80hl.hl;
     uint32_t AFp = z80.z80afprime;
     uint32_t BCp = z80.z80bcprime;
     uint32_t DEp = z80.z80deprime;
     uint32_t HLp = z80.z80hlprime;

     if (dwResult != 0x80000000)
        {
         printf("Invalid instruction at: 0x%04x\n", (unsigned int)z80.z80pc);
         printf("     Opcodes (first 4): 0x%02x 0x%02x 0x%02x 0x%02x\n",
         (uint8_t)z80mem[z80.z80pc+0], (uint8_t)z80mem[z80.z80pc+1],
         (uint8_t)z80mem[z80.z80pc+2], (uint8_t)z80mem[z80.z80pc+3]);
         done = 1;
        }   

// show the instruction counter if enabled
     if (debug.showcount)
         printf("%08x ", debug.debugcount);

// break the disassembled code up for columned output
     l = strlen(z80dis);
     i = 0;

     for (x = 0; x < 3; x++)
        {
         count = 0;
         begin = i;
         dis[0] = 0;
         while ((i < l) && (z80dis[i] != '~'))
            dis[count++] = z80dis[i++];
         dis[count] = 0;
         i++;	 // skip pass the ~ character for next iteration
         if (x < 2)
            printf("%-8s", dis);
         else   
            printf("%-16s", dis);        
        }

// show the status of the common registers     
     flags[0] = AF & B8(00000001)?'C':'-';
     flags[1] = AF & B8(00000010)?'N':'-';     
     flags[2] = AF & B8(00000100)?'P':'-';
     flags[3] = AF & B8(00001000)?'-':'-';        
     flags[4] = AF & B8(00010000)?'A':'-';
     flags[5] = AF & B8(00100000)?'-':'-';
     flags[6] = AF & B8(01000000)?'Z':'-';
     flags[7] = AF & B8(10000000)?'S':'-';
     flags[8] = 0;
     printf("FLAGS: %s AF:%04x BC:%04x DE:%04x HL:%04x", flags, AF, BC, DE, HL);

// show the index and SP registers
     if (debug.show & B8(00000001))
        {
         uint32_t IX = z80.z80ix.ix;     
         uint32_t IY = z80.z80iy.iy;
         uint32_t SP = z80.z80sp;

         SP_C = ((uint8_t)(z80mem[SP+1]) << 8) | (uint8_t)z80mem[SP]; // fetch the previous contents
         printf(" IX:%04x IY:%04x SP:%04x (%04x)\n", IX, IY, SP, SP_C);
        }
     else
        printf("\n");     

// show the alternate and I and R registers
     if (debug.show & B8(00000010))
        {
         if (debug.showcount)
            printf("%57s", "Alternate ");
         else   
            printf("%48s", "Alternate ");
         printf("AF:%04x BC:%04x DE:%04x HL:%04x I:%02x R:%02x\n", AFp, BCp, DEp, HLp, z80.z80i, z80.z80r);
        } 
    }
}
